/* tc-nios.c -- Assembler for the Altera NIOS.
   Copyright (C) 1998, 1999 Free Software Foundation.

   This file is part of GAS, the GNU Assembler.

   GAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GAS; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <ctype.h>
#include "as.h"
#include "dwarf2dbg.h"
#include "subsegs.h"     
#include "symcat.h"
#include "opcodes/nios-desc.h"
#include "opcodes/nios-opc.h"
#include "cgen.h"
#include "elf/common.h"
#include "elf/nios.h"
#include "sb.h"
#include "hash.h"
#include "macro.h"

/* Structure to hold all of the different components describing
   an individual instruction.  */
typedef struct
{
  const CGEN_INSN *	insn;
  const CGEN_INSN *	orig_insn;
  CGEN_FIELDS		fields;
#if CGEN_INT_INSN_P
  CGEN_INSN_INT         buffer [1];
#define INSN_VALUE(buf) (*(buf))
#else
  unsigned char         buffer [CGEN_MAX_INSN_SIZE];
#define INSN_VALUE(buf) (buf)
#endif
  char *		addr;
  fragS *		frag;
  int                   num_fixups;
  fixS *                fixups [GAS_CGEN_MAX_FIXUPS];
  int                   indices [MAX_OPERAND_INSTANCES];
}
nios_insn;

const char comment_chars[]        = ";";
const char line_comment_chars[]   = "#";
const char line_separator_chars[] = ""; 
const char EXP_CHARS[]            = "eE";
const char FLT_CHARS[]            = "dD";

/* describe the formal arguments to a macro */

typedef struct formal_struct
  {
    struct formal_struct *next;	/* next formal in list */
    sb name;			/* name of the formal */
    sb def;			/* the default value */
    sb actual;			/* the actual argument (changed on each expansion) */
    int index;			/* the index of the formal 0..formal_count-1 */
  }
formal_entry;

/* Other values found in the index field of a formal_entry.  */
#define QUAL_INDEX (-1)
#define NARG_INDEX (-2)
#define LOCAL_INDEX (-3)

/* describe the macro. */

typedef struct macro_struct
  {
    sb sub;			/* substitution text. */
    int formal_count;		/* number of formal args. */
    formal_entry *formals;	/* pointer to list of formal_structs */
    struct hash_control *formal_hash; /* hash table of formals. */
  }
macro_entry;

/* Default machine */

#define DEFAULT_MACHINE bfd_mach_nios16
#define DEFAULT_FLAGS	EF_NIOS_CPU_16

static unsigned long nios_mach = bfd_mach_nios16;
static int cpu_mach = (1 << MACH_NIOS16);

/* Flags to set in the elf header */
static flagword nios_flags = DEFAULT_FLAGS;

static int nios_user_set_flags_p = 0;

/* temporary place to store real insert operand fn ptr */
const char * (*real_insert_operand)
     PARAMS ((CGEN_CPU_DESC, int opindex_, CGEN_FIELDS *fields_,
              CGEN_INSN_BYTES_PTR, bfd_vma pc_));

/* Forward reference to static functions */
static const char *nios_insert_operand_hook         PARAMS ((CGEN_CPU_DESC, int opindex_, 
							      CGEN_FIELDS *fields_,
							      CGEN_INSN_BYTES_PTR, bfd_vma pc_));  
static void nios_set_flags		             PARAMS ((int));
static void nios_nword 		             PARAMS ((int));
static void nios_word  		             PARAMS ((int));
static void nios_uanword 		             PARAMS ((int));
static void nios_uaword  		             PARAMS ((int));
static void nios_add_macro                          PARAMS ((char *, char *, char **));
static void nios_load_macros                        PARAMS ((void));
static int  log2                                     PARAMS ((int));

/* routine to get macro hash table so we can add to it */
extern struct hash_control *get_macro_hash();

/* The target specific pseudo-ops which we support.  */
const pseudo_typeS md_pseudo_table[] =
{
  { "eflags",	nios_set_flags,		0 },
  { "nword",    nios_nword,            0 },
  { "word",     nios_word,             4 },
  { "long",     nios_word,             4 },
  { "byte4",    nios_word,             4 },
  { "half",     nios_word,             2 },
  { "hword",    nios_word,             2 },
  { "short",    nios_word,             2 },
  { "byte2",    nios_word,             2 },
  { "byte",     cons,                   1 },
  { "uanword",  nios_uanword,          0 },
  { "uaword",   nios_uaword,           4 },
  { "ualong",   nios_uaword,           4 },
  { "uabyte4",  nios_uaword,           4 },
  { "uahalf",   nios_uaword,           2 },
  { "uashort",  nios_uaword,           2 },
  { "uabyte2",  nios_uaword,           2 },
  { "file",	dwarf2_directive_file,	0 },
  { "loc",	dwarf2_directive_loc,	0 },
  { NULL, 	NULL,			0 }
};

/* relaxation tables */
relax_typeS *md_relax_table;

relax_typeS md_m32_relax_table[] = 
{
/* The fields are:
   1) most positive reach of this state,
   2) most negative reach of this state,
   3) how many bytes this mode will add to the size of the current frag
   4) which index into the table to try if we can't fit into this one.  */

  /* The first entry must be unused because an `rlx_more' value of zero ends
     each list.  */
  {1, 1, 0, 0},

  /* The displacement used by GAS is from the end of the 2 byte insn,
     so we subtract 2 as appropriate.  */
  /* static function bsr - only generated by compiler. */
  {2046, -2048, 0, 2 },
  /* call */
  {0x7fffffff, 0x80000000, 8, 0 },
  /* Same thing, but with leading nop for alignment.  */
};

relax_typeS md_m16_relax_table[] = 
{
/* The fields are:
   1) most positive reach of this state,
   2) most negative reach of this state,
   3) how many bytes this mode will add to the size of the current frag
   4) which index into the table to try if we can't fit into this one.  */

  /* The first entry must be unused because an `rlx_more' value of zero ends
     each list.  */
  {1, 1, 0, 0},

  /* The displacement used by GAS is from the end of the 2 byte insn,
     so we subtract 2 as appropriate.  */
  /* static function bsr - only generated by compiler. */
  {2046, -2048, 0, 2 },
  /* call */
  {0x7fff, 0x8000, 4, 0 },
};



/* assembler options */
#define OPTION_CPU_16		(OPTION_MD_BASE)
#define OPTION_CPU_32		(OPTION_MD_BASE + 1)

struct option md_longopts[] =
{  
  { "m16",         no_argument,	      NULL, OPTION_CPU_16     },
  { "m32",         no_argument,	      NULL, OPTION_CPU_32     },
  { NULL,          no_argument,       NULL, 0                 },
};

const CGEN_INSN *pfx_insn;
const CGEN_INSN *movi_insn;
const CGEN_INSN *movhi_insn;

size_t md_longopts_size = sizeof (md_longopts);
#define NIOS_SHORTOPTS ""
const char * md_shortopts = NIOS_SHORTOPTS;

int
md_parse_option (c, arg)
     int    c;
     char * arg;
{
  switch (c)
    {
    case OPTION_CPU_16:
      nios_flags = (nios_flags & ~EF_NIOS_CPU_MASK) | EF_NIOS_CPU_16;
      nios_mach = bfd_mach_nios16;
      cpu_mach = (1 << MACH_NIOS16);
      break;

    case OPTION_CPU_32:
      nios_flags = (nios_flags & ~EF_NIOS_CPU_MASK) | EF_NIOS_CPU_32;
      nios_mach = bfd_mach_nios32;
      cpu_mach = (1 << MACH_NIOS32);
      break;

    default:
      return 0;

    }

  return 1;
}

void
md_show_usage (stream)
  FILE * stream;
{
  fprintf (stream, _("NIOS specific command line options:\n"));
  fprintf (stream, _("-m16                       Nios-16 processor (16-bit)\n"));
  fprintf (stream, _("-m32                       Nios-32 processor (32-bit)\n"));
} 


void
md_begin ()
{
  flagword applicable;
  segT     seg;
  subsegT  subseg;
  CGEN_INSN_LIST *ilist;
  char *movi_name = NULL;

  /* Initialize the `cgen' interface.  */
  
  /* Set the machine number and endian.  */
  gas_cgen_cpu_desc = nios_cgen_cpu_open (CGEN_CPU_OPEN_MACHS, cpu_mach,
					 CGEN_CPU_OPEN_ENDIAN,
					 CGEN_ENDIAN_LITTLE,
					 CGEN_CPU_OPEN_END);
  nios_cgen_init_asm (gas_cgen_cpu_desc);

  /* override insert_normal handler to allow special checking */
#if 0
  real_insert_operand = gas_cgen_cpu_desc->insert_operand;
  gas_cgen_cpu_desc->insert_operand = &nios_insert_operand_hook;
#endif

  /* This is a callback from cgen to gas to parse operands.  */
  cgen_set_parse_operand_fn (gas_cgen_cpu_desc, gas_cgen_parse_operand);

  /* Set the ELF flags if desired. */
  if (nios_flags)
    bfd_set_private_flags (stdoutput, nios_flags);

  /* Set the machine type */
  bfd_default_set_arch_mach (stdoutput, bfd_arch_nios, nios_mach);

  /* Set the correct relaxation table. */
  if (nios_mach == bfd_mach_nios32)
    {
      md_relax_table = md_m32_relax_table;
      movi_name = "MOVI32";
    }
  else
    {
      md_relax_table = md_m16_relax_table;
      movi_name = "MOVI16";
    }

  /* Set up instruction addresses for use in relaxation code. */

  /* pfx instruction */
  ilist = cgen_asm_lookup_insn (gas_cgen_cpu_desc, "PFX");

  for ( ; ilist != NULL; CGEN_ASM_NEXT_INSN(ilist))
    {
      const CGEN_INSN *insn = ilist->insn;
      if (strcmp (insn->base->name, "PFX") == 0)
	{
	  pfx_insn = insn;
	  break;
	}
    }

  /* movi instruction */
  ilist = cgen_asm_lookup_insn (gas_cgen_cpu_desc, "MOVI");

  for ( ; ilist != NULL; CGEN_ASM_NEXT_INSN(ilist))
    {
      const CGEN_INSN *insn = ilist->insn;
      if (strcmp (insn->base->name, movi_name) == 0)
	{
	  movi_insn = insn;
	  break;
	}
    }

  /* movhi instruction if we are on m32 */
  if (nios_mach == bfd_mach_nios32)
    {
      ilist = cgen_asm_lookup_insn (gas_cgen_cpu_desc, "MOVHI");
      
      for ( ; ilist != NULL; CGEN_ASM_NEXT_INSN(ilist))
	{
	  const CGEN_INSN *insn = ilist->insn;
	  if (strcmp (insn->base->name, "MOVHI") == 0)
	    {
	      movhi_insn = insn;
	      break;
	    }
	}
    }

  nios_load_macros();
}

void
md_assemble (str)
     char * str;
{
  static     nios_prefix_seen = 0;
  nios_insn insn;
  char *     errmsg;
  char *     str2 = NULL;
  char       errbuf[150];

  /* Initialize GAS's cgen interface for a new instruction.  */
  gas_cgen_init_parse ();

  nios_Rbi5 = 0;

  insn.insn = nios_cgen_assemble_insn
    (gas_cgen_cpu_desc, str, & insn.fields, insn.buffer, & errmsg);


  if (!insn.insn)
    {
      nios_prefix_seen = 0;
      as_bad ("%s", errmsg);
      return;
    }

  /* code to catch Rbi5 instructions which are specified incorrectly */
  if (nios_Rbi5)
    {
      errmsg = NULL;
      if (nios_Rbi5 == NIOS_RBI5_REGISTER && nios_prefix_seen)
	errmsg =  _("prefixed instruction expects immediate data");
      else if (nios_Rbi5 == NIOS_RBI5_IMMEDIATE && !nios_prefix_seen)
	errmsg =  _("non-prefixed instruction expects two registers");

      nios_prefix_seen = 0;
      if (errmsg != NULL)
	{
	  if (strlen (str) > 50)
	    sprintf (errbuf, "%s `%.50s...'", errmsg, str);
	  else 
	    sprintf (errbuf, "%s `%.50s'", errmsg, str);
	  as_bad ("%s", errbuf);
	  return;
	}
    }

  if (nios_parsed_i11)
    nios_prefix_seen = 1;
  else
    nios_prefix_seen = 0;

  nios_parsed_i11 = 0;
  
  /* Doesn't really matter what we pass for RELAX_P here.  */
  gas_cgen_finish_insn (insn.insn, insn.buffer,
			CGEN_FIELDS_BITSIZE (& insn.fields), 1, NULL);
}

/* The syntax in the manual says constants begin with '#'.
   We just ignore it.  */

void 
md_operand (expressionP)
     expressionS * expressionP;
{
  if (* input_line_pointer == '#')
    {
      input_line_pointer ++;
      expression (expressionP);
    }
}

valueT
md_section_align (segment, size)
     segT   segment;
     valueT size;
{
  int align = bfd_get_section_alignment (stdoutput, segment);
  return ((size + (1 << align) - 1) & (-1 << align));
}

symbolS *
md_undefined_symbol (name)
  char * name;
{
  return 0;
}

/* Interface to relax_segment.  */


/* Return an initial guess of the length by which a fragment must grow to
   hold a branch to reach its destination.
   Also updates fr_type/fr_subtype as necessary.

   Called just before doing relaxation.
   Any symbol that is now undefined will not become defined.
   The guess for fr_var is ACTUALLY the growth beyond fr_fix.
   Whatever we do to grow fr_fix or fr_var contributes to our returned value.
   Although it may not be explicit in the frag, pretend fr_var starts with a
   0 value.  */

int
md_estimate_size_before_relax (fragP, segment)
     fragS * fragP;
     segT    segment;
{
  int    old_fr_fix = fragP->fr_fix;
  const CGEN_INSN * insn = fragP->fr_cgen.insn;

  /* We want to look for bsrr virtual instructions (only relaxable instruction).
     If the symbol isn't in the segment, then we want to generate a call instruction
     by default.  Otherwise, we will attempt to let relaxation take its course. */
  if (CGEN_INSN_ATTR_VALUE (insn, CGEN_INSN_RELAXABLE))
    {
      /* if symbol isn't in segment, then we will have to generate a call sequence by default. */
      if (S_GET_SEGMENT (fragP->fr_symbol) != segment)
	{
	  fragP->fr_subtype = 2;
	  if (nios_mach == bfd_mach_nios32)
	    return 8;
	  else
	    return 4;
	}
      else /* otherwise, we might be able to use a bsr */
	fragP->fr_subtype = 1;
    }
  
  return (fragP->fr_var + fragP->fr_fix - old_fr_fix);
} 

/* *fragP has been relaxed to its final size, and now needs to have
   the bytes inside it modified to conform to the new size.

   Called after relaxation is finished.
   fragP->fr_type == rs_machine_dependent.
   fragP->fr_subtype is the subtype of what the address relaxed to.  */

void
md_convert_frag (abfd, sec, fragP)
  bfd *   abfd;
  segT    sec;
  fragS * fragP;
{
  int   addend;
  int   target_address;
  int   opcode_address;
  char *opcode;

  subseg_change (sec, 0);
  opcode = fragP->fr_opcode;

  /* Address opcode resides at in file space.  */
  opcode_address = fragP->fr_address + fragP->fr_fix - 2;

  if (fragP->fr_subtype == 1)
    {
      /* we have bsr in range so we can set the address */
      fragP->fr_cgen.opindex = NIOS_OPERAND_I11;

      target_address = S_GET_VALUE (fragP->fr_symbol) + fragP->fr_offset;
      target_address += symbol_get_frag (fragP->fr_symbol)->fr_address;
      addend = ((target_address - opcode_address) - 2) >> 1;
      addend &= 0x7ff;  // handle negative offsets
      addend |= 0x8800;
      md_number_to_chars (opcode, addend, 2);
    }
  else if (fragP->fr_subtype == 2)
    {
      /* we have to generate a call sequence because the address is out of range for a bsr */

      char reg = fragP->fr_opcode[0]; /* remember we are little endian */
      int where = fragP->fr_fix - 2;

      /* pfx %hi(symbol_h) */

      gas_cgen_record_fixup (fragP,
			     where,
			     pfx_insn,
			     2 /*length*/,
			     cgen_operand_lookup_by_num (gas_cgen_cpu_desc, NIOS_OPERAND_I11),
			     BFD_RELOC_NIOS_H_HI11,
			     fragP->fr_symbol, fragP->fr_offset);


      md_number_to_chars (opcode, 0x9800, 2);

      /* movi %rx,%lo(symbol_h)  */

      gas_cgen_record_fixup (fragP,
			     where + 2,
			     movi_insn,
			     2 /*length*/,
			     cgen_operand_lookup_by_num (gas_cgen_cpu_desc, NIOS_OPERAND_I5),
			     BFD_RELOC_NIOS_H_LO5,
			     fragP->fr_symbol, fragP->fr_offset);


      md_number_to_chars (opcode + 2, 0x3400 | reg, 2);
      opcode += 4;

      /* need additional instructions if we are m32 */
      if (nios_mach == bfd_mach_nios32)
	{
	  /* pfx %xhi(symbol_h) */
	  gas_cgen_record_fixup (fragP,
				 where + 4,
				 pfx_insn,
				 2 /*length*/,
				 cgen_operand_lookup_by_num (gas_cgen_cpu_desc, NIOS_OPERAND_I11),
				 BFD_RELOC_NIOS_H_XHI11,
				 fragP->fr_symbol, fragP->fr_offset);
	  
	  md_number_to_chars (opcode, 0x9800, 2);
	  opcode += 2;
	  
	  /* movhi %rx,%xlo(symbol_h) */
	  gas_cgen_record_fixup (fragP,
				 where + 6,
				 movhi_insn,
				 2 /*length*/,
				 cgen_operand_lookup_by_num (gas_cgen_cpu_desc, NIOS_OPERAND_I5),
				 BFD_RELOC_NIOS_H_XLO5,
				 fragP->fr_symbol, fragP->fr_offset);

	  md_number_to_chars (opcode, 0x6C00 | reg, 2);
	  opcode += 2;
	}

      /* call %rx */
      md_number_to_chars (opcode, 0x7fe0 | reg, 2);

      fragP->fr_fix += opcode - fragP->fr_opcode;
    }
}


/* Functions concerning relocs.  */

/* The location from which a PC relative jump should be calculated,
   given a PC relative reloc.  */

long
md_pcrel_from_section (fixP, sec)
     fixS * fixP;
     segT   sec;
{
  long pc_addr;

  if (fixP->fx_addsy != (symbolS *) NULL
      && (! S_IS_DEFINED (fixP->fx_addsy)
	  || S_GET_SEGMENT (fixP->fx_addsy) != sec))
    {
      /* The symbol is undefined (or is defined but not in this section).
	 Let the linker figure it out.  */
      return 0;
    }

  pc_addr = fixP->fx_frag->fr_address + fixP->fx_where;

  /* for ldc on nios-32, we need to use the pc rounded
     down to a word-aligned address before the relative calculation */
  if (nios_mach == bfd_mach_nios32)
    {
      char *mnemonic = (char *)fixP->fx_cgen.insn->base->mnemonic;
      switch (*mnemonic)
	{
	case 'l':
	  return pc_addr & ~3;
	  break;
	}
    }
      
  return pc_addr & ~1;
}

/* Return the bfd reloc type for OPERAND of INSN at fixup FIXP.
   Returns BFD_RELOC_NONE if no reloc type can be found.
   *FIXP may be modified if desired.  */

bfd_reloc_code_real_type
md_cgen_lookup_reloc (insn, operand, fixP)
     const CGEN_INSN *    insn;
     const CGEN_OPERAND * operand;
     fixS *               fixP;
{
  switch (operand->type)
    {
    case NIOS_OPERAND_I5:
    case NIOS_OPERAND_I11:
    case NIOS_OPERAND_I16:
    case NIOS_OPERAND_I32:
      /* The relocation type should be recorded in opinfo */
      if (fixP->fx_cgen.opinfo != 0)
        return fixP->fx_cgen.opinfo;

    case NIOS_OPERAND_M16_I6:
    case NIOS_OPERAND_M32_I6:
      fixP->fx_pcrel = true;
      return BFD_RELOC_NIOS_PCREL6;

    case NIOS_OPERAND_M16_I8V:
    case NIOS_OPERAND_M32_I8V:
      fixP->fx_pcrel = true;
      return BFD_RELOC_NIOS_PCREL8;

    case NIOS_OPERAND_REL11:
      fixP->fx_pcrel = true;
      return BFD_RELOC_NIOS_PCREL11;
    }

  return BFD_RELOC_NONE;
}

static void
nios_add_macro (name, semantics, arguments)
     char  *name;
     char  *semantics;
     char **arguments;
{
  macro_entry *macro;
  sb macro_name;
  const char *namestr;

  macro = (macro_entry *) xmalloc (sizeof (macro_entry));
  sb_new (&macro->sub);
  sb_new (&macro_name);

  macro->formal_count = 0;
  macro->formals = 0;

  sb_add_string (&macro->sub, semantics);

  if (arguments != NULL)
    {
      formal_entry **p = &macro->formals;
      
      macro->formal_count = 0;
      macro->formal_hash = hash_new ();
      while (*arguments != NULL)
	{
	  formal_entry *formal;
	  
	  formal = (formal_entry *) xmalloc (sizeof (formal_entry));
	  
	  sb_new (&formal->name);
	  sb_new (&formal->def);
	  sb_new (&formal->actual);
	  
	  sb_add_string (&formal->name, *arguments);
	  /* Add to macro's hash table */
	  hash_jam (macro->formal_hash, sb_terminate (&formal->name), formal);
	  
	  formal->index = macro->formal_count;
	  macro->formal_count++;
	  *p = formal;
	  p = &formal->next;
	  *p = NULL;
	  ++arguments;
	}
    }

  sb_add_string (&macro_name, name);  
  namestr = sb_terminate (&macro_name);
  hash_jam (get_macro_hash(), namestr, (PTR) macro);

  macro_defined = 1;
}

/* automatically enter conditional branch macros */
static void
nios_load_macros ()
{
  char *beq_sem = "\n skps cc_ne\n br \\label\n";
  char *bne_sem = "\n skps cc_eq\n br \\label\n";
  char *bcc_sem = "\n skps cc_c \n br \\label\n";
  char *bcs_sem = "\n skps cc_nc\n br \\label\n";
  char *bvc_sem = "\n skps cc_v \n br \\label\n";
  char *bvs_sem = "\n skps cc_nv\n br \\label\n";
  char *bge_sem = "\n skps cc_lt\n br \\label\n";
  char *blt_sem = "\n skps cc_ge\n br \\label\n";
  char *ble_sem = "\n skps cc_gt\n br \\label\n";
  char *bgt_sem = "\n skps cc_le\n br \\label\n";
  char *bls_sem = "\n skps cc_hi\n br \\label\n";
  char *bhi_sem = "\n skps cc_ls\n br \\label\n";
  char *bmi_sem = "\n skps cc_pl\n br \\label\n";
  char *bpl_sem = "\n skps cc_mi\n br \\label\n";

  char *br_cc_arg_0 = "label";
  char *br_cc_args[2] = { br_cc_arg_0, NULL };

  nios_add_macro ("beq", beq_sem, br_cc_args);
  nios_add_macro ("bne", bne_sem, br_cc_args);
  nios_add_macro ("bz",  beq_sem, br_cc_args);
  nios_add_macro ("bnz", bne_sem, br_cc_args);
  nios_add_macro ("bcc", bcc_sem, br_cc_args);
  nios_add_macro ("bcs", bcs_sem, br_cc_args);
  nios_add_macro ("bnc", bcc_sem, br_cc_args);
  nios_add_macro ("bc",  bcs_sem, br_cc_args);
  nios_add_macro ("bge", bge_sem, br_cc_args);
  nios_add_macro ("blt", blt_sem, br_cc_args);
  nios_add_macro ("ble", ble_sem, br_cc_args);
  nios_add_macro ("bgt", bgt_sem, br_cc_args);
  nios_add_macro ("bls", bls_sem, br_cc_args);
  nios_add_macro ("bhi", bhi_sem, br_cc_args);
  nios_add_macro ("bv",  bvs_sem, br_cc_args);
  nios_add_macro ("bnv", bvc_sem, br_cc_args);
  nios_add_macro ("bvs", bvs_sem, br_cc_args);
  nios_add_macro ("bvc", bvc_sem, br_cc_args);
  nios_add_macro ("bpl", bpl_sem, br_cc_args);
  nios_add_macro ("bmi", bmi_sem, br_cc_args);
  nios_add_macro ("bp",  bpl_sem, br_cc_args);
  nios_add_macro ("bn",  bmi_sem, br_cc_args);
}

/* See whether we need to force a relocation into the output file. */

int
nios_force_relocation (fix)
     fixS * fix;
{
  if (fix->fx_r_type == BFD_RELOC_VTABLE_INHERIT
      || fix->fx_r_type == BFD_RELOC_VTABLE_ENTRY)
    return 1;

  return 0;
}

/* Write a value out to the object file, using the appropriate endianness.  */

void
md_number_to_chars (buf, val, n)
     char * buf;
     valueT val;
     int    n;
{
  number_to_chars_littleendian (buf, val, n);
}

/* Turn a string in input_line_pointer into a floating point constant of type
   type, and store the appropriate bytes in *litP.  The number of LITTLENUMS
   emitted is stored in *sizeP .  An error message is returned, or NULL on OK.
*/

/* Equal to MAX_PRECISION in atof-ieee.c */
#define MAX_LITTLENUMS 6

char *
md_atof (type, litP, sizeP)
     char   type;
     char * litP;
     int *  sizeP;
{
  int              i;
  int              prec;
  LITTLENUM_TYPE   words [MAX_LITTLENUMS];
  LITTLENUM_TYPE  *wordP;
  char *           t;
  char *           atof_ieee ();

  switch (type)
    {
    case 'f':
    case 'F':
    case 's':
    case 'S':
      prec = 2;
      break;

    case 'd':
    case 'D':
    case 'r':
    case 'R':
      prec = 4;
      break;

   /* FIXME: Some targets allow other format chars for bigger sizes here.  */

    default:
      * sizeP = 0;
      return _("Bad call to md_atof()");
    }

  t = atof_ieee (input_line_pointer, type, words);
  if (t)
    input_line_pointer = t;
  * sizeP = prec * sizeof (LITTLENUM_TYPE);

  /* This loops outputs the LITTLENUMs in REVERSE order; in accord with
     the nios endianness.  */
  for (wordP = words + prec - 1; prec--;)
    {
      md_number_to_chars (litP, (valueT) (*wordP--), sizeof (LITTLENUM_TYPE));
      litP += sizeof (LITTLENUM_TYPE);
    }
     
  return 0;
}

/* This is where we do the unexpected alignment check.
   This is called from HANDLE_ALIGN in tc-nios.h.  */

void
nios_align_code (fragp, count)
     fragS *fragp;
     int count;
{
  if (count >= 2
      && !(count & 1)
      && count <= 1024
      && !((long)(fragp->fr_literal + fragp->fr_fix) & 1))
    {
      unsigned short *p = (unsigned short*)(fragp->fr_literal + fragp->fr_fix);
      int i;
      
      for (i = 0; i < count; i += 2, p++)
	md_number_to_chars ((char *)p, 0x3000, 2);
      
      fragp->fr_var = count;
    }
}

boolean
nios_fix_adjustable (fixP)
   fixS * fixP;
{
  bfd_reloc_code_real_type reloc_type;

  if ((int) fixP->fx_r_type >= (int) BFD_RELOC_UNUSED)
    {
      const CGEN_INSN *insn = NULL;
      int opindex = (int) fixP->fx_r_type - (int) BFD_RELOC_UNUSED;
      const CGEN_OPERAND *operand = cgen_operand_lookup_by_num(gas_cgen_cpu_desc, opindex);
      reloc_type = md_cgen_lookup_reloc (insn, operand, fixP);
    }
  else
    reloc_type = fixP->fx_r_type;

  if (fixP->fx_addsy == NULL)
    return 1;
  
  /* Prevent all adjustments to global symbols. */
  if (S_IS_EXTERN (fixP->fx_addsy))
    return 0;
  
  if (S_IS_WEAK (fixP->fx_addsy))
    return 0;
  
  /* We need the symbol name for the VTABLE entries */
  if (   reloc_type == BFD_RELOC_VTABLE_INHERIT
      || reloc_type == BFD_RELOC_VTABLE_ENTRY)
    return 0;

  return 1;
}

/* Pseudo-op helper functions */

/* Allow user to set flags bits.  */
void
nios_set_flags (arg)
     int arg ATTRIBUTE_UNUSED;
{
  flagword new_flags = get_absolute_expression ();
  flagword new_mask = ~ (flagword)0;

  nios_user_set_flags_p = 1;
  if (*input_line_pointer == ',')
    {
      ++input_line_pointer;
      new_mask = get_absolute_expression ();
    }

  nios_flags = (nios_flags & ~new_mask) | (new_flags & new_mask);
  bfd_set_private_flags (stdoutput, nios_flags);
}

static int nios_unaligned_constant = 0;

/* This handles the native word allocation pseudo-op .nword.
   For nios16 it is equivalent to .half,  for
   nios32 it is equivalent to .word.  */

static void
nios_nword (bytes)
     int bytes;
{
  nios_word (nios_mach == bfd_mach_nios32 ? 4 : 2);
}

/* This handles the native unaligned word allocation pseudo-op .uanword.
   For nios16 it is equivalent to .uahalf,  for
   nios32 it is equivalent to .uaword.  */

static void
nios_uanword (bytes)
     int bytes;
{
  nios_unaligned_constant = 1;
  nios_word (nios_mach == bfd_mach_nios32 ? 4 : 2);
}

/* Return log2 (VALUE), or -1 if VALUE is not an exact positive power
   of two.  */
static int
log2 (value)
     int value;
{
  int shift;

  if (value <= 0)
    return -1;

  for (shift = 0; (value & 1) == 0; value >>= 1)
    ++shift;

  return (value == 1) ? shift : -1;
}

/* unaligned constant */
static void
nios_uaword (nbytes)
{
  nios_unaligned_constant = 1;
  nios_word (nbytes);
}

/* Like normal .word, except support @h */
/* clobbers input_line_pointer, checks end-of-line. */
static void
nios_word (nbytes)
     register int nbytes;
{
  expressionS exp;
  bfd_reloc_code_real_type reloc;
  char *p;
  int offset;
  int nalign;
  int unaligned;

  unaligned = nios_unaligned_constant;
  nios_unaligned_constant = 0;

  if (is_it_end_of_statement ())
    {
      demand_empty_rest_of_line ();
      return;
    }

  if (!unaligned)
    {
      nalign = log2 (nbytes);
      if (nalign == 0)
	return;
      
      assert (nalign > 0);
      
      if (now_seg == absolute_section)
	{
	  /* don't want to align, but want to continue and process expression below */
	  if ((abs_section_offset & ((1 << nalign) - 1)) != 0)
	    {
	      as_bad (_("misaligned data"));
	      return;
	    }
	}
      else
	{
	  p = frag_var (rs_align_code, 1, 1, (relax_substateT) 0,
			(symbolS *) NULL, (offsetT) nalign, (char *) NULL);
	  
	  record_alignment (now_seg, nalign);
	}
    }

  do
    {
      expression (&exp);
      if (!strncasecmp (input_line_pointer, "@h", 2))
	{
	  input_line_pointer += 2;
	  exp.X_add_number = get_single_number ();
	
	  p = frag_more (nbytes);
	  if (nbytes == 2)
	    fix_new_exp (frag_now, p - frag_now->fr_literal, nbytes, 
			 &exp, 0, BFD_RELOC_NIOS_H_16);
	  else
	    fix_new_exp (frag_now, p - frag_now->fr_literal, nbytes, 
			 &exp, 0, BFD_RELOC_NIOS_H_32);
	}
      else
	emit_expr (&exp, nbytes);
    }
  while (*input_line_pointer++ == ',');

  input_line_pointer--;		/* Put terminator back into stream. */
  demand_empty_rest_of_line ();
}

#ifdef DEBUG
#define DPRINTF1(A)	fprintf (stderr, A)
#define DPRINTF2(A,B)	fprintf (stderr, A, B)
#define DPRINTF3(A,B,C)	fprintf (stderr, A, B, C)
#else
#define DPRINTF1(A)
#define DPRINTF2(A,B)
#define DPRINTF3(A,B,C)
#endif


/* Apply a fixup to the object code.  We attempt to resolve the NIOS
   specific relocs %hi, %lo, %xhi and %xlo, both with and without @h.
   These fixups mirror the relocation code in bfd/elf32-nios.c.  For
   other types of fixup, we fallback to gas_cgen_md_apply_fix3().  */

int
nios_apply_fix3 (fixP, valueP, seg)
     fixS *fixP;
     valueT *valueP;
     segT seg;
{
  valueT value = *valueP;

  switch (fixP->fx_cgen.opinfo)
    {
    case BFD_RELOC_NIOS_LO16_LO5:
      value = (value & 0x001f);
      break;

    case BFD_RELOC_NIOS_H_LO5:
      value = ((value >> 1) & 0x001f);
      break;

    case BFD_RELOC_NIOS_LO16_HI11:
      value = ((value & 0xffff) >> 5);
      break;

    case BFD_RELOC_NIOS_H_HI11:
      value = ((value & 0x1fffe) >> 6);
      break;

    case BFD_RELOC_NIOS_HI16_LO5:
      value = ((value >> 16) & 0x001f);
      value <<= 5;
      break;

    case BFD_RELOC_NIOS_H_XLO5:
      value = ((value >> 17) & 0x001f);
      value <<= 5;
      break;
  
    case BFD_RELOC_NIOS_HI16_HI11:
      value = (value >> 21);
      break;

    case BFD_RELOC_NIOS_H_XHI11:
      value = (value >> 22);
      break;

    default:
      break;
    }

  *valueP = value;

  return gas_cgen_md_apply_fix3 (fixP, valueP, seg);
}
