/* {{{ Target definitions for the Altera Nios chip.

   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Contributed by Cygnus Solutions, a Red Hat company

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <ctype.h>

#include "config.h"
#include "system.h"
#include "tree.h"
#include "rtl.h"
#include "regs.h"
#include "obstack.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "insn-attr.h"
#include "flags.h"
#include "function.h"
#include "expr.h"
#include "recog.h"
#include "toplev.h"
#include "ggc.h"
#include "tm_p.h"
/* }}} */


// -----------------
// dvb/cv 2001 Oct 23 - adding a conditional flag to inhibit
// the generation of "jmpc" instructions. We
// may be missing this instruction in some versions
// of the CPU, and we're definitely lacking it
// now, so we'll have this option. But it's easy
// to put things back as they were: set the flag
// to zero.

#define DONT_EMIT_JMPC 1	// 0 for normal code generation

// We replace JMPC instructions with a rather straightforward
// alternative:
//
//	pfx	%hi(dest@h)
//	movi	%g0,%lo(dest@h)
//	pfx	%xhi(dest@h)	; if nios32
//	movhi	%g0%xlo(dest@h)	; if nios32
//	jmp	%g0
//	[nop]
//
// Of the three spots where JMPC is generated in this
// file, only one of them attempts to schedule a
// branch delay slot afterwards. The other two
// include an explicit NOP after the JMPC/alternative.
// (revised dvb 2001 Dec 6)
//
// -----------------




/* {{{ Local data structures, variables and macros.  */

/* Define variables for command options that have values. */
NIOS_USER_OPCODE nios_user_opcodes[NIOS_NUM_USER_OPCODES] = {
    {"mul"},
    {"div"},
    {"udiv"},
    {"mod"},
    {"umod"},
    {"fadd"},
    {"fsub"},
    {"fmul"},
    {"smin"},
    {"umin"},
    {"smax"},
    {"umax"},
    {"ffs"},
    {"extv"},
    {"extzv"},
    {"insv"},
};

char *nios_max_address = NULL;
int nios_suppress_address_xlo = 0;
int nios_suppress_address_xhi = 0;

/* Define the information needed to generate branch and scc insns.  This is
   stored from the compare operation.  */
struct rtx_def *nios_compare_op0;
struct rtx_def *nios_compare_op1;

/* Map register number to smallest register class.  */
enum reg_class regno_reg_class[FIRST_PSEUDO_REGISTER];

/* Map class letter into register class */
enum reg_class reg_class_from_letter[256];

/* Branch cost. */
int nios_branch_cost_int = DEFAULT_BRANCH_COST;

/* Name of where we pretend to think the frame pointer points.
   Normally, this is "%fp", but if we are in a leaf procedure,
   this is "%sp+something".  We record "something" separately as it may be
   too big for reg+constant addressing.  */
static const char *frame_base_name;
static int frame_base_offset;

/* Size of frame.  Need to know this to emit return insns from leaf
   procedures. ACTUAL_FSIZE is set by nios_compute_frame_size() which is
   called during the reload pass.  This is important as the value is later
   used in insn scheduling (to see what can go in a delay slot).
   APPARENT_FSIZE is the size of the stack less the register save area and
   less the outgoing argument area.  It is used when saving call preserved
   regs.  */
static int apparent_fsize;
static int actual_fsize;

#ifdef LEAF_REGISTERS
/* Vector to say how input registers are mapped to output
   registers.  FRAME_POINTER_REGNUM cannot be remapped by
   this function to eliminate it.  You must use -fomit-frame-pointer
   to get that.  */
char leaf_reg_remap[] =
{  0,  1,  2,  3,  4,  5,  6,  7,
  -1, -1, -1, -1, -1, -1, 14, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
   8,  9, 10, 11, 12, 13, -1, 15 };
#endif	/* LEAF_REGISTERS */

/* Opcode names used in nios_output_2op_insn.  */
char *nios_2op_opcode_names[][2] = NIOS_2OP_OPCODE_NAMES;

/* We may need an epilogue if we spill too many registers.
   If this is non-zero, then we branch here for the epilogue.  */
static rtx leaf_label;

extern struct obstack * saveable_obstack;

#define ROUND_ADVANCE(SIZE) (((SIZE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD)
/* }}} */

/* {{{ Local function declarations.  */
/* Various local help functions.  */
static void nios_split_constant32		PROTO ((rtx, int []));
static void nios_split_constant64		PROTO ((rtx, int []));
static void nios_print_addr_const		PROTO ((FILE *, rtx));
static void nios_print_operand_memory_reference PROTO ((FILE *, rtx));
static int nios_function_arg_slotno		PROTO ((const CUMULATIVE_ARGS *,
							enum machine_mode mode,
							tree, int, int, int *));
static HOST_WIDE_INT nios_get_value		PROTO ((rtx));
static int nios_reg_ok_for_simple_p		PROTO ((rtx, int));
static void nios_emit_stack_reservation		PROTO ((FILE *, int));
/* }}} */


/* JMR - A helper for laying assembly language comments.
 * Mostly for maintainer debugging.
 *
 *    nios_comment ("flarg is %x", blorg & 0xff);
 */
static void
nios_comment (FILE * file, const char *format, ...)
{
  va_list v;

  va_start (v, format);
  fprintf (file, "\t%s ", ASM_COMMENT_START);
  vfprintf (file, format, v);
  fprintf (file, "\n");
  va_end (v);
}


/* CBV - Add call to nios_mcount to top of every function call */
/* when compiling for GPROF */
void
nios_function_profiler (file, labelno)
     FILE *file;
     int labelno;
{
  char buf[32];
  ASM_GENERATE_INTERNAL_LABEL (buf, "LP", labelno);

  // -------------------
  // Generate a call to mcount. Note that we can
  // safely go ahead and use %o0, since we're the first thing
  // that happens in this routine and %o0 isn't
  // otherwise in use yet. (WE HOPE! !!!) -- dvb/cv 2001 10 23

  fputs ("\tpfx\t%hi(nios_mcount@h)\n", file); 
  fputs ("\tmovi\t%o0,%lo(nios_mcount@h)\n", file);

  // for nios 32 only, as set by the -m32 cmd line switch, we load the upper 16 bits of the address
  if (TARGET_32BIT)
    {
    fputs ("\tpfx\t%xhi(nios_mcount@h)\n", file);
    fputs ("\tmovhi\t%o0,%xlo(nios_mcount@h)\n", file);
    }
  //

  fputs ("\tcall\t%o0\n", file);
  fputs ("\tnop ; (branch delay slot)\n", file);  /* NOP in branch delay slot   */

}

/* {{{ Backend initialization.  */

/* {{{ nios_override_options()
   Sometimes certain combinations of command options do not make
   sense on a particular target machine.  You can define a macro
   `OVERRIDE_OPTIONS' to take account of this.  This macro, if defined, is
   executed once just after all the command options have been parsed.

   Don't use this macro to turn on various extra optimizations for `-O'.
   That is what `OPTIMIZATION_OPTIONS' is for.  */

void
nios_override_options ()
{
  int regno, i, chars_read;
  unsigned long max_address;
  int num_option_errors = 0;

  /* A C expression whose value is a register class containing hard
     register REGNO.  In general there is more than one such class; choose
     a class which is "minimal", meaning that no smaller class also
     contains the register. */
  for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
    {
      enum reg_class class = DATA_REGS;

      switch (regno)
	{
	case GPR_BP0:		/* Base pointer registers for STP/LDP */
	case GPR_BP1:
	case GPR_BP2:
	case GPR_BP3:
	case GPR_SP:		/* Stack pointer */
	case GPR_FP:		/* Frame pointer */
	  class = BP_REGS;
	  break;

	case CT10_STATUS:	/* Control registers */
	case CT11_ISTATUS:
	case CT12_WVALID:
	case CT13_CD_BANK:
	case CT14_ST_BANK:
	case CT15_LD_BANK:
	  class = CT_REGS;
	  break;

	case GPR_RETURN:	/* Link registers */
	case GPR_LINK:
	  class = LN_REGS;
	  break;
	}

      regno_reg_class[regno] = class;
    }

  /* A C expression which defines the machine-dependent operand
     constraint letters for register classes.  If CHAR is such a
     letter, the value should be the register class corresponding to
     it.  Otherwise, the value should be `NO_REGS'.  The register
     letter `r', corresponding to class `GENERAL_REGS', will not be
     passed to this macro; you do not need to handle it.

     The following letters are unavailable, due to being used as
     constraints:
	'0'..'9'
	'<', '>'
	'E', 'F', 'G', 'H'
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P'
	'Q', 'R', 'S', 'T', 'U'
	'V', 'X'
	'g', 'i', 'm', 'n', 'o', 'p', 'r', 's' */

  for (i = 0; i < 256; i++)
    reg_class_from_letter[i] = NO_REGS;

  reg_class_from_letter['b'] = BP_REGS;
  reg_class_from_letter['d'] = DATA_REGS;
  reg_class_from_letter['c'] = CT_REGS;
  reg_class_from_letter['l'] = LN_REGS;

  /* Parse -muser-opcode-???= options */
  for (i = 0; i < NIOS_NUM_USER_OPCODES; ++i)
    if (nios_user_opcodes[i].s != NULL)
      {
	nios_user_opcodes[i].pfx = 0;

	if (sscanf(nios_user_opcodes[i].s, "usr%i%n",
		   &nios_user_opcodes[i].usr,
		   &chars_read) == 1
	      && chars_read == strlen(nios_user_opcodes[i].s)
	    || sscanf(nios_user_opcodes[i].s, "pfx%i,usr%i%n",
			&nios_user_opcodes[i].pfx,
			&nios_user_opcodes[i].usr,
			&chars_read) == 2
	      && chars_read == strlen(nios_user_opcodes[i].s))
	  {
	    if (nios_user_opcodes[i].pfx >  2047 ||
		nios_user_opcodes[i].pfx < -1024)
	      {
		error ("Bad option: -muser-opcode-%s=%s, "
			      "valid range is pfx-1024 to pfx2047",
		       nios_user_opcodes[i].name,
		       nios_user_opcodes[i].s);

		++num_option_errors;
	      }

	    if (nios_user_opcodes[i].usr > 4 ||
		nios_user_opcodes[i].usr < 0)
	      {
		error ("Bad option: -muser-opcode-%s=%s, "
				"valid range is usr0-usr4",
		       nios_user_opcodes[i].name,
		       nios_user_opcodes[i].s);

		++num_option_errors;
	      }
	  }

	else
	  {
	    error ("Bad option: -muser-opcode-%s=%s, "
				"use usr# or pfx#,usr#",
		   nios_user_opcodes[i].name,
		   nios_user_opcodes[i].s);

	    ++num_option_errors;
	  }
      }

  /* Parse -mmax-address= options */
  if (nios_max_address != NULL)
    if (sscanf(nios_max_address, "%lx%n", &max_address, &chars_read) == 1
	&& chars_read == strlen(nios_max_address))
      {
	if ((max_address & 0xffff0000) == 0) nios_suppress_address_xlo = 1;
	if ((max_address & 0xffe00000) == 0) nios_suppress_address_xhi = 1;
      }
    else
      {
	error ("Bad value \"%s\" for option -mmax-address", nios_max_address);
	++num_option_errors;
      }

  if (num_option_errors > 0)
    fatal("Bad -m options");
}

/* }}} */

/* {{{ nios_optimization_options()
   Some machines may desire to change what optimizations are performed for
   various optimization levels.  This macro, if defined, is executed once just
   after the optimization level is determined and before the remainder of the
   command options have been parsed.  Values set in this macro are used as the
   default values for the other command line options.

   LEVEL is the optimization level specified; 2 if `-O2' is specified, 1 if
   `-O' is specified, and 0 if neither is specified.

   SIZE is non-zero if `-Os' is specified, 0 otherwise.  

   You should not use this macro to change options that are not
   machine-specific.  These should uniformly selected by the same optimization
   level on all supported machines.  Use this macro to enable machbine-specific
   optimizations.

   *Do not examine `write_symbols' in this macro!* The debugging options are
   *not supposed to alter the generated code.  */

void
nios_optimization_options (level, size)
  int level;
  int size ATTRIBUTE_UNUSED;
{
  if (level >= 2)
    {
      target_flags |= MASK_ANDN;
      target_flags |= MASK_NEG_CONST;
      target_flags |= MASK_NEW_MULHISI;
    }
}

/* }}} */

/* }}} */

/* {{{ stdarg/varargs support.  */

/* {{{ nios_builtin_saveregs()
   Do what is necessary for `va_start'.  We look at the current function
   to determine if stdarg or varargs is used and return the address of the
   first unnamed parameter.  */

rtx
nios_builtin_saveregs ()
{
  int first_reg = current_function_args_info.words;
  rtx address;
  int regno;

  for (regno = first_reg; regno < NIOS_NUM_ARG_REGS; regno++)
    emit_move_insn (gen_rtx_MEM (word_mode,
			     gen_rtx_PLUS (Pmode,
				      frame_pointer_rtx,
				      GEN_INT (STACK_POINTER_OFFSET
					       + UNITS_PER_WORD * regno))),
		    gen_rtx_REG (word_mode,
			     NIOS_INCOMING_INT_ARG_FIRST + regno));

  address = gen_rtx_PLUS (Pmode,
		     frame_pointer_rtx,
		     GEN_INT (STACK_POINTER_OFFSET
			      + UNITS_PER_WORD * first_reg));

  if (current_function_check_memory_usage
      && first_reg < NIOS_NUM_ARG_REGS)
    emit_library_call (chkr_set_right_libfunc, 1, VOIDmode, 3,
		       address, ptr_mode,
		       GEN_INT (UNITS_PER_WORD 
			 	* (NIOS_NUM_ARG_REGS - first_reg)),
		       TYPE_MODE (sizetype), GEN_INT (MEMORY_USE_RW),
		       TYPE_MODE (integer_type_node));

  return address;
}
/* }}} */

/* {{{ nios_va_start()
   Implement `va_start' for varargs and stdarg.  */

void
nios_va_start (stdarg_p, valist, nextarg)
     int stdarg_p ATTRIBUTE_UNUSED;
     tree valist;
     rtx nextarg;
{
  nextarg = expand_builtin_saveregs ();
  std_expand_builtin_va_start (1, valist, nextarg);
}
/* }}} */

/* {{{ nios_va_arg()
   Implement `va_arg'.  */

rtx
nios_va_arg (valist, type)
     tree valist, type;
{
  HOST_WIDE_INT size, rsize;
  tree addr, incr;
  rtx addr_rtx;
  int indirect = 0;

  /* Round up sizeof(type) to a word.  */
  size = int_size_in_bytes (type);
  rsize = (size + UNITS_PER_WORD - 1) & -UNITS_PER_WORD;

  /* Large types are passed by reference (this is the same decision taken
     by nios_function_arg_pass_by_reference).  */
  if (size > 4 * UNITS_PER_WORD)
    {
      indirect = 1;
      size = rsize = UNITS_PER_WORD;
    }

  incr = valist;
  addr = incr = save_expr (incr);
  incr = fold (build (PLUS_EXPR, ptr_type_node, incr,
		      build_int_2 (rsize, 0)));

  incr = build (MODIFY_EXPR, ptr_type_node, valist, incr);
  TREE_SIDE_EFFECTS (incr) = 1;
  expand_expr (incr, const0_rtx, VOIDmode, EXPAND_NORMAL);

  addr_rtx = expand_expr (addr, NULL, Pmode, EXPAND_NORMAL);

  if (indirect)
    {
      addr_rtx = force_reg (Pmode, addr_rtx);
      addr_rtx = gen_rtx_MEM (Pmode, addr_rtx);
      MEM_ALIAS_SET (addr_rtx) = get_varargs_alias_set ();
    }

  return addr_rtx;
}
/* }}} */

/* }}} */

/* {{{ Trampoline support.  */

/* {{{ nios_trampoline_size()
   
   FIXME - Trampoline templates have not been defined yet.

   A C expression for the size in bytes of the trampoline, as an integer.  */

int
nios_trampoline_size ()
{
  return 5 /* instructions */ * 4 /* instruction size */;
}
/* }}} */

/* {{{ nios_initialize_trampoline()
   
   FIXME - Trampoline templates have not been defined yet.
   
   A C statement to initialize the variable parts of a trampoline.  ADDR is
   an RTX for the address of the trampoline; FNADDR is an RTX for the
   address of the nested function; STATIC_CHAIN is an RTX for the static
   chain value that should be passed to the function when it is called.  */

void
nios_initialize_trampoline (addr, fnaddr, static_chain)
     rtx addr;
     rtx fnaddr;
     rtx static_chain;
{
  rtx sc_reg = force_reg (Pmode, static_chain);

  emit_library_call (gen_rtx_SYMBOL_REF (SImode, "__trampoline_setup"),
		     FALSE, VOIDmode, 4,
		     addr, Pmode,
		     GEN_INT (nios_trampoline_size ()), SImode,
		     fnaddr, Pmode,
		     sc_reg, Pmode);
}
/* }}} */

/* }}} */

/* {{{ Memory and register addressing support.  */

/* {{{ nios_hard_regno_mode_ok()
   
   A C expression that is nonzero if it is permissible to store a value
   of mode MODE in hard register number REGNO (or in several registers
   starting with that one).  For a machine where all registers are
   equivalent, a suitable definition is

        #define HARD_REGNO_MODE_OK(REGNO, MODE) 1

   It is not necessary for this macro to check for the numbers of fixed
   registers, because the allocation mechanism considers them to be
   always occupied.

   On some machines, double-precision values must be kept in even/odd
   register pairs.  The way to implement that is to define this macro to
   reject odd register numbers for such modes.

   The minimum requirement for a mode to be OK in a register is that the
   `movMODE' instruction pattern support moves between the register and
   any other hard register for which the mode is OK; and that moving a
   value into the register and back out not alter it.

   Since the same instruction used to move `SImode' will work for all
   narrower integer modes, it is not necessary on any machine for
   `HARD_REGNO_MODE_OK' to distinguish between these modes, provided you
   define patterns `movhi', etc., to take advantage of this.  This is
   useful because of the interaction between `HARD_REGNO_MODE_OK' and
   `MODES_TIEABLE_P'; it is very desirable for all integer modes to be
   tieable.

   Many machines have special registers for floating point arithmetic.
   Often people assume that floating point machine modes are allowed
   only in floating point registers.  This is not true.  Any registers
   that can hold integers can safely *hold* a floating point machine
   mode, whether or not floating arithmetic can be done on it in those
   registers. Integer move instructions can be used to move the values.

   On some machines, though, the converse is true: fixed-point machine
   modes may not go in floating registers.  This is true if the floating
   registers normalize any value stored in them, because storing a
   non-floating value there would garble it.  In this case,
   `HARD_REGNO_MODE_OK' should reject fixed-point machine modes in
   floating registers.  But if the floating registers do not
   automatically normalize, if you can store any bit pattern in one and
   retrieve it unchanged without a trap, then any machine mode may go in
   a floating register, so you can define this macro to say so.

   The primary significance of special floating registers is rather that
   they are the registers acceptable in floating point arithmetic
   instructions. However, this is of no concern to `HARD_REGNO_MODE_OK'.
   You handle it by writing the proper constraints for those
   instructions.

   On some machines, the floating registers are especially slow to
   access, so that it is better to store a value in a stack frame than
   in such a register if floating point arithmetic is not being done.
   As long as the floating registers are not in class `GENERAL_REGS',
   they will not be used unless some pattern's constraint asks for one.  */

int
nios_hard_regno_mode_ok (regno, mode)
     int regno;
     enum machine_mode mode;
{
  return (((regno) & 1) == 0 || GET_MODE_SIZE (mode) <= UNITS_PER_WORD);
}
/* }}} */

/* {{{ nios_mode_tieable_p()
   
   A C expression that is nonzero if it is desirable to choose register
   allocation so as to avoid move instructions between a value of mode
   MODE1 and a value of mode MODE2.

   If `HARD_REGNO_MODE_OK (R, MODE1)' and `HARD_REGNO_MODE_OK (R, MODE2)'
   are ever different for any R, then `MODES_TIEABLE_P (MODE1, MODE2)' must
   be zero.  */

int
nios_modes_tieable_p (mode1, mode2)
    enum machine_mode mode1;
    enum machine_mode mode2;
{
  return (mode1 == mode2
	  || (GET_MODE_SIZE (mode1) <= UNITS_PER_WORD
	      && GET_MODE_SIZE (mode2) <= UNITS_PER_WORD));
}
/* }}} */

/* {{{ nios_legitimate_constant_p()
   
   A C expression that is nonzero if X is a legitimate constant for an
   immediate operand on the target machine.  You can assume that X
   satisfies `CONSTANT_P', so you need not check this.  In fact, `1' is
   a suitable definition for this macro on machines where anything
   `CONSTANT_P' is valid.  */

int
nios_legitimate_constant_p (x)
     rtx x ATTRIBUTE_UNUSED;
{
  return 1;
}
/* }}} */

/* {{{ nios_return_addr_rtx()
   
   A C expression whose value is RTL representing the value of the
   return address for the frame COUNT steps up from the current frame,
   after the prologue.  FRAMEADDR is the frame pointer of the COUNT
   frame, or the frame pointer of the COUNT - 1 frame if
   `RETURN_ADDR_IN_PREVIOUS_FRAME' is defined.

   The value of the expression must always be the correct address when
   COUNT is zero, but may be `NULL_RTX' if there is not way to determine
   the return address of other frames.  */

rtx
nios_return_addr_rtx (count, frame)
     int count ATTRIBUTE_UNUSED;
     rtx frame ATTRIBUTE_UNUSED;
{
  if (count == 0)
    return gen_rtx_REG (Pmode, GPR_RETURN);
  else
    return NULL_RTX;
}
/* }}} */

/* {{{ nios_go_if_legitimate_address()
   
   A C compound statement with a conditional `goto LABEL;' executed if X
   (an RTX) is a legitimate memory address on the target machine for a
   memory operand of mode MODE.

   It usually pays to define several simpler macros to serve as
   subroutines for this one.  Otherwise it may be too complicated to
   understand.

   This macro must exist in two variants: a strict variant and a
   non-strict one.  The strict variant is used in the reload pass.  It
   must be defined so that any pseudo-register that has not been
   allocated a hard register is considered a memory reference.  In
   contexts where some kind of register is required, a pseudo-register
   with no hard register must be rejected.

   The non-strict variant is used in other passes.  It must be defined
   to accept all pseudo-registers in every context where some kind of
   register is required.

   Compiler source files that want to use the strict variant of this
   macro define the macro `REG_OK_STRICT'.  You should use an `#ifdef
   REG_OK_STRICT' conditional to define the strict variant in that case
   and the non-strict variant otherwise.

   Subroutines to check for acceptable registers for various purposes
   (one for base registers, one for index registers, and so on) are
   typically among the subroutines used to define
   `GO_IF_LEGITIMATE_ADDRESS'.  Then only these subroutine macros need
   have two variants; the higher levels of macros may be the same
   whether strict or not.

   Normally, constant addresses which are the sum of a `symbol_ref' and
   an integer are stored inside a `const' RTX to mark them as constant.
   Therefore, there is no need to recognize such sums specifically as
   legitimate addresses.  Normally you would simply recognize any
   `const' as legitimate.

   Usually `PRINT_OPERAND_ADDRESS' is not prepared to handle constant
   sums that are not marked with `const'.  It assumes that a naked
   `plus' indicates indexing.  If so, then you *must* reject such naked
   constant sums as illegitimate addresses, so that none of them will be
   given to `PRINT_OPERAND_ADDRESS'.

   On some machines, whether a symbolic address is legitimate depends on
   the section that the address refers to.  On these machines, define
   the macro `ENCODE_SECTION_INFO' to store the information into the
   `symbol_ref', and then check for it here.  When you see a `const',
   you will have to look inside it to find the `symbol_ref' in order to
   determine the section.

   The best way to modify the name string is by adding text to the
   beginning, with suitable punctuation to prevent any ambiguity.
   Allocate the new name in `saveable_obstack'.  You will have to modify
   `ASM_OUTPUT_LABELREF' to remove and decode the added text and output
   the name accordingly, and define `STRIP_NAME_ENCODING' to access the
   original name string.

   You can check the information stored here into the `symbol_ref' in
   the definitions of the macros `GO_IF_LEGITIMATE_ADDRESS' and
   `PRINT_OPERAND_ADDRESS'.  */

#define RTX_OK_FOR_BASE_P(X, REG_OK_STRICT_P)				\
  ((REG_P (X) && nios_reg_ok_for_base_p (X, REG_OK_STRICT_P))		\
   || (GET_CODE (X) == SUBREG						\
       && REG_P (SUBREG_REG (X))					\
       && nios_reg_ok_for_base_p (SUBREG_REG (X), REG_OK_STRICT_P)))

#define RTX_OK_FOR_SIMPLE_P(X, REG_OK_STRICT_P)				\
  ((REG_P (X) && nios_reg_ok_for_simple_p (X, REG_OK_STRICT_P))	\
   || (GET_CODE (X) == SUBREG						\
       && REG_P (SUBREG_REG (X))					\
       && nios_reg_ok_for_simple_p (SUBREG_REG (X), REG_OK_STRICT_P)))

int
nios_go_if_legitimate_address (mode, x, reg_ok_strict_p)
    enum machine_mode mode;
    rtx x;
    int reg_ok_strict_p;
{
  int ret = 0;

  /* Any general purpose register can serve as a memory address.  */
  if (RTX_OK_FOR_SIMPLE_P (x, reg_ok_strict_p))
    ret = 1;

  /* Base + offset addressing. Only these types are allowed:
     	Pointer: Base register (%l0 - %l3) with 16-bit word-offset.
	Simple:  Any register with signed 11-bit word-offset.
	Stack:   Stack pointer with unsigned 8-bit word offset or unsigned
		 10-bit byte offset or unsigned 9-bit half-word offset.  */
  else if (GET_CODE (x) == PLUS 
	   && REG_P (XEXP (x, 0))
           && GET_CODE (XEXP (x, 1)) == CONST_INT
	   /* We only allow unaligned offsets for %fp and %sp because
	      we know that they're always aligned.  */
	   && (INTVAL (XEXP (x, 1)) % UNITS_PER_WORD == 0
	       || REGNO (XEXP (x, 0)) == GPR_FP
	       || REGNO (XEXP (x, 0)) == GPR_SP))
    {
      int word_offset = INTVAL (XEXP (x, 1)) / UNITS_PER_WORD;

      /* Pointer. In this case we cannot allow sub-word memory references
	 because nios does not support sub-word store operations using
	 `stp'. We need to force a `simple' memory reference and hope that
	 the offset is within range for `pfx'.  */
      if (RTX_OK_FOR_BASE_P (XEXP (x, 0), reg_ok_strict_p)
	  && CONST_OK_FOR_T (word_offset)
	  && GET_MODE_SIZE (mode) >= UNITS_PER_WORD)
	ret = 1;

      /* Simple, %sp and %fp.  */
      else if (RTX_OK_FOR_SIMPLE_P (XEXP (x, 0), reg_ok_strict_p) 
	       && CONST_OK_FOR_U (word_offset))
	ret = 1;
    }

  if (TARGET_DEBUG_ADDR)
    {
      fprintf (stderr, "\n[ GO_IF_LEGITIMATE_ADDRESS ], mode = %s, "
	       "result = %d, addresses are %sstrict\n",
	       GET_MODE_NAME (mode), ret,
	       (reg_ok_strict_p) ? "" : "not ");
      debug_rtx (x);
    }

  return ret;
}
/* }}} */

/* {{{ nios_reg_ok_for_simple_p()
   
   Return nonzero if X (assumed to be a `reg' RTX) is valid for use in
   simple addressing mode. Nios accepts any general purpose register to
   serve as a simple address with up to 11-bit wide word offsets
   (indicated with PFX).  */

int
nios_reg_ok_for_simple_p (x, reg_ok_strict_p)
    rtx x;
    int reg_ok_strict_p;
{
  return ((reg_ok_strict_p)
	   ? (GPR_P (REGNO (x)))
	   : (GPR_P (REGNO (x)) || REGNO (x) >= FIRST_PSEUDO_REGISTER));
}
/* }}} */

/* {{{ nios_reg_ok_for_base_p()
   
   A C expression that is nonzero if X (assumed to be a `reg' RTX) is
   valid for use as a base register.  For hard registers, it should
   always accept those which the hardware permits and reject the others.
   Whether the macro accepts or rejects pseudo registers must be
   controlled by `REG_OK_STRICT' as described above.  This usually
   requires two variant definitions, of which `REG_OK_STRICT' controls
   the one actually used. */

int
nios_reg_ok_for_base_p (x, reg_ok_strict_p)
    rtx x;
    int reg_ok_strict_p;
{
  return ((reg_ok_strict_p)
	   ? (REGNO_OK_FOR_BASE_P (REGNO (x)))
	   : (BP_REG_P (REGNO (x)) || REGNO (x) >= FIRST_PSEUDO_REGISTER));
}
/* }}} */

/* {{{ nios_reg_ok_for_index_p()
   
   A C expression that is nonzero if X (assumed to be a `reg' RTX) is
   valid for use as an index register.

   The difference between an index register and a base register is that
   the index register may be scaled.  If an address involves the sum of
   two registers, neither one of them scaled, then either one may be
   labeled the "base" and the other the "index"; but whichever labeling
   is used must fit the machine's constraints of which registers may
   serve in each capacity. The compiler will try both labelings, looking
   for one that is valid, and will reload one or both registers only if
   neither labeling works.  */

int
nios_reg_ok_for_index_p (x, reg_ok_strict_p)
    rtx x;
    int reg_ok_strict_p;
{
  return ((reg_ok_strict_p)
	   ? 0
	   : (REGNO (x) >= FIRST_PSEUDO_REGISTER));
}
/* }}} */

/* }}} */

/* {{{ Stack and frame support.  */

/* {{{                        The Nios Stack
		      (Offsets are for 16-bit words)


             Before call		       	      After call
 HIGH   +-----------------------+		+-----------------------+
 MEMORY |			|		|			|
	|   Caller's locals	|       	|    Caller's locals	|
        |			|		|			|
        +-----------------------+		+-----------------------+
        |			|		|			|
 	|   alloca() space	|		|    alloca() space	|
        |			|		|			|
        +-----------------------+		+-----------------------+
        |			|		|			|
	|   Arguments past   	|		|   Arguments past   	|
        |  the 6th argument.    |		|  the 6th argument.    |
        |			|		|			|
SP+46-> +-----------------------+      FP+46->	+-----------------------+
        |			|		|			|
 	|   6 words to save	|		|   6 words to save	|
	|   arguments passed	|		|   arguments passed	|
	|  in registers, even	|		|  in registers, even	|
       	|    if not passed.	|      		|    if not passed.	|
        |			|		|			|
SP+34-> +-----------------------+      FP+34->	+-----------------------+
        |			|		|			|
        | 1 word hidden address |		| 1 word hidden address |
	|  for callee to store	|               |  for callee to store	|
	|   aggregate return	|               |   aggregate return	|
	|        values.	|               |        values.	|
        |			|		|			|
SP+32-> +-----------------------+      FP+32->	+-----------------------+
        |			|		|			|
        | 16 word reg save area	|		| 16 word reg save area |
       	|                       |      		|			|
   SP-> +-----------------------+         FP->	+-----------------------+
				        	|			|
				        	|    Callee's locals	|
				        	|			|
				        	+-----------------------+
				        	|		        |
                                        	|    registers saved	|
                                        	|       		|
				        	+-----------------------+
				        	|			|
                                        	|    alloca() space	|
        			        	|			|
				        	+-----------------------+
				        	|			|
                                                |   Arguments past   	|
        			                |  the 6th argument.    |
                                                |			|
					SP+46-> +-----------------------+	
                                                |			|
                                         	|   6 words to save	|
                                        	|   arguments passed	|
                                        	|  in registers, even	|
                                               	|    if not passed.	|
                                                |			|
                                        SP+34-> +-----------------------+
                                                |			|
                                                | 1 word hidden address |
                                        	|  for callee to store	|
                                        	|   aggregate return	|
                                        	|        values.	|
                                                |			|
                                        SP+32-> +-----------------------+
                                                |			|
                                                | 16 word reg save area	|
 LOW                                           	|                       |
 MEMORY                                    SP-> +-----------------------+

}}}  */

/* {{{ nios_compute_frame_size()
   
   Returns the number of bytes offset between the frame pointer and the
   stack pointer for the current function.  SIZE is the number of bytes
   of space needed for local variables.  */

int
nios_compute_frame_size (size, leaf_function)
     int size;
     int leaf_function;
{
  int n_regs = 0, i;
  int outgoing_args_size = (current_function_outgoing_args_size
			    + REG_PARM_STACK_SPACE (current_function_decl));

  /* Check which global registers (%r0 - %r7) need to be saved.  */
  for (i = 0; i < 8; i += 2)
    if ((regs_ever_live[i] && ! call_used_regs[i])
	|| (regs_ever_live[i + 1] && ! call_used_regs[i + 1]))
      n_regs += 2;

  /* Set up values for use in `nios_function_epilogue'.  */
  if (leaf_function
      && n_regs == 0
      && size == 0 
      && current_function_outgoing_args_size == 0)
    {
      actual_fsize = apparent_fsize = 0;
    }
  else
    {
      /* We subtract STARTING_FRAME_OFFSET, remember it's negative.  */
      apparent_fsize = NIOS_STACK_ALIGN (size - STARTING_FRAME_OFFSET);

      /* Add the registers that need saving.  */
      apparent_fsize += n_regs * UNITS_PER_WORD;

      /* Add arguments past the 6th argument and space for the 6 register
	 arguments even if they're all passed in registers.  */
      actual_fsize = apparent_fsize + NIOS_STACK_ALIGN (outgoing_args_size);

      /* Reserve 1 word of space for a hidden pointer to an aggregate value
	 returned by our callees.  */
      actual_fsize += UNITS_PER_WORD;

      /* Make sure nothing can clobber our register windows. If a SAVE must
	 be done, or there is a stack-local variable, the register window
	 area must be allocated.  */
      actual_fsize += (NIOS_REGISTER_WINDOW_SIZE * UNITS_PER_WORD);
    }

  if (TARGET_DEBUG_STACK)
    {
      fprintf (stderr, "\n\nStack size information for function %s:\n",
	   ((current_function_decl && DECL_NAME (current_function_decl))
	    ? IDENTIFIER_POINTER (DECL_NAME (current_function_decl))
	    : "<unknown>"));

      fprintf (stderr, "\tApparent frame size              %4d bytes\n", 
	       apparent_fsize);
      fprintf (stderr, "\n");
      fprintf (stderr, "\tStarting frame offset            %4d bytes\n",
	       STARTING_FRAME_OFFSET);
      fprintf (stderr, "\tSize of locals                   %4d bytes\n", size);
      fprintf (stderr, "\tRegisters to save (%2d)           %4d bytes\n", 
	       n_regs,
	       n_regs * UNITS_PER_WORD);
      fprintf (stderr, "\tSize of outgoing arguments       %4d bytes\n", 
	       current_function_outgoing_args_size);
      fprintf (stderr, "\tSpace reserved for reg arguments %4d bytes\n",
	       (actual_fsize) 
	        ? REG_PARM_STACK_SPACE (current_function_decl)
		: 0);
      fprintf (stderr, "\tSpace reserved for hidden ptr    %4d bytes\n",
	       (actual_fsize) 
	        ? UNITS_PER_WORD
		: 0);
      fprintf (stderr, "\tSpace reserved for window spills %4d bytes\n", 
	       (actual_fsize) 
	        ? (NIOS_REGISTER_WINDOW_SIZE * UNITS_PER_WORD)
		: 0);
      fprintf (stderr, "\t                                 ----\n");
      fprintf (stderr, "\tActual frame size                %4d bytes\n",
	       NIOS_STACK_ALIGN (actual_fsize));
      fprintf (stderr, "\n");
    }

  return NIOS_STACK_ALIGN (actual_fsize);
}
/* }}} */

/* {{{ nios_function_prologue()
   
   A C compound statement that outputs the assembler code for entry to a
   function.  The prologue is responsible for setting up the stack
   frame, initializing the frame pointer register, saving registers that
   must be saved, and allocating SIZE additional bytes of storage for
   the local variables.  SIZE is an integer.  FILE is a stdio stream to
   which the assembler code should be output.

   The label for the beginning of the function need not be output by
   this macro.  That has already been done when the macro is run.

   To determine which registers to save, the macro can refer to the
   array `regs_ever_live': element R is nonzero if hard register R is
   used anywhere within the function.  This implies the function
   prologue should save register R, provided it is not one of the
   call-used registers. (`FUNCTION_EPILOGUE' must likewise use
   `regs_ever_live'.)

   On machines that have "register windows", the function entry code
   does not save on the stack the registers that are in the windows,
   even if they are supposed to be preserved by function calls; instead
   it takes appropriate steps to "push" the register stack, if any
   non-call-used registers are used in the function.

   On machines where functions may or may not have frame-pointers, the
   function entry code must vary accordingly; it must set up the frame
   pointer if one is wanted, and not otherwise.  To determine whether a
   frame pointer is in wanted, the macro can refer to the variable
   `frame_pointer_needed'.  The variable's value will be 1 at run time
   in a function that needs a frame pointer.

   The function entry code is responsible for allocating any stack space
   required for the function.  This stack space consists of the regions
   listed below.  In most cases, these regions are allocated in the
   order listed, with the last listed region closest to the top of the
   stack (the lowest address if `STACK_GROWS_DOWNWARD' is defined, and
   the highest address if it is not defined).  You can use a different
   order for a machine if doing so is more convenient or required for
   compatibility reasons.  Except in cases where required by standard or
   by a debugger, there is no reason why the stack layout used by GCC
   need agree with that used by other compilers for a machine.

      * A region of `current_function_pretend_args_size' bytes of
	uninitialized space just underneath the first argument arriving
	on the stack.  (This may not be at the very start of the
	allocated stack region if the calling sequence has pushed
	anything else since pushing the stack arguments.  But usually,
	on such machines, nothing else has been pushed yet, because the
	function prologue itself does all the pushing.) This region is
	used on machines where an argument may be passed partly in
	registers and partly in memory, and, in some cases to support
	the features in `varargs.h' and `stdargs.h'.

      * An area of memory used to save certain registers used by the
	function.  The size of this area, which may also include space
	for such things as the return address and pointers to previous
	stack frames, is machine-specific and usually depends on which
	registers have been used in the function. Machines with register
	windows often do not require a save area.

      * A region of at least SIZE bytes, possibly rounded up to an
	allocation boundary, to contain the local variables of the
	function.  On some machines, this region and the save area may
	occur in the opposite order, with the save area closer to the
	top of the stack.

      * Optionally, when `ACCUMULATE_OUTGOING_ARGS' is defined, a region
	of `current_function_outgoing_args_size' bytes to be used for
	outgoing argument lists of the function.

   Normally, it is necessary for the macros `FUNCTION_PROLOGUE' and
   `FUNCTION_EPILOGUE' to treat leaf functions specially.  The C
   variable `leaf_function' is nonzero for such a function.  */

void
nios_function_prologue (stream, size, leaf_function)
     FILE *stream;
     int size;
     int leaf_function;
{
  /* Need to use actual_fsize, since we are also allocating
     space for our callee (and our own register save area).  */
  actual_fsize = nios_compute_frame_size (size, leaf_function);

  if (leaf_function)
    {
      frame_base_name = "%sp";
      frame_base_offset = actual_fsize;
    }
  else
    {
      frame_base_name = "%fp";
      frame_base_offset = 0;
    }

  /* This is only for the human reader.  */
  fprintf (stream, "\t%s start prologue\n", ASM_COMMENT_START);

  nios_emit_stack_reservation (stream, leaf_function);

  if (dwarf2out_do_frame () && actual_fsize)
    {
      char *label = dwarf2out_cfi_label ();

      /* The canonical frame address refers to the top of the frame.  */
      dwarf2out_def_cfa (label, (leaf_function ? STACK_POINTER_REGNUM
				 : FRAME_POINTER_REGNUM),
			 frame_base_offset);

      if (! leaf_function)
	{
	  /* Note the register window save.  This tells the unwinder that
	     it needs to restore the window registers from the previous
	     frame's window save area at 0(cfa).  */
	  dwarf2out_window_save (label);

	  /* The return address (-8) is now in %i7.  */
	  dwarf2out_return_reg (label, 31);
	}
    }

  /* If doing anything with PIC, do it now.  */
  if (! flag_pic)
    fprintf (stream, "\t%s end prologue\n", ASM_COMMENT_START);

  leaf_label = 0;
  if (leaf_function && actual_fsize != 0)
      warning ("leaf procedure with frame size %d", actual_fsize);
}
/* }}} */

/* {{{ nios_emit_stack_reservation()
   
   Emit code to reserve actual_fsize words on the stack.  */

static void
nios_emit_stack_reservation (stream, leaf_function)
    FILE *stream;
    int leaf_function;
{
  int reserve_with_save, reserve_extra;
     
  if (actual_fsize == 0)
    return;

  /* Frame must be word aligned.  */
  if (actual_fsize % UNITS_PER_WORD != 0)
    abort();

  /* Non-leaf functions should save the first NIOS_MAX_SAVE_SIZE words with
     `save'. The remaining words are saved using pfx/subi as necessary.  */
  if (! leaf_function)
    {
      /* We can only save NIOS_MAX_SAVE_SIZE words with the 'save'
	 instruction. Anything beyond NIOS_MAX_SAVE_SIZE must be saved
	 by explicitly subtracting from %sp.  */
      actual_fsize /= UNITS_PER_WORD;

      if (actual_fsize > NIOS_MAX_SAVE_SIZE)
	{
	  reserve_with_save = NIOS_MAX_SAVE_SIZE;
	  /* Additional frame must be reserved in bytes.  */
	  reserve_extra = (actual_fsize - NIOS_MAX_SAVE_SIZE) * UNITS_PER_WORD;
	}
      else
	{
	  reserve_with_save = actual_fsize;
	  reserve_extra = 0;
	}

      fprintf (stream, "\tsave\t%%sp, -%d\n", reserve_with_save);
    }
  else /* leaf function */
    reserve_extra = actual_fsize;

  if (reserve_extra == 0)
    return;

  /* Whatever space we couldn't reserve with `save', must be preserved with
     explicit manipulation of %sp.  */
  if (CONST_OK_FOR_L (reserve_extra))
    fprintf (stream, "\tsubi\t%%sp, %d\n", reserve_extra);

  else if (CONST_OK_FOR_N (reserve_extra))
    fprintf (stream, "\tpfx\t%%hi(%d)\n\tsubi\t%%sp, %%lo(%d)\n",
	     reserve_extra, reserve_extra);

  else
    {
      /* Subtract values larger than 0xffff from %sp.  */
      if (TARGET_32BIT)
	{
	  /* On delta-32 we move the value on %r0 and subtract. Since the
	     size of the stack cannot be a number with more than 32 bits,
	     the movi/movhi combination will always work.  */
	  fprintf (stream, "\tpfx\t%%hi(%d)\n\tmovi\t%%r0, %%lo(%d)\n",
		   reserve_extra, reserve_extra);

	  fprintf (stream, "\tpfx\t%%xhi(%d)\n\tmovhi\t%%r0, %%xlo(%d)\n",
		   reserve_extra, reserve_extra);

	  fputs ("\tsub\t%sp, %r0\n", stream);
	}
      else
	{
	  /* For delta-16, pfx cannot deal with values larger than 0xffff.
	     So, we issue several pfx/subi combinations as necessary.  */
	  while (reserve_extra > 0)
	    {
	      int val = (reserve_extra <= 0xffff) ? reserve_extra : 0xffff;

	      if (CONST_OK_FOR_L (val))
		fprintf (stream, "\tsubi\t%%sp, %d\n", val);
	      else
		fprintf (stream, "\tpfx\t%%hi(%d)\n\tsubi\t%%sp, %%lo(%d)\n",
			 val, val);

	      reserve_extra -= val;
	    }
	}
    }
}
/* }}} */

/* {{{ nios_function_epilogue()
   
   A C compound statement that outputs the assembler code for exit from
   a function.  The epilogue is responsible for restoring the saved
   registers and stack pointer to their values when the function was
   called, and returning control to the caller.  This macro takes the
   same arguments as the macro `FUNCTION_PROLOGUE', and the registers to
   restore are determined from `regs_ever_live' and
   `CALL_USED_REGISTERS' in the same way.

   On some machines, there is a single instruction that does all the
   work of returning from the function.  On these machines, give that
   instruction the name `return' and do not define the macro
   `FUNCTION_EPILOGUE' at all.

   Do not define a pattern named `return' if you want the
   `FUNCTION_EPILOGUE' to be used.  If you want the target switches to
   control whether return instructions or epilogues are used, define a
   `return' pattern with a validity condition that tests the target
   switches appropriately.  If the `return' pattern's validity condition
   is false, epilogues will be used.

   On machines where functions may or may not have frame-pointers, the
   function exit code must vary accordingly.  Sometimes the code for
   these two cases is completely different.  To determine whether a
   frame pointer is wanted, the macro can refer to the variable
   `frame_pointer_needed'. The variable's value will be 1 when compiling
   a function that needs a frame pointer.

   Normally, `FUNCTION_PROLOGUE' and `FUNCTION_EPILOGUE' must treat leaf
   functions specially.  The C variable `leaf_function' is nonzero for
   such a function.

   On some machines, some functions pop their arguments on exit while
   others leave that for the caller to do.  For example, the 68020 when
   given `-mrtd' pops arguments in functions that take a fixed number of
   arguments.

   Your definition of the macro `RETURN_POPS_ARGS' decides which
   functions pop their own arguments.  `FUNCTION_EPILOGUE' needs to know
   what was decided. The variable that is called
   `current_function_pops_args' is the number of bytes of its arguments
   that a function should pop.  */

void
nios_function_epilogue (stream, size, leaf_function)
     FILE *stream;
     int size ATTRIBUTE_UNUSED;
     int leaf_function;
{
  if (! leaf_function)
    fprintf (stream, "\t" 
	     "ret\n\t"
	     "restore\n");
  else
    {
      if (CONST_OK_FOR_L (actual_fsize))
	fprintf (stream, "\t"
	         "lret\n\t"
		 "addi\t%%sp, %d\n", 
		 actual_fsize);
      else
	fprintf (stream, "\t"
	         "pfx\t%%hi(%d)\n\t"
		 "lret\n\t"
		 "addi\t%%sp, %%lo(%d)\n",
		 actual_fsize, actual_fsize);
    }
}
/* }}} */

/* {{{ nios_direct_return_p()
   
   Return true if the bare return instruction can be used outside of the
   epilogue code.  For nios, we only do it if there was no stack
   allocation.  */

int
nios_direct_return_p ()
{
  if (! reload_completed)
    return FALSE;

  return (current_function_uses_only_leaf_regs);
}
/* }}} */

/* }}} */

/* {{{ Argument passing.  */

/* {{{ nios_init_cumulative()

   A C statement (sans semicolon) for initializing the variable CUM for
   the state at the beginning of the argument list.  The variable has
   type `CUMULATIVE_ARGS'.  The value of FNTYPE is the tree node for the
   data type of the function which will receive the args, or 0 if the
   args are to a compiler support library function.  The value of
   INDIRECT is nonzero when processing an indirect call, for example a
   call through a function pointer. The value of INDIRECT is zero for a
   call to an explicitly named function, a library function call, or
   when `INIT_CUMULATIVE_ARGS' is used to find arguments for the
   function being compiled.

   When processing a call to a compiler support library function,
   LIBNAME identifies which one.  It is a `symbol_ref' rtx which
   contains the name of the function, as a string.  LIBNAME is 0 when an
   ordinary C function call is being processed.  Thus, each time this
   macro is called, either LIBNAME or FNTYPE is nonzero, but never both
   of them at once.  */

void
nios_init_cumulative_args (cum, fntype, libname, indirect, incoming)
     CUMULATIVE_ARGS *cum;
     tree fntype;
     rtx libname;
     int indirect;
     int incoming;
{
  cum->words = 0;

  if (TARGET_DEBUG_ARG)
    {
      fprintf (stderr, "\nnios_init_cumulative_args:");
      if (indirect)
	fputs (" indirect", stderr);

      if (incoming)
	fputs (" incoming", stderr);

      if (fntype)
	{
	  tree ret_type = TREE_TYPE (fntype);
	  fprintf (stderr, " return=%s,",
		   tree_code_name[ (int)TREE_CODE (ret_type) ]);
	}

      if (libname && GET_CODE (libname) == SYMBOL_REF)
	fprintf (stderr, " libname=%s", XSTR (libname, 0));

      if (cfun->returns_struct)
	fprintf (stderr, " return-struct");

      putc ('\n', stderr);
    }
}
/* }}} */

/* {{{ nios_function_arg_pass()

   A C expression that indicates when an argument must be passed by
   reference. If nonzero for an argument, a copy of that argument is
   made in memory and a pointer to the argument is passed instead of the
   argument itself.  The pointer is passed in whatever way is
   appropriate for passing a pointer to that type.  */

int
nios_function_arg_pass_by_reference (cum, mode, type, named)
     CUMULATIVE_ARGS *cum ATTRIBUTE_UNUSED;
     enum machine_mode mode ATTRIBUTE_UNUSED;
     tree type;
     int named ATTRIBUTE_UNUSED;
{
  return (type && int_size_in_bytes (type) > 4 * UNITS_PER_WORD);
}
/* }}} */

/* {{{ nios_function_arg_boundary()

   If defined, a C expression that gives the alignment boundary, in
   bits, of an argument with the specified mode and type.  If it is not
   defined, `PARM_BOUNDARY' is used for all arguments.  */

int
nios_function_arg_boundary (mode, type)
     enum machine_mode mode ATTRIBUTE_UNUSED;
     tree type ATTRIBUTE_UNUSED;
{
  return BITS_PER_WORD;
}
/* }}} */

/* {{{ nios_function_arg_slotno()

   Compute the slot number to pass an argument in.
   Returns the slot number or -1 if passing on the stack.

   CUM is a variable of type CUMULATIVE_ARGS which gives info about
    the preceding args and about the function being called.
   MODE is the argument's machine mode.
   TYPE is the data type of the argument (as a tree).
    This is null for libcalls where that information may
    not be available.
   NAMED is nonzero if this argument is a named parameter
    (otherwise it is an extra parameter matching an ellipsis).
   INCOMING_P is zero for FUNCTION_ARG, nonzero for FUNCTION_INCOMING_ARG.
   *PREGNO records the register number to use if scalar type.  */

static int
nios_function_arg_slotno (cum, mode, type, named, incoming_p, pregno)
     const CUMULATIVE_ARGS *cum;
     enum machine_mode mode;
     tree type;
     int named ATTRIBUTE_UNUSED;
     int incoming_p;
     int *pregno;
{
  int regbase = (incoming_p)
		 ? NIOS_INCOMING_INT_ARG_FIRST
		 : NIOS_OUTGOING_INT_ARG_FIRST;
  int slotno = cum->words;

  if (MUST_PASS_IN_STACK (mode, type))
    return -1;

  if (slotno >= NIOS_NUM_ARG_REGS)
    return -1;

  *pregno = regbase + slotno;

  return slotno;
}
/* }}} */

/* {{{ nios_function_arg()

   A C expression that controls whether a function argument is passed in
   a register, and which register.

   The arguments are CUM, of type CUMULATIVE_ARGS, which summarizes (in
   a way defined by INIT_CUMULATIVE_ARGS and FUNCTION_ARG_ADVANCE) all
   of the previous arguments so far passed in registers; MODE, the
   machine mode of the argument; TYPE, the data type of the argument as
   a tree node or 0 if that is not known (which happens for C support
   library functions); and NAMED, which is 1 for an ordinary argument
   and 0 for nameless arguments that correspond to `...' in the called
   function's prototype.

   The value of the expression should either be a `reg' RTX for the hard
   register in which to pass the argument, or zero to pass the argument
   on the stack.

   The usual way to make the ANSI library `stdarg.h' work on a machine
   where some arguments are usually passed in registers, is to cause
   nameless arguments to be passed on the stack instead.  This is done
   by making `FUNCTION_ARG' return 0 whenever NAMED is 0.

   You may use the macro `MUST_PASS_IN_STACK (MODE, TYPE)' in the
   definition of this macro to determine if this argument is of a type
   that must be passed in the stack.  If `REG_PARM_STACK_SPACE' is not
   defined and `FUNCTION_ARG' returns non-zero for such an argument, the
   compiler will abort.  If `REG_PARM_STACK_SPACE' is defined, the
   argument will be computed in the stack and then loaded into a
   register.

   INCOMING_P is zero for FUNCTION_ARG, nonzero for FUNCTION_INCOMING_ARG.  */

rtx
nios_function_arg (cum, mode, type, named, incoming_p)
     const CUMULATIVE_ARGS *cum;
     enum machine_mode mode;
     tree type;
     int named;
     int incoming_p;
{
  int slotno, regno;
  rtx reg;

  slotno = nios_function_arg_slotno (cum, mode, type, named, incoming_p, 
                                      &regno);

  if (slotno == -1)
    reg = NULL_RTX;
  else
    reg = gen_rtx_REG (mode, regno);

  return reg;
}
/* }}} */

/* {{{ nios_function_arg_advance()
   
   A C statement (sans semicolon) to update the summarizer variable CUM
   to advance past an argument in the argument list.  The values MODE,
   TYPE and NAMED describe that argument.  Once this is done, the
   variable CUM is suitable for analyzing the *following* argument with
   `FUNCTION_ARG', etc.

   This macro need not do anything if the argument in question was
   passed on the stack.  The compiler knows how to track the amount of
   stack space used for arguments without any special help.  */

void
nios_function_arg_advance (cum, mode, type, named)
     CUMULATIVE_ARGS *cum;
     enum machine_mode mode;
     tree type ATTRIBUTE_UNUSED;
     int named;
{
  int slotno, regno;

  /* We pass 0 for incoming_p here, it doesn't matter.  */
  slotno = nios_function_arg_slotno (cum, mode, type, named, 0, &regno);

  cum->words += (mode != BLKmode
		 ? ROUND_ADVANCE (GET_MODE_SIZE (mode))
		 : ROUND_ADVANCE (int_size_in_bytes (type)));

  if (TARGET_DEBUG_ARG)
    fprintf (stderr,
	     "nios_function_arg_advance: words = %2d, mode = %4s, "
	     "named = %d, size = %3d\n",
	     cum->words, GET_MODE_NAME (mode), named, 
	     cum->words * UNITS_PER_WORD);
}
/* }}} */

/* {{{ nios_function_arg_partial_nregs()
   
   A C expression for the number of words, at the beginning of an
   argument, must be put in registers.  The value must be zero for
   arguments that are passed entirely in registers or that are entirely
   pushed on the stack.

   On some machines, certain arguments must be passed partially in
   registers and partially in memory.  On these machines, typically the
   first N words of arguments are passed in registers, and the rest on
   the stack.  If a multi-word argument (a `double' or a structure)
   crosses that boundary, its first few words must be passed in
   registers and the rest must be pushed. This macro tells the compiler
   when this occurs, and how many of the words should go in registers.

   `FUNCTION_ARG' for these arguments should return the first register
   to be used by the caller for this argument; likewise
   `FUNCTION_INCOMING_ARG', for the called function.  */

int
nios_function_arg_partial_nregs (cum, mode, type, named)
     CUMULATIVE_ARGS *cum;
     enum machine_mode mode;
     tree type ATTRIBUTE_UNUSED;
     int named ATTRIBUTE_UNUSED;
{
  int slotno, regno;
  int ret;

  /* We pass 0 for incoming_p here, it doesn't matter.  */
  slotno = nios_function_arg_slotno (cum, mode, type, named, 0, &regno);

  if (slotno == -1)
    ret = 0;
  else if ((slotno + (mode == BLKmode
		      ? ROUND_ADVANCE (int_size_in_bytes (type))
		      : ROUND_ADVANCE (GET_MODE_SIZE (mode))))
	    > NIOS_NUM_ARG_REGS)
    ret = NIOS_NUM_ARG_REGS - slotno;
  else
    ret = 0;

  if (TARGET_DEBUG_ARG)
    fprintf (stderr, "nios_function_arg_partial_nregs: %d\n", ret);

  return ret;
}
/* }}} */

/* }}} */

/* {{{ Assembly output support.  */

/* {{{ nios_output_bit_test_branch()
   
  Output code for a conditional branch using bit-tests. On the nios we
  can do short jumps within +/-1K of the jump location. If we can't do a
  short jump, we reverse the sign of the conditional and do a long jump.

  Argument `is_true_branch' determines whether we are emitting an
  if-true branch or an if-false branch.  */



char *
nios_output_bit_test_branch (is_true_branch, insn, op, cmp_op)
    int is_true_branch;
    rtx insn;
    rtx *op;
    enum rtx_code cmp_op;
{
  static char ret[256];

  if (get_attr_length (insn) == 2)
    {
      /* Emit code for short conditional branches.  */

      /* When testing for zero bits, we want to avoid executing the branch
	 when the bit tested is not zero. Thus we issue skp1 for if-true
	 EQ comparisons or if-false NE comparisons.  */
      if ((cmp_op == EQ && is_true_branch)
	   || (cmp_op == NE && !is_true_branch))
	strcpy (ret, "skp1\t%0, %1\n\tbr\t%l2%#");
      else
	strcpy (ret, "skp0\t%0, %1\n\tbr\t%l2%#");
    }
  else
    {
      /* Emit code for long conditional branches.  */

      rtx real_label;
      rtx opposite_label;
      rtx xop[6];

      /* Generate labels for the real and opposite branch.   */
      real_label = gen_label_rtx ();
      opposite_label = gen_label_rtx ();

      /* Save the operands so that we can output the branch to the original
	 label after filling in the delay slot.  This is needed because
	 final_scan_insn overwrites the operand array.  The two extra
	 operands are the labels for jumping around the real branching code
	 and storing the address for the target location.  */
      xop[0] = op[0];		/* Register to test.  */
      xop[1] = op[1];		/* Bit number to extract.  */
      xop[2] = op[2];		/* Target location.  */
      xop[3] = real_label;	/* Label to hold target address.  */
      xop[4] = opposite_label;	/* Opposite branch. */
      xop[5] = NULL;


      /* Conditional branch with the opposite conditional value to jump
	 around the real branching code. Since we need to do a long
	 conditional jump, we do two things:
       
	 1- Emit an unconditional long jump to the target label
	    (real_label). 
	 2- Conditionally jump around the unconditional jump using the
	    opposite sign from the original condition.  

	 Original code		Transformed code
	  'if (a&8)'

	 skp0	%i0, 3		skp1	%i0, 3
	 br	target		br	opposite_label
				nop		;; or delay slot filler.
	 			jmpc	[real_label]
				nop
				.align 2
		    real_label:
				.nword target@h
		opposite_label:						 */

      if ((cmp_op == EQ && is_true_branch)
	   || (cmp_op == NE && !is_true_branch))
	strcpy (ret, "skp0\t%0, %1\n\tbr\t%l4%#");
      else
	strcpy (ret, "skp1\t%0, %1\n\tbr\t%l4%#");

      output_asm_insn (ret, xop);

      /* If the delay slot has been filled, emit the instruction into the
	 delay slot of the conditional branch emitted above.  */
      if (dbr_sequence_length () > 0)
	{
	  rtx delay_slot_insn = NEXT_INSN (insn);
	  final_scan_insn (delay_slot_insn, asm_out_file, 1, 0, 1);

	  /* Delete the delay slot insn to avoid emitting it again.  */
	  PUT_CODE (delay_slot_insn, NOTE);
	  NOTE_LINE_NUMBER (delay_slot_insn) = NOTE_INSN_DELETED;
	  NOTE_SOURCE_FILE (delay_slot_insn) = 0;
	}
      else
	{
	  /* Delay slot cannot be filled. Output a `nop'. */
	  output_asm_insn ("nop", xop);
	}

      /* Branch to the original target. Emit code to do an indirect jump.
	 Also write out the opposite label for the skip jump we emitted
	 before.  */

// -------
// dvb/cv 2001 Oct 23 -- conditionally generate an alternative to jmpc
// the else clause is the normal code generation as provided by cygnus
// we use "ret" since it looks safe enough

#if DONT_EMIT_JMPC

      sprintf(ret,"%s",         "\t pfx %%hi(%l2@h) \n");
      sprintf(ret,"%s%s",ret,   "\t movi %%g0,%%lo(%l2@h) \n");
      if(TARGET_32BIT)
        {
	  if (!nios_suppress_address_xhi)
	    sprintf(ret,"%s%s",ret, "\t pfx %%xhi(%l2@h) \n");

	  if (!nios_suppress_address_xlo)
	    sprintf(ret,"%s%s",ret, "\t movhi %%g0,%%xlo(%l2@h) \n");
        }
      sprintf(ret,"%s%s",ret,   "\t jmp %%g0 \n");
      sprintf(ret,"%s%s",ret,   "\t nop ; (delay slot) \n");

// we are careful to emit the same number of labels as the code we are
// replacing, else the compiler gets confused or something.

      sprintf(ret,"%s%s",ret,   "%l3:\n%l4:");
      output_asm_insn(ret,xop);


#else
      output_asm_insn (
	  "jmpc\t[%l3]\n\tnop\n\t.align 2\n%l3:\n\t.nword %l2@h\n%l4:",
	  xop);
#endif
// ------

      ret[0] = '\0';		/* Don't need to emit anything else. */
    }

  return ret;
}
/* }}} */

/* {{{ nios_output_cond_branch()
   
  Output code for a conditional branch. On the nios we can do short
  jumps within +/-1K of the jump location. If we can't do a short jump,
  we reverse the sign of the conditional and do a long jump.

  Argument `is_true_branch' determines whether we are emitting an
  if-true branch or an if-false branch. 

  Argument `is_bit_test' is true if the conditional comes from a
  bit-test.  */

char *
nios_output_cond_branch (is_true_branch, insn, op)
    int is_true_branch;
    rtx insn;
    rtx *op;
{
  static char ret[256];

  if (get_attr_length (insn) == 2)
    {
      /* Emit code for short conditional branches.  */

      if ((GET_CODE (op[0]) == EQ || GET_CODE (op[0]) == NE)
	   && GET_CODE (op[2]) == CONST_INT
	   && INTVAL (op[2]) == 0)
	{
	  /* Special case for zero-tests. Use skprz/skprnz.  */
	  if ((GET_CODE (op[0]) == EQ && is_true_branch)
	      || (GET_CODE (op[0]) == NE && !is_true_branch))
	    strcpy (ret, "skprnz\t%1\n\tbr\t%l3%#");
	  else
	    strcpy (ret, "skprz\t%1\n\tbr\t%l3%#");
	}
      else
	{
	  /* Regular compare and branch.  */
	  strcpy (ret, nios_output_2op_insn (NIOS_CMP, op, 1, 2, insn));
	  if (is_true_branch)
	    sprintf (ret, "%s\n\tb%%B0\t%%l3%%#", ret);
	  else
	    sprintf (ret, "%s\n\tb%%b0\t%%l3%%#", ret);
	}
    }
  else
    {
      /* Emit code for long conditional branches.  */

      rtx real_label;
      rtx opposite_label;
      rtx xop[7];

      /* Generate labels for the real and opposite branch.   */
      real_label = gen_label_rtx ();
      opposite_label = gen_label_rtx ();

      /* Save the operands so that we can output the branch to the original
	 label after filling in the delay slot.  This is needed because
	 final_scan_insn overwrites the operand array.  The two extra
	 operands are the labels for jumping around the real branching code
	 and storing the address for the target location.  */
      xop[0] = op[0];		/* Comparison operator.  */
      xop[1] = op[1];		/* First comparison operand.  */
      xop[2] = op[2];		/* Second comparison operand.  */
      xop[3] = op[3];		/* Target location.  */
      xop[4] = real_label;	/* Label to hold target address.  */
      xop[5] = opposite_label;	/* Opposite branch. */
      xop[6] = NULL;


      /* Conditional branch with the opposite conditional value to jump
	 around the real branching code. Since we need to do a long
	 conditional jump, we do two things:
       
	 1- Emit an unconditional long jump to the target label
	    (real_label). 
	 2- Conditionally jump around the unconditional jump using the
	    opposite sign from the original condition.  

	 Original code		Transformed code

	 beq target		bne opposite_label
				nop	;; or delay slot filler.
	 			jmpc [real_label]
				nop
				.align 2
		    real_label:
				.nword target@h
		opposite_label:
       */

      if ((GET_CODE (op[0]) == EQ || GET_CODE (op[0]) == NE)
	   && GET_CODE (op[2]) == CONST_INT
	   && INTVAL (op[2]) == 0)
	{
	  /* Special case for zero-tests. Use skprz/skprnz.  */
	  if ((GET_CODE (op[0]) == EQ && is_true_branch)
	      || (GET_CODE (op[0]) == NE && !is_true_branch))
	    strcpy (ret, "skprz\t%1\n\tbr\t%l5%#");
	  else
	    strcpy (ret, "skprnz\t%1\n\tbr\t%l5%#");
	}
      else
	{
	  /* Regular compare and branch.  */
	  strcpy (ret, nios_output_2op_insn (NIOS_CMP, xop, 1, 2, insn));
	  if (is_true_branch)
	    sprintf (ret, "%s\n\tb%%b0\t%%l5%%#", ret);
	  else
	    sprintf (ret, "%s\n\tb%%B0\t%%l5%%#", ret);
	}

      output_asm_insn (ret, xop);

      /* If the delay slot has been filled, emit the instruction into the
	 delay slot of the conditional branch emitted above.  */
      if (dbr_sequence_length () > 0)
	{
	  rtx delay_slot_insn = NEXT_INSN (insn);
	  final_scan_insn (delay_slot_insn, asm_out_file, 1, 0, 1);

	  /* Delete the delay slot insn to avoid emitting it again.  */
	  PUT_CODE (delay_slot_insn, NOTE);
	  NOTE_LINE_NUMBER (delay_slot_insn) = NOTE_INSN_DELETED;
	  NOTE_SOURCE_FILE (delay_slot_insn) = 0;
	}
      else
	{
	  /* Delay slot cannot be filled. Output a `nop'. */
	  output_asm_insn ("nop", xop);
	}

      /* Branch to the original target. Emit code to do an indirect jump.
	 Also write out the opposite label for the skip jump we emitted
	 before.  */


// ------
// dvb/cv 2001 Oct 23 -- conditionally generate an alternative to jmpc
// the else clause is the normal code generation as provided by cygnus
// we use "ret" since it looks safe enough

#if DONT_EMIT_JMPC

      sprintf(ret,"%s",         "\t pfx %%hi(%l3@h) \n");
      sprintf(ret,"%s%s",ret,   "\t movi %%g0,%%lo(%l3@h) \n");
      if(TARGET_32BIT)
        {
	  if (!nios_suppress_address_xhi)
	    sprintf(ret,"%s%s",ret, "\t pfx %%xhi(%l3@h) \n");

	  if (!nios_suppress_address_xlo)
	    sprintf(ret,"%s%s",ret, "\t movhi %%g0,%%xlo(%l3@h) \n");
        }
      sprintf(ret,"%s%s",ret,   "\t jmp %%g0 \n");
      sprintf(ret,"%s%s",ret,   "\t nop ; (delay slot) \n");

// we are careful to emit the same number of labels as the code we are
// replacing, else the compiler gets confused or something.

      sprintf(ret,"%s%s",ret,   "%l4:\n%l5:");
      output_asm_insn(ret,xop);


#else
      output_asm_insn (
	  "jmpc\t[%l4]\n\tnop\n\t.align 2\n%l4:\n\t.nword %l3@h\n%l5:",
	  xop);
#endif
// ------

      ret[0] = '\0';		/* Don't need to emit anything else. */
    }

  return ret;
}
/* }}} */

/* {{{ nios_output_branch()
   
   Output an unconditional branch.  */

char *
nios_output_branch (insn, op)
    rtx insn;
    rtx *op;
{
  static char ret[256];

  if (get_attr_length (insn) == 2)
    {
      /* Short branch.  */
      sprintf (ret, "br\t%%l0%%#");
      return ret;
    }
  else
    {
      /* Long branch.  */
      rtx xop[3];

      /* Save the operand array so that we can output the original
	 label after filling the delay slot.  This is needed because
	 final_scan_insn overwrites the operand array.  Also add an operand
	 for the label needed to save the jump location.  */
      xop[0] = op[0];
      xop[1] = gen_label_rtx ();
      xop[2] = NULL;

      /* Output long jump and fill its delay slot if possible.  */
// ------
// dvb/cv 2001 Oct 23 -- conditionally generate an alternative to jmpc
// the else clause is the normal code generation as provided by cygnus

#if DONT_EMIT_JMPC

      sprintf(ret,"%s",         "\t pfx %%hi(%l0@h) \n");
      sprintf(ret,"%s%s",ret,   "\t movi %%g0,%%lo(%l0@h) \n");
      if(TARGET_32BIT)
        {
	  if (!nios_suppress_address_xhi)
	    sprintf(ret,"%s%s",ret, "\t pfx %%xhi(%l0@h) \n");

	  if (!nios_suppress_address_xlo)
	    sprintf(ret,"%s%s",ret, "\t movhi %%g0,%%xlo(%l0@h) \n");
        }
      sprintf(ret,"%s%s",ret,   "\t jmp %%g0 \n");

#else

      strcpy (ret, "jmpc\t[%l1]");

#endif
// ------
      output_asm_insn (ret, xop);

      /* If the delay slot has been filled, emit the instruction into the
	 delay slot of the conditional branch emitted above.  */
      if (dbr_sequence_length () > 0)
	{
	  rtx delay_slot_insn = NEXT_INSN (insn);
	  final_scan_insn (delay_slot_insn, asm_out_file, 1, 0, 1);

	  /* Delete the delay slot insn to avoid emitting it again.  */
	  PUT_CODE (delay_slot_insn, NOTE);
	  NOTE_LINE_NUMBER (delay_slot_insn) = NOTE_INSN_DELETED;
	  NOTE_SOURCE_FILE (delay_slot_insn) = 0;
	}
      else
	{
	  /* Delay slot cannot be filled. Output a `nop'. */
	  output_asm_insn ("nop", xop);
	}

      /* Output storage for jump location.  */
      strcpy (ret, 
	      ".align 2\n"
	      "%l1:\n\t"
	      ".nword %l0@h");
      output_asm_insn (ret, xop);

      return "";		/* Don't need to emit anything else. */
    }
}
/* }}} */

/* {{{ nios_output_2op_insn()
   
   Output various two operand instructions with operands 'reg, reg' or
   'reg, imm'. The enumeration nios_2op_opcodes contains the valid
   opcodes and the table nios_2op_opcode_names gives the names for each
   opcode and operand combination.  */

char *
nios_output_2op_insn (opcode, op, op0ix, op1ix, insn)
    enum nios_2op_opcodes opcode;
    rtx *op;
    int op0ix;	/* Index into `op' array for the first operand.  */
    int op1ix;	/* Index into `op' array for the second operand.  */
    rtx insn;
{
  static char ret[256];
  rtx op0, op1;

  if ((int)opcode >= (int)NIOS_LAST_2OP_OPCODE)
    fatal_insn ("Bad opcode to nios_output_2op_insn: ", insn);

  op0 = op[op0ix];
  op1 = op[op1ix];

  /* Register to register operation.  */
  if (REG_P (op0) && REG_P (op1))
    {
      sprintf (ret, "%s\t%%%d, %%%d", nios_2op_opcode_names[opcode][0], 
	       op0ix, op1ix);

      /* Adjust destination register if this was a sub-word operation.  */
      if (GET_MODE (op1) == QImode)
	sprintf (ret, "%s\n\text8s\t%%%d, 0", ret, op0ix);
      else if (TARGET_32BIT && GET_MODE (op1) == HImode)
	sprintf (ret, "%s\n\text16s\t%%%d, 0", ret, op0ix);

      return ret;
    }

  /* Immediate to register operation.  */
  else if (REG_P (op0) 
      && GET_CODE (op1) == CONST_INT || GET_CODE (op1) == CONST_DOUBLE)
    {
      HOST_WIDE_INT val = nios_get_value (op1);

      if (CONST_OK_FOR_L (val)
	  /* AND, OR and XOR must always be prefixed.  */
	  && opcode != NIOS_AND 
	  && opcode != NIOS_OR 
	  && opcode != NIOS_XOR)
	{
	  /* Immediate fits in 5 bits. Use directly.  */
	  sprintf (ret, "%s\t%%%d, %%%d", nios_2op_opcode_names[opcode][1], 
		   op0ix, op1ix);
	  output_asm_insn (ret, op);
	}
      else if (CONST_OK_FOR_P (val) && opcode == NIOS_MOV)
	{
	  /* Move operations using immediates that are an exact power of
	     two can be done more efficiently using 'bgen'.  */
	  sprintf (ret, "bgen\t%%%d, %d", op0ix, exact_log2 (val));
	  output_asm_insn (ret, op);
	}
      else if (CONST_OK_FOR_L (-val) && opcode == NIOS_MOV && TARGET_NEG_CONST)
	{
	  /* Negative of MOVI fits in 5 bits. */
	  sprintf (ret, "movi\t%%%d, %ld", op0ix, (long) -val);
	  output_asm_insn (ret, op);

	  sprintf (ret, "neg\t%%%d", op0ix);
	  output_asm_insn (ret, op);
	}
      else if (CONST_OK_FOR_P (~val) && opcode == NIOS_MOV && TARGET_NEG_CONST)
	{
	  /* Masks with single zero bit are inverse of exact power of 2,
	     as in 0xffff7fff. */
	  sprintf (ret, "bgen\t%%%d, %d", op0ix, exact_log2 (~val));
	  output_asm_insn (ret, op);

	  sprintf (ret, "not\t%%%d", op0ix);
	  output_asm_insn (ret, op);
	}
      else if (CONST_OK_FOR_P (-val) && opcode == NIOS_MOV && TARGET_NEG_CONST)
	{
	  /* Masks with right-justified field of zeroes are negative
	     of exact power of 2, as in 0xffffff00. */
	  sprintf (ret, "bgen\t%%%d, %d", op0ix, exact_log2 (-val));
	  output_asm_insn (ret, op);

	  sprintf (ret, "neg\t%%%d", op0ix);
	  output_asm_insn (ret, op);
	}
      else if (TARGET_32BIT && CONST_OK_FOR_N (-val) && opcode == NIOS_MOV &&
	       TARGET_NEG_CONST)
	{
	  /* For nios32, negative of immediate up to 16 bits. */
	  sprintf (ret, "pfx\t%%%%hi(%ld)", (long) -val);
	  output_asm_insn (ret, op);

	  sprintf (ret, "movi\t%%%d, %%%%lo(%ld)", op0ix, (long) -val);
	  output_asm_insn (ret, op);

	  sprintf (ret, "neg\t%%%d", op0ix);
	  output_asm_insn (ret, op);
	}
      else
	{
	  /* Immediate is wider than 5 bits or it's one of AND, OR, XOR.
	     Prefix it.  */
	  sprintf (ret, "pfx\t%%%%hi(%%%d)", op1ix);
	  output_asm_insn (ret, op);

	  sprintf (ret, "%s\t%%%d, %%%%lo(%%%d)", nios_2op_opcode_names[opcode][1], 
		   op0ix, op1ix);
	  output_asm_insn (ret, op);

	  /* For nios32 moves also move the upper 16 bits. Only do this if
	     upper 16 bits are not zero because the previous move cleared
	     the upper 16 bits of the destination register.  */
	  if (TARGET_32BIT && ! CONST_OK_FOR_N (val))
	    {
	      if (opcode == NIOS_MOV)
		{
		  if (val & 0xffe00000)
		  {
		    sprintf (ret, "pfx\t%%%%xhi(%%%d)", op1ix);
		    output_asm_insn (ret, op);
		  }

		  sprintf (ret, "movhi\t%%%d, %%%%xlo(%%%d)", op0ix, op1ix);
		  output_asm_insn (ret, op);
		}
	      else
		{
		  char msg[80];
		  sprintf (msg, "Operands to '%s' cannot be wider than 16 bits: ", 
		      nios_2op_opcode_names[opcode][0]);
		  fatal_insn (msg, insn);
		}
	    }
	}

	return "";
    }

  /* Constant to register operation. We only allow moves in this case.  */
  else if (REG_P (op[op0ix]) && CONSTANT_P (op[op1ix]))
    {
      if (opcode != NIOS_MOV)
	{
	  char msg[80];
	  sprintf (msg, "Invalid operand to '%s': ",
		   nios_2op_opcode_names[opcode][0]);
	  fatal_insn (msg, insn);
	}

      sprintf (ret, "pfx\t%%%%hi(%%%d)\n\t%s\t%%%d, %%%%lo(%%%d)", 
	       op1ix, nios_2op_opcode_names[opcode][1], op0ix, op1ix);
      output_asm_insn (ret, op);

      /* For nios32 moves also move the upper 16 bits of the constant.  */
      if (TARGET_32BIT)
	{
	  if (!nios_suppress_address_xhi)
	    {
	      sprintf (ret, "pfx\t%%%%xhi(%%%d)", op1ix);
	      output_asm_insn (ret, op);
	    }

	  if (!nios_suppress_address_xlo)
	    {
	      sprintf (ret, "movhi\t%%%d, %%%%xlo(%%%d)", op0ix, op1ix);
	      output_asm_insn (ret, op);
	    }
	}

      return "";
    }

  fatal_insn ("Bad operands to nios_output_2op_insn: ", insn);
}
/* }}} */

/* {{{ nios_output_memory_access()
   
   Output memory access operations.  */

char *
nios_output_memory_access (op, insn)
    rtx *op;
    rtx insn;
{
  rtx op0 = op[0];
  rtx op1 = op[1];
  rtx mem_rtx, base_reg, offset_rtx;
  int byte_offset, halfword_offset, word_offset, offset, subreg_offset;
  rtx xop[2];
  char tmp_str[128];

  /* Sanity check. We only deal with reg<->mem moves here.  */
  if (! ((GET_CODE (op0) == MEM && REG_P (op1))
         || (REG_P (op0) && GET_CODE (op1) == MEM)))
    fatal_insn ("Bad operands to nios_output_memory_access:", insn);

  /* Determine base address and offset of the memory operand.  */
  mem_rtx = (GET_CODE (op0) == MEM) ? op0 : op1;
  if (GET_CODE (XEXP (mem_rtx, 0)) == PLUS)
    {
      base_reg = XEXP (XEXP (mem_rtx, 0), 0);
      offset_rtx = XEXP (XEXP (mem_rtx, 0), 1);
      offset = INTVAL (offset_rtx);
    }
  else
    {
      base_reg = XEXP (mem_rtx, 0);
      offset_rtx = NULL_RTX;
      offset = 0;
    }

  /* Determine if the base register is a subreg.  */
  subreg_offset = 0;
  while (GET_CODE (base_reg) == SUBREG)
    {
      subreg_offset += SUBREG_WORD (base_reg);
      base_reg = SUBREG_REG (base_reg);
    }

  /* Offsets are always given in words. We only allow unaligned offsets
     when loading subwords with %fp or %sp which we already know to be
     aligned.  */
  if (offset % UNITS_PER_WORD != 0
      && REGNO (base_reg) != GPR_FP
      && REGNO (base_reg) != GPR_SP)
    fatal_insn ("Offsets for memory references must be multiple of word size:",
		insn);

  word_offset = offset / UNITS_PER_WORD;

  /* Half-word and byte offsets needed when using %fp or %sp to access
     subwords in memory.  */
  if ((REGNO (base_reg) == GPR_FP || REGNO (base_reg) == GPR_SP)
      && GET_MODE_SIZE (GET_MODE (mem_rtx)) < UNITS_PER_WORD)
    {
      halfword_offset = (offset % UNITS_PER_WORD) / 2;
      byte_offset = offset % UNITS_PER_WORD;

      /* Adjust for negative offsets. Compute a positive offset from the
	 previous half-word.  */
      if (byte_offset < 0 || halfword_offset < 0)
	{
	  word_offset--;
	  if (byte_offset < 0)
	    byte_offset += UNITS_PER_WORD;

	  if (halfword_offset < 0)
	    halfword_offset += (UNITS_PER_WORD / 2);
	}
    }
  else
    halfword_offset = byte_offset = 0;


  if (REG_P (op0))
    {
      rtx scratch_dest = NULL;

      /* Register loads. If loading a sub-word from memory with
	 registers other than %fp and %sp, the base register and
	 destination register must be different. Otherwise, if we need
	 to use ext8d or ext16d to extract the sub-word, the load
	 operation will have trashed the base register. 

	 We're trying to avoid the following situation:

	 	ldp %l0, [%l0, 3]	; This trashes l0
		ext8d %l0, %l0		;   and extracts the wrong byte.

	 We use %r0 as a scratch register and convert the above to:

	 	mov %r0, %l0
		ldp %r0, [%l0, 3]
		ext8d %r0, %l0
		mov %l0, %r0

	 Not the nicest piece of code, but it does the trick. 

	 FIXME: We should find a better way for dealing with this.  */

      if (REGNO (base_reg) != GPR_FP 
	  && REGNO (base_reg) != GPR_SP
	  && GET_MODE_SIZE (GET_MODE (mem_rtx)) < UNITS_PER_WORD
	  && REGNO (op0) == REGNO (base_reg))
	{
	  scratch_dest = gen_rtx_REG (word_mode, GPR_R0);
	  xop[0] = scratch_dest;
	  xop[1] = op[0];
	  output_asm_insn ("mov\t%0, %1", xop);
	}


      /* Load a full word from memory and then do the appropriate
	 extractions if needed.  */
      if ((REGNO_OK_FOR_BASE_P (REGNO (base_reg))) &&
          (TARGET_FEWER_OPCODES == 0))    /* Avoid LDP ...*/
	{
	  /* For registers %l0 - %l3 we can use ldp.  Load to the
	     scratch register if it has been defined (see above).  */
	  if (CONST_OK_FOR_L (word_offset))
	    {
	      xop[0] = (scratch_dest) ? scratch_dest : op[0];
	      xop[1] = op[1];
	      if (TARGET_DCACHE && MEM_VOLATILE_P(mem_rtx))
                {
                  /* Ordinarily no PFX needed, but need to force     */
                  /* non-cached access with PFXIO, so use "PFXIO 0". */
                  output_asm_insn ("pfxio\t0", xop);
                }
	      output_asm_insn ("ldp\t%0, %M1", xop);
	    }
	  else
	    {
	      xop[0] = gen_rtx_CONST_INT (word_mode, word_offset);

	      if (TARGET_DCACHE && MEM_VOLATILE_P(mem_rtx))
                {
                  /* Force non-cached access with PFXIO instead of PFX. */
                  output_asm_insn ("pfxio\t%%hi(%0)", xop);
                }
              else
                {
                  output_asm_insn ("pfx\t%%hi(%0)", xop);
                }

	      xop[0] = (scratch_dest) ? scratch_dest : op[0];
	      xop[1] = op[1];
	      output_asm_insn ("ldp\t%0, %M1", xop);
	    }
	}
     else
	{
	  /* Otherwise, we use `pfx' to specify the word offset. The offset
	     range is more restricted in this case. If the base register is
	     %sp, we use `lds' when the offset is in the range [0, 255].  */

          /* Since LDS doesn't use PFX, we cannot force an uncached access
             by using PFXIO.  So don't use LDS for uncached accesses. */

	  if ((REGNO (base_reg) == GPR_SP)
	      && (CONST_OK_FOR_K (word_offset))
              && (TARGET_FEWER_OPCODES == 0)  /* Avoid LDS also... */
              && (!(TARGET_DCACHE && MEM_VOLATILE_P(mem_rtx))))
	    {
	      /* Almost unnecessary sanity check. scratch_dest should
		 never be set if base_reg is %sp.  */
	      if (scratch_dest)
		abort ();

	      xop[0] = op[0];
	      xop[1] = gen_rtx_CONST_INT (word_mode, word_offset);
	      output_asm_insn ("lds\t%0, [%%sp, %1]", xop);
	    }
	  else
	    {
	      if (! CONST_OK_FOR_U (word_offset))
		fatal_insn ("Offset for `ld' must be between -1024 and 1023:",
			    insn);

	      if (word_offset != 0)
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, word_offset);
                  if (TARGET_DCACHE && MEM_VOLATILE_P(mem_rtx))
                    {
                      /* Force non-cached access with PFXIO instead of PFX. */
                      output_asm_insn ("pfxio\t%0", xop);
                    }
                  else
                    {
                      output_asm_insn ("pfx\t%0", xop);
                    }
		}
	      else if (TARGET_DCACHE && MEM_VOLATILE_P(mem_rtx))
                {
                  /* Ordinarily no PFX needed, but need to force     */
                  /* non-cached access with PFXIO, so use "PFXIO 0". */
                  output_asm_insn ("pfxio\t0", xop);
                }

	      xop[0] = (scratch_dest) ? scratch_dest : op[0];
	      xop[1] = base_reg;
	      output_asm_insn ("ld\t%0, [%1]", xop);
	    }
	}

      /* Now do the appropriate extractions for sub-words. Deal with the
	 special case when %fp or %sp are the base register. In this case
	 we use ext8s or ext16s.  */
      if (GET_MODE (mem_rtx) == QImode)
	{
	  if (REGNO (base_reg) == GPR_FP
	      || REGNO (base_reg) == GPR_SP)
	    {
	      xop[0] = op[0];
	      xop[1] = gen_rtx_CONST_INT (word_mode, byte_offset);
	      output_asm_insn ("ext8s\t%0, %1", xop);
	    }
	  else
	    {
	      xop[0] = (scratch_dest) ? scratch_dest : op[0];
	      xop[1] = base_reg;
	      output_asm_insn ("ext8d\t%0, %1", xop);

	      /* If we used %r0 as a scratch register, move it back to
		 the original destination.  */
	      if (scratch_dest)
		{
		  xop[0] = op[0];
		  xop[1] = scratch_dest;
		  output_asm_insn ("mov\t%0, %1", xop);
		}
	    }
	}

      else if (TARGET_32BIT 
	       && GET_MODE (mem_rtx) == HImode)
	{
	  if (REGNO (base_reg) == GPR_FP
	      || REGNO (base_reg) == GPR_SP)
	    {
	      xop[0] = op[0];
	      xop[1] = gen_rtx_CONST_INT (word_mode, halfword_offset);
	      output_asm_insn ("ext16s\t%0, %1", xop);
	    }
	  else
	    {
	      xop[0] = (scratch_dest) ? scratch_dest : op[0];
	      xop[1] = base_reg;
	      output_asm_insn ("ext16d\t%0, %1", xop);

	      /* If we used %r0 as a scratch register, move it back to
		 the original destination.  */
	      if (scratch_dest)
		{
		  xop[0] = op[0];
		  xop[1] = scratch_dest;
		  output_asm_insn ("mov\t%0, %1", xop);
		}
	    }
	}
    }
  else
    {
      /* Register stores. Stores using one of the base pointer registers
	 %l0 - %l3 can use `stp' only for full-word stores.  */
      if (((REGNO_OK_FOR_BASE_P (REGNO (base_reg))
	  && GET_MODE_SIZE (GET_MODE (mem_rtx)) >= UNITS_PER_WORD))
	  && (TARGET_FEWER_OPCODES == 0))
	{
	  if (! CONST_OK_FOR_L (word_offset))
	    {
	      xop[0] = gen_rtx_CONST_INT (word_mode, word_offset);
	      output_asm_insn ("pfx\t%%hi(%0)", xop);
	    }

	  output_asm_insn ("stp\t%M0, %1", op);
	}
      else
	{
	  /* Use `pfx' to specify the offset, if needed.  */
	  char prefix[80];

	  if (! CONST_OK_FOR_U (word_offset))
	    fatal_insn ("Offset for st/st8d/st16d must be between -1024 "
			"and 1023:", insn);

	  /* The decision to use the prefix is a bit more elaborate now.
	     When storing to an unaligned offset with %sp, we have several
	     choices:

	     1- Byte stores can use 10-bit byte offsets without a prefix.
	     2- Half-word stores can use 9-bit half-word offsets w/o a prefix.
	     3- Word stores we can use 8-bit word offsets without a prefix.

	     In any other case, we must issue a prefix instruction. Notice
	     that the byte and half-word offsets are computed using the
	     actual offset given as argument, not the byte_offset and
	     halfword_offset computed above. This is because byte_offset
	     and halfword_offset are computed relative to the word_offset.
	     Here we need absolute byte and half-word offsets.  */

	  if ((word_offset == 0)
	      || ((TARGET_FEWER_OPCODES == 0)
	         && (REGNO (base_reg) == GPR_SP
		  && (CONST_OK_FOR_R (offset)
		      || CONST_OK_FOR_Q (offset / 2)
		      || CONST_OK_FOR_K (word_offset)))))
	    prefix[0] = '\0';
	  else
	    sprintf (prefix, "pfx\t%d", word_offset);

	  if (GET_MODE (mem_rtx) == QImode)
	    {
	      /* Byte stores. If storing with %fp we need to use st8s when
		 using unaligned offsets.  If storing with %sp, we use
		 sts8s if the original offset fits in 10 bits.  */
	      xop[0] = gen_rtx_REG (word_mode, GPR_R0);
	      xop[1] = op[1];
	      output_asm_insn ("fill8\t%0, %1", xop);
	      output_asm_insn (prefix, NULL);

	      if ((REGNO (base_reg) == GPR_SP && CONST_OK_FOR_R (offset))
	          && (TARGET_FEWER_OPCODES == 0))
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, offset);
		  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
		  output_asm_insn ("sts8s\t[%%sp, %0], %1", xop);
		}
	      else if ((REGNO (base_reg) == GPR_FP && byte_offset != 0)
	              && (TARGET_FEWER_OPCODES == 0))
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, byte_offset);
                  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
                  sprintf (tmp_str, "st8s\t[%%%s], %%1, %%0", reg_names[GPR_FP]);
		  output_asm_insn (tmp_str, xop);
		}
	      else
		{
		  xop[0] = base_reg;
		  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
		  output_asm_insn ("st8d\t[%0], %1", xop);
		}
	    }
	  else if (TARGET_32BIT
	           && GET_MODE (mem_rtx) == HImode)
	    {
	      /* Half-word stores for nios-32. If storing with %fp we need
		 to use st16s when using unaligned offsets. If storing with
		 %sp, we can use sts16s without a prefix if the original
		 offset expressed in half-words fits in 9 bits.  */
	      xop[0] = gen_rtx_REG (word_mode, GPR_R0);
	      xop[1] = op[1];
	      output_asm_insn ("fill16\t%0, %1", xop);
	      output_asm_insn (prefix, NULL);

	      if ((REGNO (base_reg) == GPR_SP
		  && CONST_OK_FOR_Q (offset / 2))
	          && (TARGET_FEWER_OPCODES == 0))
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, offset / 2);
		  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
		  output_asm_insn ("sts16s\t[%%sp, %0], %1", xop);
		}
	      else if ((REGNO (base_reg) == GPR_FP
		       && halfword_offset != 0)
	          && (TARGET_FEWER_OPCODES == 0))
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, halfword_offset);
                  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
                  sprintf (tmp_str, "st16s\t[%%%s], %%1, %%0", reg_names[GPR_FP]);
		  output_asm_insn (tmp_str, xop);
		}
	      else
		{
		  xop[0] = base_reg;
		  xop[1] = gen_rtx_REG (word_mode, GPR_R0);
		  output_asm_insn ("st16d\t[%0], %1", xop);
		}
	    }
	  else
	    {
	      /* Full-word stores. If the base register is %sp and the
		 offset is in the range [0, 255] we can use `sts' instead
		 of the pfx/st combination.  */
	      if ((REGNO (base_reg) == GPR_SP
		  && CONST_OK_FOR_K (word_offset))
	          && (TARGET_FEWER_OPCODES == 0))
		{
		  xop[0] = gen_rtx_CONST_INT (word_mode, word_offset);
		  xop[1] = op[1];
		  output_asm_insn ("sts\t[%%sp, %0], %1", xop);
		}
	      else
		{
		  output_asm_insn (prefix, NULL);
		  xop[0] = base_reg;
		  xop[1] = op[1];
		  output_asm_insn ("st\t[%0], %1", xop);
		}
	    }
	}
    }


  return "";
}
/* }}} */

/* {{{ nios_output_move()
   
   Output move operations.  */

char *
nios_output_move (op, insn)
    rtx *op;
    rtx insn;
{
  static char ret[512];

  /* Register to register move.  */
  if (REG_P (op[0]) && REG_P (op[1]))
    return nios_output_2op_insn (NIOS_MOV, op, 0, 1, insn);

  /* Immediate to register move.  */
  else if (REG_P (op[0])
	  && (GET_CODE (op[1]) == CONST_INT 
	      || GET_CODE (op[1]) == CONST_DOUBLE))
      return nios_output_2op_insn (NIOS_MOV, op, 0, 1, insn);

  /* Memory to register moves.  */
  if ((GET_CODE (op[0]) == MEM && REG_P (op[1]))
      || (REG_P (op[0]) && GET_CODE (op[1]) == MEM))
    return nios_output_memory_access (op, insn);

  /* Constant to register move.  */
  else if (REG_P (op[0]) && CONSTANT_P (op[1]))
    return nios_output_2op_insn (NIOS_MOV, op, 0, 1, insn);

  /* Unrecognized operands.  */
  else
    fatal_insn ("nios_output_move:", insn);


  return ret;
}
/* }}} */

/* {{{ nios_output_user_opcode()
   
   Output user opcodes.
   All operations supported by this function use op[0] and op[2]. */

char *
nios_output_user_opcode (opcode, op, insn)
    enum nios_user_opcodes opcode;
    rtx *op;	/* operand array */
    rtx insn;
{
  static char ret[256];
  int starting_bit, width, pfx;

  if ((int)opcode >= (int)NIOS_NUM_USER_OPCODES)
    fatal_insn ("Bad opcode to nios_output_user_opcode: ", insn);

  switch (opcode)
    {
    case NIOS_USER_OPCODE_MUL:
    case NIOS_USER_OPCODE_DIV:
    case NIOS_USER_OPCODE_UDIV:
    case NIOS_USER_OPCODE_MOD:
    case NIOS_USER_OPCODE_UMOD:
    case NIOS_USER_OPCODE_FADD:
    case NIOS_USER_OPCODE_FSUB:
    case NIOS_USER_OPCODE_FMUL:
    case NIOS_USER_OPCODE_SMIN:
    case NIOS_USER_OPCODE_UMIN:
    case NIOS_USER_OPCODE_SMAX:
    case NIOS_USER_OPCODE_UMAX:
      /* Register to register operation using op[0] and op[2]. */
      if (REG_P (op[0]) && REG_P (op[2]))
	{
	  if (nios_user_opcodes[opcode].usr == 0)
	    {
	      if (nios_user_opcodes[opcode].pfx != 0)
		{
		  sprintf (ret, "pfx\t%d", nios_user_opcodes[opcode].pfx);
		  output_asm_insn (ret, op);
		}

	      sprintf (ret, "usr0\t%%0, %%2");
	      output_asm_insn (ret, op);
	    }

	  else
	    {
	      sprintf (ret, "mov\t%%%%r0, %%2");
	      output_asm_insn (ret, op);

	      if (nios_user_opcodes[opcode].pfx != 0)
		{
		  sprintf (ret, "pfx\t%d", nios_user_opcodes[opcode].pfx);
		  output_asm_insn (ret, op);
		}

	      sprintf (ret, "usr%d\t%%0", nios_user_opcodes[opcode].usr);
	      output_asm_insn (ret, op);
	    }

	  return "";
	}
      break;

    case NIOS_USER_OPCODE_FFS:
      /* Register to register operation using op[0] and op[1]. */
      if (REG_P (op[0]) && REG_P (op[1]))
	{
	  if (nios_user_opcodes[opcode].usr == 0)
	    {
	      if (nios_user_opcodes[opcode].pfx != 0)
		{
		  sprintf (ret, "pfx\t%d", nios_user_opcodes[opcode].pfx);
		  output_asm_insn (ret, op);
		}

	      sprintf (ret, "usr0\t%%0, %%1");
	      output_asm_insn (ret, op);
	    }

	  else
	    {
	      sprintf (ret, "mov\t%%%%r0, %%1");
	      output_asm_insn (ret, op);

	      if (nios_user_opcodes[opcode].pfx != 0)
		{
		  sprintf (ret, "pfx\t%d", nios_user_opcodes[opcode].pfx);
		  output_asm_insn (ret, op);
		}

	      sprintf (ret, "usr%d\t%%0", nios_user_opcodes[opcode].usr);
	      output_asm_insn (ret, op);
	    }

	  return "";
	}
      break;

    case NIOS_USER_OPCODE_EXTV:
    case NIOS_USER_OPCODE_EXTZV:
      /* Register to register operation using op[0] and op[1]. */
      /* op[2] is bit-field width, and op[3] is starting bit. */
      if (REG_P (op[0]) && REG_P (op[1]) &&
	  (GET_CODE (op[2]) == CONST_INT || GET_CODE (op[2]) == CONST_DOUBLE) &&
	  (GET_CODE (op[3]) == CONST_INT || GET_CODE (op[3]) == CONST_DOUBLE))
	{
	  width = nios_get_value (op[2]);
	  starting_bit = nios_get_value (op[3]);

	  if (CONST_OK_FOR_L (width) && CONST_OK_FOR_L (starting_bit))
	    {
	      pfx = starting_bit | (width << 5);

	      if (opcode == NIOS_USER_OPCODE_EXTV)
		pfx |= (1 << 10); /* specify sign extension in pfx so that */
				  /* same opcode can be used for extv & extzv */

	      if (nios_user_opcodes[opcode].usr == 0)
		{
		  sprintf (ret, "pfx\t%d", pfx);
		  output_asm_insn (ret, op);

		  sprintf (ret, "usr0\t%%0, %%1");
		  output_asm_insn (ret, op);
		}

	      else
		{
		  sprintf (ret, "mov\t%%%%r0, %%1");
		  output_asm_insn (ret, op);

		  sprintf (ret, "pfx\t%d", pfx);
		  output_asm_insn (ret, op);

		  sprintf (ret, "usr%d\t%%0", nios_user_opcodes[opcode].usr);
		  output_asm_insn (ret, op);
		}

	      return "";
	    }
	}
      break;

    case NIOS_USER_OPCODE_INSV:
      /* Register to register operation using op[0] and op[3]. */
      /* op[1] is bit-field width, and op[2] is starting bit. */
      if (REG_P (op[0]) && REG_P (op[3]) &&
	  (GET_CODE (op[1]) == CONST_INT || GET_CODE (op[1]) == CONST_DOUBLE) &&
	  (GET_CODE (op[2]) == CONST_INT || GET_CODE (op[2]) == CONST_DOUBLE))
	{
	  width = nios_get_value (op[1]);
	  starting_bit = nios_get_value (op[2]);

	  if (CONST_OK_FOR_L (width) && CONST_OK_FOR_L (starting_bit))
	    {
	      pfx = starting_bit | (width << 5);

	      if (nios_user_opcodes[opcode].usr == 0)
		{
		  sprintf (ret, "pfx\t%d", pfx);
		  output_asm_insn (ret, op);

		  sprintf (ret, "usr0\t%%0, %%3");
		  output_asm_insn (ret, op);
		}

	      else
		{
		  sprintf (ret, "mov\t%%%%r0, %%3");
		  output_asm_insn (ret, op);

		  sprintf (ret, "pfx\t%d", pfx);
		  output_asm_insn (ret, op);

		  sprintf (ret, "usr%d\t%%0", nios_user_opcodes[opcode].usr);
		  output_asm_insn (ret, op);
		}
	    }

	  return "";
	}
      break;
    }

  fatal_insn ("Bad operands to nios_output_user_opcode: ", insn);
}

/* }}} */

/* {{{ nios_print_int()
   
   Output the appropriate code to write to file STREAM an integer
   constant or address rtx VALUE that is NBYTES wide.  */

void
nios_print_int (stream, value, nbytes)
     FILE *stream;
     rtx value;
     int nbytes;
{
  int is_fnptr = (GET_CODE (value) == SYMBOL_REF && SYMBOL_REF_FLAG (value));

  switch (nbytes)
    {
    case 1:
      fputs ("\t.byte\t", stream);
      break;
  
    case 2:
      if (is_fnptr && !TARGET_32BIT)
	fputs ("\t.nword\t", stream);
      else
	fputs ("\t.uashort\t", stream);
      break;

    case 4:
      if (is_fnptr && TARGET_32BIT)
	fputs ("\t.nword\t", stream);
      else
	fputs ("\t.uaword\t", stream);
      break;

    default:
      fatal_insn ("Bad value to nios_print_int", value);
    }

  output_addr_const (stream, value);

  /* Mark functions with @h.  */
  if (is_fnptr)
    fprintf (stream, "@h");

  fputc ('\n', stream);
}
/* }}} */

/* {{{ nios_print_operand_address()
   
   Print a memory address as an operand to reference that memory location.  */

void
nios_print_operand_address (stream, x)
     FILE *stream;
     rtx x;
{
  int offset;

  if (GET_CODE (x) == MEM)
    x = XEXP (x, 0);

  switch (GET_CODE (x))
    {
    case SUBREG:
      offset = 0;
      do {
	offset += SUBREG_WORD (x);
	x = SUBREG_REG (x);
      } while (GET_CODE (x) == SUBREG);
      fputs (reg_names[ REGNO (x) + offset ], stream);
      return;

    case REG:
      fputs (reg_names[ REGNO (x) ], stream);
      return;

    case CONST_INT:
      fprintf (stream, "%d", SIGN_EXTEND_SHORT (INTVAL (x)));
      return;

    case SYMBOL_REF:
    case CONST:
    case CODE_LABEL:
    case LABEL_REF:
      nios_print_addr_const (stream, x);
      return;

    default:
      break;
    }

  fatal_insn ("Bad insn to nios_print_operand_address:", x);
}
/* }}} */

/* {{{ nios_print_operand()
   
   A C compound statement to output to stdio stream STREAM the assembler
   syntax for an instruction operand X.  X is an RTL expression.

   LETTER is a value that can be used to specify one of several ways of
   printing the operand.  It is used when identical operands must be
   printed differently depending on the context.  LETTER comes from the
   `%' specification that was used to request printing of the operand.
   If the specification was just `%DIGIT' then LETTER is 0; if the
   specification was `%LTR DIGIT' then LETTER is the ASCII code for LTR.

   If X is a register, this macro should print the register's name. The
   names can be found in an array `reg_names' whose type is `char *[]'.
   `reg_names' is initialized from `REGISTER_NAMES'.

   When the machine description has a specification `%PUNCT' (a `%'
   followed by a punctuation character), this macro is called with a
   null pointer for X and the punctuation character for LETTER.  */

void
nios_print_operand (stream, x, letter)
   FILE *stream;
   rtx x;
   int letter;
{
  enum rtx_code code = (x) ? GET_CODE (x) : NIL;

  switch (letter)
    {
      case '#':
	/* Emit `nop' if the delay slot cannot be filled.  */
	if (dbr_sequence_length () == 0)
	  fputs ("\n\tnop", stream);
	break;

      case 'F':
	/* Operand is a local function name used in a `bsrr' instruction,
	   output the name without the @h suffix. We just assemble the name
	   without calling nios_print_operand_address().  */
	assemble_name (stream, XSTR (x, 0));
	break;

      case 'b':
      case 'B':
	/* These are normal ('B') and reversed ('b') branches.  */
	switch (letter == 'B' ? GET_CODE (x) : reverse_condition (GET_CODE (x)))
	  {
	  case NE:
	    fprintf (stream, "ne");
	    break;
	  case EQ:
	    fprintf (stream, "eq");
	    break;
	  case GE:
	    fprintf (stream, "ge");
	    break;
	  case GT:
	    fprintf (stream, "gt");
	    break;
	  case LE:
	    fprintf (stream, "le");
	    break;
	  case LT:
	    fprintf (stream, "lt");
	    break;
	  case GEU:
	    fprintf (stream, "cc");
	    break;
	  case GTU:
	    fprintf (stream, "hi");
	    break;
	  case LEU:
	    fprintf (stream, "ls");
	    break;
	  case LTU:
	    fprintf (stream, "cs");
	    break;
	  default:
	    abort ();
	  }
	break;

      /* Memory reference.  */
      case 'M':
	if (code != MEM)
	  fatal_insn ("Bad insn to nios_print_operand 'M' modifier:", x);

	nios_print_operand_memory_reference (stream, XEXP (x, 0));
	break;

      /* 1's complement of integer.  */
      case 'N':
	if (code != CONST_INT && code != CONST_DOUBLE)
	  fatal_insn ("Non-constant for nios_print_operand 'N' modifier:", x);

	else
	  {
	    HOST_WIDE_INT value = ~nios_get_value (x);
	    fprintf (stream, "%ld", (long) value);
	  }
	break;

      /* These are the least significant word in a 32bit value.
	 'o' allows us to sign extend a constant if doing so
	 makes for more compact code.  */
      case 'L':
      case 'o':
	switch (GET_CODE (x))
	  {
	  case MEM:
	    fputc ('(', stream);
	    output_address (XEXP (x, 0));
	    fputc (')', stream);
	    break;

	  case REG:
	    fprintf (stream, "%s", reg_names[REGNO (x)]);
	    break;

	  case SUBREG:
	    fprintf (stream, "%s",
		     reg_names[REGNO (SUBREG_REG (x)) + SUBREG_WORD (x)]);
	    break;

	  case CONST_DOUBLE:
	    if (code == 'L')
	      {
		HOST_WIDE_INT val = nios_get_value (x);
		nios_print_operand_address (stream, GEN_INT (val & 0xffff));
	      }
	    else
	      {
		HOST_WIDE_INT val = nios_get_value (x);

		val &= 0xffff;
		val = (((val) & 0xffff) ^ (~0x7fff)) + 0x8000;
		nios_print_operand_address (stream, GEN_INT (val));
	      }
	    break;

	  case CONST_INT:
	    if (code == 'L')
	      nios_print_operand_address (stream, GEN_INT ((INTVAL (x) & 0xffff)));
	    else
	      {
		unsigned int val = INTVAL (x) & 0xffff;
		val = (((val) & 0xffff) ^ (~0x7fff)) + 0x8000;
		nios_print_operand_address (stream, GEN_INT (val));
	      }
	    break;
	  default:
	    abort ();
	  }
	break;

      /* Similarly, but for the most significant word.  */
      case 'H':
      case 'h':
	switch (GET_CODE (x))
	  {
	  case MEM:
	    fputc ('(', stream);
	    x = adj_offsettable_operand (x, 2);
	    output_address (XEXP (x, 0));
	    fputc (')', stream);
	    break;

	  case REG:
	    fprintf (stream, "%s", reg_names[REGNO (x) + 1]);
	    break;

	  case SUBREG:
	    fprintf (stream, "%s",
		     reg_names[REGNO (SUBREG_REG (x)) + SUBREG_WORD (x)] + 1);
	    break;

	  case CONST_DOUBLE:
	    if (code == 'H')
	      {
		HOST_WIDE_INT val = nios_get_value (x);
		nios_print_operand_address (stream, GEN_INT ((val >> 16) & 0xffff));
	      }
	    else
	      {
		HOST_WIDE_INT val = nios_get_value (x);
		val = (val >> 16) & 0xffff;
		val = (((val) & 0xffff) ^ (~0x7fff)) + 0x8000;

		nios_print_operand_address (stream, GEN_INT (val));
	      }
	    break;

	  case CONST_INT:
	    if (code == 'H')
	      nios_print_operand_address (stream,
				     GEN_INT ((INTVAL (x) >> 16) & 0xffff));
	    else
	      {
		unsigned int val = (INTVAL (x) >> 16) & 0xffff;
		val = (((val) & 0xffff) ^ (~0x7fff)) + 0x8000;

		nios_print_operand_address (stream, GEN_INT (val));
	      }
	    break;
	  default:
	    abort ();
	  }
	break;


      case 0:
	if (code == REG)
	  fputs (reg_names[ REGNO (x) ], stream);

	else if (code == CONST_INT || code == CONST_DOUBLE)
	  {
	    HOST_WIDE_INT value = nios_get_value (x);
	    fprintf (stream, "%ld", (long) value);
	  }

	else if (code == MEM)
	  nios_print_operand_address (stream, XEXP (x, 0));

	else if (CONSTANT_P (x))
	  nios_print_operand_address (stream, x);

	else
	  fatal_insn ("Bad insn in print_operand, 0 case", x);

	return;

      default:
	{
	  char buf[80];

	  sprintf (buf, "Invalid asm template character '%%%c'", letter);
	  fatal_insn (buf, x);
	}
    }
}
/* }}} */

/* {{{ nios_print_addr_const()
   
   Print an integer constant expression in assembler syntax.  */

static void
nios_print_addr_const (stream, x)
     FILE *stream;
     rtx x;
{
  char buf[256];

 restart:
  switch (GET_CODE (x))
    {
    case SYMBOL_REF:
      assemble_name (stream, XSTR (x, 0));
      /* Append @h to symbols representing user functions or libcalls.  */
      if (SYMBOL_REF_FLAG (x) || SYMBOL_REF_USED (x))
	fprintf (stream, "@h");
      break;

    case LABEL_REF:
      ASM_GENERATE_INTERNAL_LABEL (buf, "L", CODE_LABEL_NUMBER (XEXP (x, 0)));
      assemble_name (stream, buf);
      break;

    case CODE_LABEL:
      ASM_GENERATE_INTERNAL_LABEL (buf, "L", CODE_LABEL_NUMBER (x));
      assemble_name (stream, buf);
      break;

    case CONST_INT:
      fprintf (stream, "%d", INTVAL (x));
      break;

    case CONST:
      nios_print_addr_const (stream, XEXP (x, 0));
      break;

    case PLUS:
      /* Some assemblers need integer constants to appear last (eg masm).  */
      if (GET_CODE (XEXP (x, 0)) == CONST_INT)
	{
	  nios_print_addr_const (stream, XEXP (x, 1));
	  if (INTVAL (XEXP (x, 0)) >= 0)
	    fprintf (stream, "+");
	  nios_print_addr_const (stream, XEXP (x, 0));
	}
      else
	{
	  nios_print_addr_const (stream, XEXP (x, 0));
	  if (INTVAL (XEXP (x, 1)) >= 0)
	    fprintf (stream, "+");
	  nios_print_addr_const (stream, XEXP (x, 1));
	}
      break;

    case MINUS:
      /* Avoid outputting things like x-x or x+5-x,
	 since some assemblers can't handle that.  */
      x = simplify_subtraction (x);
      if (GET_CODE (x) != MINUS)
	goto restart;

      nios_print_addr_const (stream, XEXP (x, 0));
      fprintf (stream, "-");
      if (GET_CODE (XEXP (x, 1)) == CONST_INT
	  && INTVAL (XEXP (x, 1)) < 0)
	{
	  fprintf (stream, ASM_OPEN_PAREN);
	  nios_print_addr_const (stream, XEXP (x, 1));
	  fprintf (stream, ASM_CLOSE_PAREN);
	}
      else
	nios_print_addr_const (stream, XEXP (x, 1));
      break;

    case ZERO_EXTEND:
    case SIGN_EXTEND:
      nios_print_addr_const (stream, XEXP (x, 0));
      break;

    default:
      output_addr_const (stream, x);
      break;
    }
}
/* }}} */

/* {{{ nios_print_operand_memory_reference()
   
   Print a memory reference suitable for the ldp/stp instructions.  */

static void
nios_print_operand_memory_reference (stream, x)
     FILE *stream;
     rtx x;
{
  rtx x0, x1;
  int offset;

  switch (GET_CODE (x))
    {
    case SUBREG:
      offset = 0;
      do {
	offset += SUBREG_WORD (x);
	x = SUBREG_REG (x);
      } while (GET_CODE (x) == SUBREG);
      fprintf (stream, "[%s, 0]", reg_names[ REGNO (x) + offset ]);
      break;

    case REG:
      fprintf (stream, "[%s, 0]", reg_names[ REGNO (x) ]);
      break;

    case PLUS:
      x0 = XEXP (x, 0);
      x1 = XEXP (x, 1);
      offset = 0;

      while (GET_CODE (x0) == SUBREG)
	{
	  offset += SUBREG_WORD (x0);
	  x0 = SUBREG_REG (x0);
	}

      if (GET_CODE (x0) == REG
	  && BP_REG_P (REGNO (x0))
	  && GET_CODE (x1) == CONST_INT)
	{
	  int value = nios_get_value (x1);

	  value /= UNITS_PER_WORD;

	  fprintf (stream, "[%s, ", reg_names[REGNO (x0) + offset]);

	  if (! CONST_OK_FOR_L (value))
	    fprintf(stream, "%%lo(%d)]", value);
	  else
	    fprintf(stream, "%d]", value);
	}
      else
	fatal_insn ("Bad base register in nios_print_operand_memory_reference:", x);
      break;

    default:
      fatal_insn ("Bad insn to nios_print_operand_memory_reference:", x);
      break;
    }
}
/* }}} */

/* }}} */

/* {{{ Various helper functions.  */

/* {{{ nios_split_constant32()
   
   Split a 32-bit constant into 2 16-bit values. The nios is
   little-endian, so the little word goes first.  */

static void
nios_split_constant32 (x, values)
     rtx x;
     int values[];
{
  enum rtx_code code = GET_CODE (x);
  HOST_WIDE_INT v;

  if (code == CONST_INT)
    {
      v = nios_get_value (x);
      values[0] = SIGN_EXTEND_SHORT (v);
      values[1] = SIGN_EXTEND_SHORT (v >> 16);
    }

  else if (code == CONST_DOUBLE && GET_MODE (x) == SFmode)
    {
      REAL_VALUE_TYPE rv;

      REAL_VALUE_FROM_CONST_DOUBLE (rv, x);
      REAL_VALUE_TO_TARGET_SINGLE (rv, v);
      values[0] = SIGN_EXTEND_SHORT (v);
      values[1] = SIGN_EXTEND_SHORT (v >> 16);
    }

  else
    fatal_insn ("Bad rtx passed to nios_split_constant32:", x);
}
/* }}} */

/* {{{ nios_split_constant64()
   
   Split a 64-bit constant into 4 16-bit values.  The nios is
   little-endian, so the little word goes first.  */

static void
nios_split_constant64 (x, values)
     rtx x;
     int values[];
{
  enum rtx_code code = GET_CODE (x);
  HOST_WIDE_INT v;

  if (code == CONST_INT)
    {
      v = nios_get_value (x);
      if (v < 0)
	values[2] = values[3] = -1;
      else
	values[2] = values[3] = 0;

      values[0] = SIGN_EXTEND_SHORT (v);
      values[1] = SIGN_EXTEND_SHORT (v >> 16);
    }

  else if (code == CONST_DOUBLE && GET_MODE (x) == DFmode)
    {
      long t[2];
      REAL_VALUE_TYPE rv;

      REAL_VALUE_FROM_CONST_DOUBLE (rv, x);
      REAL_VALUE_TO_TARGET_DOUBLE (rv, t);
      values[0] = SIGN_EXTEND_SHORT (t[1]);
      values[1] = SIGN_EXTEND_SHORT (t[1] >> 16);
      values[2] = SIGN_EXTEND_SHORT (t[0]);
      values[3] = SIGN_EXTEND_SHORT (t[0] >> 16);
    }

  else if (code == CONST_DOUBLE
	   && (GET_MODE (x) == DImode || GET_MODE (x) == VOIDmode))
    {
      v = CONST_DOUBLE_HIGH (x);
      values[0] = SIGN_EXTEND_SHORT (v);
      values[1] = SIGN_EXTEND_SHORT (v >> 16);

      v = CONST_DOUBLE_LOW (x);
      values[2] = SIGN_EXTEND_SHORT (v);
      values[3] = SIGN_EXTEND_SHORT (v >> 16);
    }

  else
    fatal_insn ("Bad rtx passed to nios_split_constant64:", x);
}
/* }}} */

/* {{{ nios_subword()
   
   Return an RTX that refers to the nth subword of a multiword
   expression.  This is similar to operand_subword, but that
   has a bias that SImode or SFmode is a word.  */

rtx
nios_subword (op, word_num, target_mode, source_mode)
     rtx op;
     int word_num;
     enum machine_mode target_mode;
     enum machine_mode source_mode;
{
  enum rtx_code code = GET_CODE (op);
  enum machine_mode mode = GET_MODE (op);
  int split_values[4];

  if (! IN_RANGE_P (word_num, 0, 3))
    error ("nios_subword: Internal error -- word_num (%d) "
	   "should be >= 0 && <= 3", word_num);

  if (target_mode == SImode 
      || target_mode == SFmode
      && (word_num != 0 && word_num != 2))
    error ("nios_subword: Internal error -- word_num (%d) should be 0 or 2 "
	   "for SImode", word_num);

  if (target_mode != HImode && target_mode != SImode)
    error ("nios_subword: Internal error -- target_mode (%s) must be HImode "
	   "or SImode", GET_MODE_NAME (target_mode));

  if (code == REG)
    {
      if ((UNITS_PER_WORD * word_num) >= GET_MODE_SIZE (mode))
	error ("nios_subword: Internal error -- word_num (%d) is too large "
	       "for mode %s", word_num, GET_MODE_NAME (mode));

      return gen_rtx (REG, target_mode, REGNO (op) + word_num);
    }

  else if (code == SUBREG)
    {
      if ((UNITS_PER_WORD * word_num) >= GET_MODE_SIZE (mode))
	error ("nios_subword: Internal error -- word_num (%d) is too large "
	       "for mode %s", word_num, GET_MODE_NAME (mode));

      do
	{
	  word_num += SUBREG_WORD (op);
	  op = SUBREG_REG (op);
	}
      while (GET_CODE (op) == SUBREG);

      return gen_rtx (REG, target_mode, REGNO (op) + word_num);
    }

  else if ((code == CONST_INT || code == CONST_DOUBLE) 
           && GET_MODE_SIZE (source_mode) == 2)
    {
      if (word_num > 1)
	error ("nios_subword: Internal error -- word_num (%d) is too large for "
	       "32-bit constants", word_num);

      nios_split_constant32 (op, split_values);
      return GEN_INT (SIGN_EXTEND_SHORT (split_values[word_num]));
    }

  else if ((code == CONST_INT || code == CONST_DOUBLE)
           && GET_MODE_SIZE (source_mode) == 4)
    {
      nios_split_constant32 (op, split_values);
      return GEN_INT (SIGN_EXTEND_SHORT (split_values[word_num]));
    }

  else if ((code == CONST_INT || code == CONST_DOUBLE) 
           && GET_MODE_SIZE (source_mode) == 8)
    {
      HOST_WIDE_INT value;

      nios_split_constant64 (op, split_values);
      value = SIGN_EXTEND_SHORT (split_values[word_num]);
      if (GET_MODE_SIZE (target_mode) == 4)
	value = (value << 16) & (split_values[word_num+1] & 0xffff);

      return GEN_INT (value);
    }

  else if (code == MEM)
    return change_address (op, target_mode, 
	                   plus_constant (XEXP (op, 0), word_num * UNITS_PER_WORD));

  else
    fatal_insn ("Cannot get subword of: ", op);

  return NULL_RTX;
}
/* }}} */

/* {{{ nios_get_value()

   Return the given a CONST_INT or CONST_DOUBLE as HOST_WIDE_INT
   constant.  */
   
HOST_WIDE_INT
nios_get_value (op)
    rtx op;
{
  HOST_WIDE_INT value;

  if (GET_CODE (op) != CONST_INT && GET_CODE (op) != CONST_DOUBLE)
    fatal_insn ("Cannot get value for: ", op);


  if (GET_CODE (op) == CONST_DOUBLE)
    {
      REAL_VALUE_TYPE rv;

      REAL_VALUE_FROM_CONST_DOUBLE (rv, op);
      REAL_VALUE_TO_TARGET_SINGLE (rv, value);
    }
  else
    value = INTVAL (op);

  return value;
}
/* }}} */

/* {{{ nios_power_of_two()
   
   Return true if 'value' is a valid power of two <= 2**32.  */

int
nios_power_of_two (value)
    HOST_WIDE_INT value;
{
  int power = exact_log2 (value);
  return (power >= 0 && power <= 32);
}
/* }}} */

/* {{{ nios_encode_label()
   
   Encode a function declaration name to mark it as a local function for
   potential bsr relaxation.  */

void
nios_encode_label (sym)
     rtx sym;
{
  char *str = XSTR (sym, 0);
  int len = strlen (str);
  char *newstr;

  newstr = (char *) obstack_alloc (saveable_obstack, len + 2);
  strcpy (newstr + 1, str);
  *newstr = LOCAL_FUNCTION_MARKER;
  XSTR (sym, 0) = newstr;
}
/* }}} */

/* {{{ nios_local_function_operand()

   Return true if OP is a reference to a function defined locally in this
   compilation unit.  */

int
nios_local_function_operand (op, mode)
   rtx op;
   enum machine_mode mode ATTRIBUTE_UNUSED;
{
  rtx symbol = (GET_CODE (op) == MEM) ? XEXP (op, 0) : op;

  return ((GET_CODE (symbol) == SYMBOL_REF)
	   && (XSTR (symbol, 0))[0] == LOCAL_FUNCTION_MARKER);
}

/* }}} */
/* }}} */

/* {{{ Reload helpers.  */

/* {{{ nios_secondary_reload_class()
   What (if any) secondary registers are needed to move IN with mode
   MODE into a register from in register class CLASS.  */

enum reg_class
nios_secondary_reload_class (class, mode, in)
     enum reg_class class;
     enum machine_mode mode;
     rtx in;
{
  /* Memory loads less than a full word wide can't have an address or stack
     pointer destination.  They must use a data register as an intermediate
     register.  */
  if (GET_CODE (in) == MEM
      && GET_MODE_SIZE (mode) < UNITS_PER_WORD
      && class == BP_REGS)
    return DATA_REGS;

  /* Otherwise assume no secondary reloads are needed.  */
  return NO_REGS;
}
/* }}} */


/* Split a move into two smaller pieces.
   MODE indicates the reduced mode.  OPERANDS[0] is the original destination
   OPERANDS[1] is the original src.  The new destinations are
   OPERANDS[2] and OPERANDS[4], while the new sources are OPERANDS[3]
   and OPERANDS[5].  */

void
nios_split_words (nmode, omode, operands)
     enum machine_mode nmode,omode;
     rtx *operands;
{
  rtx dl,dh;	/* src/dest pieces */
  rtx sl,sh;
  int	move_high_first = 0;	/* Assume no overlap */

  switch (GET_CODE (operands[0])) /* DEST */
    {

    case SUBREG:
    case REG:
      if ((GET_CODE (operands[1]) == REG
	   || GET_CODE (operands[1]) == SUBREG)
	  && true_regnum (operands[0]) > true_regnum (operands[1]))
	move_high_first = 1;

      if (GET_CODE (operands[0]) == SUBREG)
	{
	  dl = gen_rtx_SUBREG (nmode, SUBREG_REG (operands[0]), 0);
	  dh = gen_rtx_SUBREG (nmode, SUBREG_REG (operands[0]), 1);
	}
      else if (GET_CODE (operands[0]) == REG && ! IS_PSEUDO_P (operands[0]))
	{
	  int	r = REGNO (operands[0]);
	  dh = gen_rtx_REG (nmode, r + HARD_REGNO_NREGS (r, nmode));
	  dl = gen_rtx_REG (nmode, r );
	}
      else
	{
	  dh = gen_rtx_SUBREG (nmode, operands[0], 1);
	  dl = gen_rtx_SUBREG (nmode, operands[0], 0);
	}
      break;

    case MEM:
      switch (GET_CODE (XEXP (operands[0], 0)))
	{
	case POST_INC:
	case POST_DEC:
	  abort ();
	default:
	  dl = operand_subword (operands[0], 0,	0, omode);
	  dh = operand_subword (operands[0], 
				GET_MODE_SIZE (nmode)/UNITS_PER_WORD,
				0, omode);
	}
      break;
    default:
      abort ();
    }

  switch (GET_CODE (operands[1]))
    {
    case REG:
      if (! IS_PSEUDO_P (operands[1]))
	{
	  int r = REGNO (operands[1]);

	  sh = gen_rtx_REG (nmode, r + HARD_REGNO_NREGS (r, nmode));
	  sl = gen_rtx_REG (nmode, r);
	}
      else
	{
	  sh = gen_rtx_SUBREG (nmode, operands[1], GET_MODE_SIZE (nmode));
	  sl = gen_rtx_SUBREG (nmode, operands[1], 0);
	}
      break;

    case CONST_DOUBLE:
      if (operands[1] == const0_rtx)
	sh = sl = const0_rtx;
      else
	{
	  REAL_VALUE_TYPE rv;
	  long value;

	  REAL_VALUE_FROM_CONST_DOUBLE (rv, operands[1]);
	  REAL_VALUE_TO_TARGET_SINGLE (rv, value);

	  sh = GEN_INT (CONST_DOUBLE_HIGH (operands[1]));
	  sl = GEN_INT (CONST_DOUBLE_LOW (operands[1]));
	}
      break;

    case CONST_INT:
      if (operands[1] == const0_rtx)
	sh = sl = const0_rtx;
      else
	{
	  int vl, vh;

	  switch (nmode)
	    {
	    default:
	      abort ();
	    }
	    
	  sl = GEN_INT (vl);
	  sh = GEN_INT (vh);
	}
      break;

    case SUBREG:
      sl = gen_rtx_SUBREG (nmode, SUBREG_REG (operands[1]), 0);
      sh = gen_rtx_SUBREG (nmode, SUBREG_REG (operands[1]), 1);
      break;

    case MEM:
      switch (GET_CODE (XEXP (operands[1], 0)))
	{
	case POST_DEC:
	case POST_INC:
	  abort ();
	  break;
	default:
	  sl = operand_subword (operands[1], 0, 0, omode);
	  sh = operand_subword (operands[1], 
				GET_MODE_SIZE (nmode)/UNITS_PER_WORD,
				0, omode);
	  
	  /* Check if the DF load is going to clobber the register
             used for the address, and if so make sure that is going
             to be the second move */
	  if (GET_CODE (dl) == REG && regno_use_in (true_regnum (dl), sl))
	    {
	      move_high_first = 1;
	      if (GET_CODE (dh) == REG && regno_use_in (true_regnum (dh), sl))
		{
		  /* Cannot win by swapping the order, so give up */
		  abort ();
		}
	    }
	}
      break;
    default:
      abort ();
    }

  if (move_high_first)
    {
      operands[2] = dh;
      operands[3] = sh;
      operands[4] = dl;
      operands[5] = sl;
    }
  else
    {
      operands[2] = dl;
      operands[3] = sl;
      operands[4] = dh;
      operands[5] = sh;
    }
  return;
}


/* }}} */

/* {{{ Flat helpers */
/* {{{ nios_immediate_insn()
 * Helper for nios_addi() and nios_subi().
 */

static void
nios_immediate_insn (stream, instruction, reg, value)
     FILE *stream;
     const char *instruction;
     const char *reg;
     int value;
{
  if (CONST_OK_FOR_L (value))
    {
      fprintf (stream, "\t" "%s\t%s, %d\n", instruction, reg, value);
    }
  else if (CONST_OK_FOR_N (value))
    {
      fprintf (stream,
	       "\t"
	       "pfx\t%%hi(%d)\n\t"
	       "%s\t%s, %%lo(%d)\n", value, instruction, reg, value);
    }
  else
    {
      abort();
    }
}

/* }}} */

static void
nios_subi (stream, reg, value)
     FILE *stream;
     const char *reg;
     int value;
{
  if (0 == value)
    {
      return;
    }

  nios_immediate_insn (stream, "subi", reg, value);
}

static void
nios_addi (stream, reg, value)
     FILE *stream;
     const char *reg;
     int value;
{
  if (0 == value)
    {
      return;
    }

  nios_immediate_insn (stream, "addi", reg, value);
}

/* }}} */

/* {{{ Nios flat code
 *
 * Notes:
 *
 *   - Flat code uses the same ABI as round code.  There isn't really
 *     a flat ABI.  This means you should be able to safely link flat
 *     code with round code.
 *
 *   - As a result, it is bad to use "%fp" explicitly.  Always use
 *     reg_names[GRP_FP].
 *
 *   - MULTILIB_OPTIONS has changed in t-nios.  This means the entire
 *     cygwin tree should be reconfigured and recompiled.  That will
 *     build the mflat c library, etc.
 *
 *   - The old Nios SDK interrupt handler will not work with flat
 *     code.  That handler assumes the ISR does a SAVE.  Now all
 *     call clobbered regies need to be saved.
 *
 *   - The compiler generally wants to use a BP register like %l0
 *     in LDP cases where there is a zero offset.  It would be better
 *     to use LD and a call clobbered register instead.  This would
 *     save having to save and restore the %l0 register in the prologue
 *     epilogue.
 *
 * -- JMR
 *
 * TODO: debugging.
 */

static struct
{
  /* All in bytes */
  int locals;
  int fp_save;
  int saves;
  int args;
  int aggregate;
  int cwp;
  int total;

  /* Which registers to save or'd into a word */
  int save_me;

  /* Boolean */
  int need_new_frame;
}
stack_info;

#define NIOS_FLAT_SAVE_SP              (frame_pointer_needed)

/* Byte %sp offset */
#define NIOS_FLAT_SAVE_BASE \
    (stack_info.args + stack_info.aggregate + stack_info.cwp)

/* %sp offset in words for the frame pointer, mostly for gdb to look at. */
#define NIOS_FLAT_FP_SAVE_OFFSET       (NIOS_FLAT_SAVE_BASE / UNITS_PER_WORD)

/* %sp offset in words for saving caller's regs */
#define NIOS_FLAT_CALLER_SAVE_OFFSET   (NIOS_FLAT_FP_SAVE_OFFSET + 1)

/* Number of a call-clobbered register which may be clobbered
 * in the epliogue. */
#define NIOS_FLAT_EPILOGUE_SCRATCH      7 /* %g7 */


/* {{{ nios_flat_conditional_register_usage()
 * Used by the CONDITIONAL_REGISTER_USAGE macro for flat targets.
 */
void
nios_flat_conditional_register_usage ()
{
    /* Not used at the moment. */
}

/* }}} */

/* {{{ nios_flat_must_save_register()
 * See if register needs to be saved in the prologue.
 */

static int
nios_flat_must_save_register (long regno)
{
  if (regs_ever_live[regno] && !call_used_regs[regno])
    {
      /*
       * The callee used it and the caller was counting on
       * it being preserved.
       */
      return 1;
    }

  switch (regno)
  {
  case GPR_RETURN:
  case GPR_LINK:
      /*
       * The callee probably makes a call or bsr.  The callee
       * needs to preserve its own return address.
       */
      if (regs_ever_live[regno])
      {
          return 1;
      }
  default:
      break;
  }

  return 0;
}

/* }}} */

/* {{{ nios_flat_compute_frame_size()
 * Add up all the bytes of a flat stack frame.
 */
static void
nios_flat_compute_frame_size (size, leaf_function)
     int size;
     int leaf_function;
{
  int regno;
  int save_space = 0;
  int total = 0;
  int arg_space;

  if (TARGET_DEBUG_STACK)
    {
      fprintf (stderr,
	       "\n\nStack info for flat function %s()\n"
	       "\tRegisters to save [", CURRENT_FUNCTION_NAME);
    }

  memset (&stack_info, 0, sizeof (stack_info));

  /* Which of the caller's register to save? */
  for (regno = GPR_FIRST; regno < GPR_LAST; regno++)
    {
      if (nios_flat_must_save_register (regno))
	{
	  if (TARGET_DEBUG_STACK)
	    {
	      fprintf (stderr, " %s", reg_names[regno]);
	    }

	  save_space += UNITS_PER_WORD;
	  stack_info.save_me |= (1 << regno);
	}
    }

  if (TARGET_DEBUG_STACK)
    {
      fprintf (stderr, " ]\n");
    }

#define TOTAL(var, val)  do {                                              \
    int val_ = (val);                                                      \
                                                                           \
    (var) = val_;                                                          \
    total += val_;                                                         \
                                                                           \
    if (TARGET_DEBUG_STACK)                                                \
    {                                                                      \
        fprintf(stderr, " %32s = %-4d bytes, %-3d words\n",                \
                        #var, (val_), (val_) / UNITS_PER_WORD);            \
    }                                                                      \
} while (0)

  /* Space to stash local vars */
  TOTAL (stack_info.locals, size);

  /* Save the frame pointer, the debugger wants it. */
  TOTAL (stack_info.fp_save, UNITS_PER_WORD);

  /* Caller's registers which must be saved */
  TOTAL (stack_info.saves, save_space);

  /* Always make space for %o0..%o5 */
  arg_space = REG_PARM_STACK_SPACE (current_function_decl);

  /* Some args are passed on the stack, space is allocated by the caller */
  TOTAL (stack_info.args, current_function_outgoing_args_size + arg_space);

  /* The system area for reg file overflow */
  TOTAL (stack_info.cwp, NIOS_REGISTER_WINDOW_SIZE * UNITS_PER_WORD);

  /* One pointer to a struct return value */
  TOTAL (stack_info.aggregate, UNITS_PER_WORD);

#undef TOTAL

  /* Is a new register frame needed? */
  stack_info.need_new_frame = (size || save_space);

  if (stack_info.need_new_frame)
    {
      /* Make room for %fp */
      stack_info.saves += UNITS_PER_WORD;
    }

  /* STARTING_FRAME_OFFSET is probably zero? */

  if (STARTING_FRAME_OFFSET)
  {
      abort();
  }

  total -= STARTING_FRAME_OFFSET;
  stack_info.total = total;

  if (TARGET_DEBUG_STACK)
    {
      fprintf (stderr, " %32s = %-4d bytes, %-3d words\n",
	       "TOTAL", total, total / UNITS_PER_WORD);
    }
}

/* }}} */

/* {{{ nios_flat_sp_operation()
 * Helper for nios_flat_save_restore_regs().
 */
static void
nios_flat_sp_operation (stream, store_or_load, regno, offset)
     FILE *stream;
     int store_or_load;
     int regno;
     int offset;		/* words */
{
  const char *reg = reg_names[regno];

  if (CONST_OK_FOR_K (offset))
    {
      if (store_or_load)
	{
          fprintf (stream, "\tsts\t[%%sp,%d], %s\n", offset, reg);
	}
      else
	{
	  fprintf (stream, "\tlds\t%s, [%%sp,%d]\n", reg, offset);
	}
    }
  else
    {
      /* This should never happen.  The saved regs are stored near the %sp. */
      abort ();
    }
}

/* }}} */

/* {{{ nios_flat_save_restore_regs()
 * Save or store the caller's registers which need to be preserved.
 * Used by nios_flat_function_prologue() and nios_flat_function_epilogue()
 */
static void
nios_flat_save_restore_regs (stream, regs, start_offset, store_or_load)
     FILE *stream;
     int regs;
     int start_offset;		/* words */
     int store_or_load;
{
  int regno;
  int offset = start_offset;	/* words */

  for (regno = GPR_FIRST; regno < GPR_LAST; regno++)
    {
      if (regs & (1 << regno))
	{
          nios_flat_sp_operation (stream, store_or_load, regno, offset);

          /*
          if (store_or_load && dwarf2out_do_frame ())
          {
            dwarf2out_return_save ("", offset - stack_info.total);
          }
         */
          offset++;
        }
    }
}

/* }}} */

/* {{{ nios_flat_emit_stack_reservation()
 * Expand the flat stack.  Called in nios_flat_function_prologue()
 */
static void
nios_flat_emit_stack_reservation (stream, save_me)
FILE *stream;
int save_me;
{
  nios_comment (stream, "Expand callers frame %d words.",
		stack_info.total / UNITS_PER_WORD);
  fprintf (stream, "\tmov\t%%g1, %%sp\n");
  nios_subi (stream, "%sp", stack_info.total);
  fprintf (stream, "\tsts \t[%%sp,%d], %%fp\n", NIOS_FLAT_FP_SAVE_OFFSET);
}

/* }}} */

static void
nios_flat_prologue_dwarf (void)
{
  char *label = dwarf2out_cfi_label ();

  /* The canonical frame address refers to the top of the frame.  */
//  dwarf2out_def_cfa (label, STACK_POINTER_REGNUM, NIOS_FLAT_FP_SAVE_OFFSET);
}

/* {{{ nios_flat_function_prologue()
 * Emit the prologue for -mflat code.
 */
void
nios_flat_function_prologue (stream, size, leaf_function)
     FILE *stream;
     int size;
     int leaf_function;
{
  nios_flat_compute_frame_size (size, leaf_function);

  if (stack_info.need_new_frame)
    {
      nios_comment (stream, "flat prologue begin");
      nios_flat_emit_stack_reservation (stream, stack_info.save_me);

  /*    if (dwarf2out_do_frame())
        {
          nios_flat_prologue_dwarf ();
        }
  */
      if (stack_info.save_me)
	{
	  nios_comment (stream, "save call preserved regs");
	  nios_flat_save_restore_regs (stream,
				       stack_info.save_me,
                                       NIOS_FLAT_CALLER_SAVE_OFFSET,
                                       1);
	}

      nios_comment (stream, "Set frame pointer");
      /* gdb looks for this at the end of the prologue. */
      fprintf (stream, "\tmov \t%s, %%g1\n", reg_names[GPR_FP]);
      nios_comment (stream, "flat prologue end");
    }
}

/* }}} */

/* {{{ nios_flat_eligible_for_epilogue_delay()
 * Check the instruction 'trial' to see if it can be used in
 * the epilogue delay slot after the 'lret'.
 */

int
nios_flat_eligible_for_epilogue_delay (trial, slot)
     rtx trial;
     int slot;
{
  rtx pat = PATTERN (trial);
  int attr_len = get_attr_length (trial);
  int regno;

  if (attr_len != 2)
    {
      /* Avoid instructions with pfx, etc. */
      return 0;
    }

  if (GET_CODE (trial) != INSN)
    {
      /* No jumps, calls, or oddball things */
      return 0;
    }

  if (reg_mentioned_p (stack_pointer_rtx, pat) ||
      reg_mentioned_p (frame_pointer_rtx, pat))
    {
      /* %sp and %fp belong to the caller at this point. */
      return 0;
    }

  if (regno_use_in (NIOS_FLAT_EPILOGUE_SCRATCH, pat))
    {
      return 0;
    }

  /*
   * Only allow %g1..%g6 or %o0..%o5. It is not known which of
   * the other registers may have been restored in the epilogue.
   * That info is calculated after this function call.
   */

  for (regno = 14; regno < 32; regno++)
    {
      if (regno_use_in (regno, pat))
	{
	  return 0;
	}
    }

  /* All tests passed, ok. */
  return 1;
}

/* }}} */

/* {{{ nios_flat_function_epilogue()
 * Emit the epilogue for flat code.
 */
void
nios_flat_function_epilogue (stream, size)
     FILE *stream;
     int size;
{
  int have_delay_insn = (int) current_function_epilogue_delay_list;
  const char *delay_insn = "\tnop\n";

  nios_comment (stream, "flat epilogue begin");

  if (stack_info.need_new_frame)
    {
      if (NIOS_FLAT_SAVE_SP)
	{
	  /*
	   * A temporary reg is used so that the %sp modification
	   * is done in one step.  The alternative:
	   *
	   *    mov   %sp, %fp
	   *    pfx   %hi(123)
	   *    subi  %sp, %lo(123)
	   *
	   * causes %sp to be wrong for 1 instruction.  An ISR can
	   * then come in and scribble on important things.
	   */
	  nios_comment (stream, "restore callee's %%sp");
	  fprintf (stream, "\tmov\t%s, %s\n",
		   reg_names[NIOS_FLAT_EPILOGUE_SCRATCH], reg_names[GPR_FP]);
	  nios_subi (stream,
		     reg_names[NIOS_FLAT_EPILOGUE_SCRATCH], stack_info.total);
	  fprintf (stream, "\tmov\t%%sp, %s\n",
		   reg_names[NIOS_FLAT_EPILOGUE_SCRATCH]);
	}

      if (stack_info.save_me)
	{
	  nios_comment (stream, "restore caller's regs");
	  nios_flat_save_restore_regs (stream,
				       stack_info.save_me,
				       NIOS_FLAT_CALLER_SAVE_OFFSET, 0);
	}

      nios_comment (stream, "restore %s", reg_names[GPR_FP]);
      fprintf (stream,
               "\tlds\t%s, [%%sp,%d]\n",
               reg_names[GPR_FP], NIOS_FLAT_FP_SAVE_OFFSET);

      nios_comment (stream, "restore caller's %%sp");
      nios_addi (stream, "%sp", stack_info.total);
    }

  fprintf (stream, "\tjmp\t%%o7\n");

  /* Fill the return delay slot */
  if (!have_delay_insn)
    {
      fprintf (stream, "%s", delay_insn);
    }
  else
    {
      final_scan_insn (XEXP (current_function_epilogue_delay_list, 0),
		       stream, 1, 0, 1);
    }

  nios_comment (stream, "flat epilogue end");
}

/* }}} */

/* }}} */
