/* Target-dependent code for the Altera NIOS, for GDB, the GNU Debugger.
   Copyright 2000 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <stdarg.h>
#include "defs.h"
#include "symtab.h"
#include "symfile.h"
#include "gdbtypes.h"
#include "inferior.h"
#include "gdbcore.h"
#include "arch-utils.h"

/*
 * Extra frame info needed for flat code.  Things can get complicated
 * due to the possibiliy of mixing flat and round code.
 */
struct frame_extra_info
{
    int flat_stack_space;    /* Save space in bytes for flat frames. */
    int flat_sp_offsets[32]; /* Prologue save offsets from %sp.      */
    int save_words;          /* From "SAVE %sp,-x" in round code.    */
    const char *name;        /* Function name.                       */
};

/* A flat frame does not have a SAVE / RESTORE instruction pair. */
#define FRAME_FLAT(frame)         \
    (0 == (frame)->extra_info->save_words)

/* A 'simple' frame is flat and does not alter %sp/%fp or make calls: */
#define FRAME_SIMPLE(frame)    \
    (FRAME_FLAT(frame) && (0 == (frame)->extra_info->flat_stack_space))

/* For development use. */
#define FRAME_NAME(f)                                             \
    ((f) ?                                                        \
    ((f)->extra_info ?                                            \
    ((f)->extra_info->name ? (f)->extra_info->name : "<no name>") \
    : "<no extra info>")                                          \
    : "<null frame>")

extern void _initialize_nios_tdep (void);


/* Quirks and eccentricities of the Nios:

   On the Nios, the PC is one bit longer than all the other
   registers, and its low bit is zero.  Whenever the PC is used as
   data (i.e., return addresses, indirect jumps), the data is the true
   address shifted right one bit.  So if we have int (*f) (int) or
   something, then if f holds the bit pattern 0x4040, f refers to the
   function at 0x8080.

   Here's how we deal with this in GDB:

   - All pointer types, both code and data, are the same size.
     Nios-16: TYPE_LENGTH (any_ptr_type) = 2
     Nios-32: TYPE_LENGTH (any_ptr_type) = 4

   - In `struct value' objects, both code and data pointers use the
     same format the processor uses in memory.  That is, the bit
     pattern 0x4040 refers to the address 0x4040 when used as a data
     pointer, but refers to address 0x8080 when used as a code
     pointer.

   - Values in registers use the same representation, except for the
     PC.  On both the Nios-16 and the Nios-32, the raw form of the
     PC is four bytes long, and its contents are the code address,
     unchanged.  We define a raw / virtual conversion for the PC, so
     that referring to $pc gives you a sixteen-bit value, the same as
     any other code pointer.

   - To convert between pointer values and actual CORE_ADDR addresses,
     we provide nios-specific functions for POINTER_TO_ADDRESS and
     ADDRESS_TO_POINTER; look for nios_pointer_to_address and
     nios_address_to_pointer, below.

   For the Nios-32 architecture, the simulator only gives us a 32-bit
   PC, and we use ELF-32 object files, so even though the code address
   space is 8Gb, we don't really deal with code addresses in the upper
   4Gb.  */


/* The structure describing a particular nios variant.

   This structure holds the parameters that differ between one nios
   variant and another.  The current gdbarch structure points to this.  */

struct gdbarch_tdep {

  /* The layout of the register file.  */
  struct register_map *registers;

};

#define CURRENT_REGISTER_MAP (gdbarch_tdep (current_gdbarch)->registers)



/* The Nios register set.  */

enum {
  nios_num_regs = 36,

  first_gr_regnum = 0,
  last_gr_regnum = 31,

  first_out_regnum   =  8,
  last_out_regnum    = 15,

  first_local_regnum = 16,
  last_local_regnum  = 23,

  first_in_regnum    = 24,
  last_in_regnum     = 31,

  pc_regnum = 32,
  first_ctl_regnum = 33,
  last_ctl_regnum = 35,

  /* Special names for registers mentioned above that have designated
     purposes.  */
  return_regnum = 8,
  sp_regnum = 14,
  fp_regnum = 30,

  /* The `call' and `callc' instructions save the return address in
     the LINK register.  However, most functions start with a `save'
     instruction, and do a `restore' before returning.  A `save'
     instruction moves the LINK register, $o7, to the SAVE register,
     $i7.  So where you find the return address depends on whether
     you've executed the prologue or not.  */
  link_regnum = 15,
  save_regnum = 31,

  status_regnum = 33,
  wvalid_regnum = 35
};

/* This gets initialized on startup.  */
static char *nios_register_names[] = {
  "g0", "g1", "g2", "g3", "g4", "g5", "g6", "g7",
  "o0", "o1", "o2", "o3", "o4", "o5", "sp", "o7",
  "l0", "l1", "l2", "l3", "l4", "l5", "l6", "l7",
  "i0", "i1", "i2", "i3", "i4", "i5", "fp", "i7",
  "pc", "ctl0", "ctl1", "ctl2"
};

static char *
nios_register_name (int reg)
{
  if (reg < 0 && reg < nios_num_regs)
    return "?toosmall?";
  if (reg >= nios_num_regs)
    return "?toolarge?";

  return nios_register_names[reg];
}

/*
 * This debug code was added for the flat port.  Please don't remove it
 * until we are sure flat debugging is ok.
 */
#if 0
static void __attribute__((format (printf, 1, 2)))
verbose (const char *s, ...)
{
    va_list v;

    va_start (v, s);
    fprintf (stderr, "* ");
    vfprintf (stderr, s, v);
    fprintf (stderr, "\n");
    va_end (v);
}
#else
#define verbose(args...)
#endif


/* Information about a single register.  */
struct one_register {
  int raw_size;			/* the register's raw size */
  int offset;			/* its offset in the register file */
  struct type *virtual_type;	/* its virtual type  */
};

/* A map of a register set.  */
struct register_map {
  struct one_register reg[nios_num_regs];
  int total_size;		/* the total register files size, in bytes  */
};

/* Return a register map for the variant whose BFD machine number is MACH.  */
static struct register_map *
make_register_map (int mach)
{
  struct register_map *map = xmalloc (sizeof (*map));
  struct type *void_ptr = lookup_pointer_type (builtin_type_void);
  struct type *void_func_ptr
    = lookup_pointer_type (lookup_function_type (builtin_type_void));
  int i;

  memset (map, 0, sizeof (*map));


  if (mach == bfd_mach_nios16)
    {
      for (i = first_gr_regnum; i <= last_gr_regnum; i++)
	{
	  struct one_register *r = &map->reg[i];

	  r->raw_size = 2;
	  r->virtual_type = builtin_type_int16;
	}
    }
  else if (mach == bfd_mach_nios32)
    {
      for (i = first_gr_regnum; i <= last_gr_regnum; i++)
	{
	  struct one_register *r = &map->reg[i];

	  r->raw_size = 4;
	  r->virtual_type = builtin_type_int32;
	}
    }
  else
    internal_error ("nios-tdep.c (make_register_map): "
		    "unrecognized machine");

  /* The SP and the FP are both void * values.  */
  map->reg[sp_regnum].virtual_type = void_ptr;
  map->reg[fp_regnum].virtual_type = void_ptr;

  /* The LINK and SAVE registers hold code addresses.  */
  map->reg[link_regnum].virtual_type = void_func_ptr;
  map->reg[save_regnum].virtual_type = void_func_ptr;

  /* On the Nios-16, the PC is 17 bits long.  So we claim it's four
     bytes long on both variants, and then use the "virtual / raw"
     conversion macros to put it in the same format as everything else.  */
  map->reg[pc_regnum].raw_size = 4;
  map->reg[pc_regnum].virtual_type = void_func_ptr;

  for (i = first_ctl_regnum; i <= last_ctl_regnum; i++)
    {
      struct one_register *r = &map->reg[i];

      r->raw_size = 2;
      r->virtual_type = builtin_type_uint16;
    }

  /* Assume that there are no gaps between registers in the register
     file, and fill in map->offset and map->total_size.  */
  {
    int here = 0;

    for (i = 0; i < nios_num_regs; i++)
      {
	struct one_register *r = &map->reg[i];

	r->offset = here;
	here += r->raw_size;
      }

    map->total_size = here;
  }

  /* Sanity check: make sure that we've filled in data for every
     register.  */
  for (i = 0; i < nios_num_regs; i++)
    {
      struct one_register *r = &map->reg[i];

      if (r->raw_size == 0 || ! r->virtual_type)
	internal_error ("nios-tdep.c (make_register_map): "
			"forgot to initialize register %d in register map", i);
    }

  return map;
}


static int
nios_register_raw_size (int reg)
{
  struct register_map *map = CURRENT_REGISTER_MAP;

  if (0 <= reg && reg < nios_num_regs)
    return map->reg[reg].raw_size;

  internal_error ("nios-tdep.c (nios_register_raw_size): "
		  "bad register number: %d", reg);
}

static int
nios_register_byte (int reg)
{
  struct register_map *map = CURRENT_REGISTER_MAP;

  if (0 <= reg && reg < nios_num_regs)
    return map->reg[reg].offset;

  internal_error ("nios-tdep.c (nios_register_byte): "
		  "bad register number: %d", reg);
}

static struct type *
nios_register_virtual_type (int reg)
{
  struct register_map *map = CURRENT_REGISTER_MAP;

  if (0 <= reg && reg < nios_num_regs)
    return map->reg[reg].virtual_type;

  internal_error ("nios-tdep.c (nios_register_virtual_type): "
		  "bad register number: %d", reg);
}

static int
nios_register_virtual_size (int reg)
{
  struct register_map *map = CURRENT_REGISTER_MAP;

  if (0 <= reg && reg < nios_num_regs)
    return TYPE_LENGTH (map->reg[reg].virtual_type);

  internal_error ("nios-tdep.c (nios_register_virtual_size): "
		  "bad register number: %d", reg);
}

static int
nios_register_convertible (int reg)
{
  return reg == pc_regnum;
}

static void
nios_convert_to_virtual (int reg, struct type *type, char *from, char *to)
{
  if (reg == pc_regnum)
    {
      ULONGEST val = extract_unsigned_integer (from, 4);

      if (val & 1)
	warning ("PC register seems to contain an unaligned address");

      store_unsigned_integer (to, TYPE_LENGTH (type), val >> 1);
    }
  else
    internal_error ("nios-tdep.c (nios_convert_to_virtual): "
		    "register %d isn't convertible", reg);
}

static void
nios_convert_to_raw (struct type *type, int reg, char *from, char *to)
{
  if (reg == pc_regnum)
    {
      ULONGEST val = extract_unsigned_integer (from, TYPE_LENGTH (type));

      // !!! dont multiply by two? dvb--fixes /2 error in pc entry store_unsigned_integer (to, 4, val << 1);
      store_unsigned_integer (to, 4, val);
    }
  else
    internal_error ("nios-tdep.c (nios_convert_to_raw): "
		    "register %d isn't convertible", reg);
}


/* Return the value of the current window pointer, CWP.  */
static int
nios_get_cwp ()
{
  return (read_register (status_regnum) >> 4) & 0x1f;
}

static int
nios_get_hi_limit ()
{
  return (read_register (wvalid_regnum) >> 5) & 0x1f;
}

/* Set the current window pointer to WIN.  */
static void
nios_set_cwp (int win)
{
  if (win & -32)
    internal_error ("nios-tdep.c (nios_set_cwp): "
		    "register window value out of range: %d",
		    win);

  write_register (status_regnum,
		  ((read_register (status_regnum) & ~(0x1f << 4))
		   | (win << 4)));

  /* Storing a new value in the CWP causes the register windows to
     shift around, so we need to flush our cache of register values.  */
  registers_changed ();
}



/* The Nios uses odd pointer formats.  */

static CORE_ADDR
nios_pointer_to_address (struct type *type, char *buf)
{
  /* We know TYPE is either a pointer or a reference.  If it's a code
     pointer, we need to shift it left one bit; if it's a data
     pointer, it's fine as it is.  */
  enum type_code target = TYPE_CODE (TYPE_TARGET_TYPE (type));
  CORE_ADDR addr = extract_address (buf, TYPE_LENGTH (type));

  if (target == TYPE_CODE_FUNC || target == TYPE_CODE_METHOD)
    addr <<= 1;

  return addr;
}

static void
nios_address_to_pointer (struct type *type, char *buf, CORE_ADDR addr)
{
  /* We know TYPE is either a pointer or a reference.  If it's a code
     pointer, we need to shift it right one bit; if it's a data
     pointer, it's fine as it is.  */
  enum type_code target = TYPE_CODE (TYPE_TARGET_TYPE (type));
  
  if (target == TYPE_CODE_FUNC || target == TYPE_CODE_METHOD)
    addr >>= 1;

  store_address (buf, TYPE_LENGTH (type), addr);
}


/* Instruction format macros.  */

/* The insn_* macros are named after the instruction formats in the
   Altera Nios processor spec.  It seems more readable to use these
   than to write out the bits directly.

   All these macros mask off all but the needed bits from their
   operands, so you can specify -1 as an operand to get a field filled
   with ones.  */

/* Note argument order!  */
#define insn_RR(op6, a, b) \
  (((op6) & 0x3f) << 10 | ((b) & 0x1f) << 5 | ((a) & 0x1f))

#define insn_i11(op5, k) \
  ((((op5) & 0x1f) << 1) | ((k) & 0x7ff))

#define insn_i5v(op6, op2v, v) \
  ((((op6) & 0x3f) << 10) | (((op2v) & 3) << 8) | ((v) & 0x1f))

#define insn_i8v(op6, op2v, v) \
  ((((op6) & 0x3f) << 10) | (((op2v) & 3) << 8) | ((v) & 0xff))

/* ... except for these macros, which yield the low and high bytes
   of an instruction.  */
#define low(insn)  ((insn) & 0xff)
#define high(insn) (((insn) >> 8) & 0xff)

#define OP_MOV    0xc  /* 001100    */
#define OP_SAVE   0x1e /* 011110    */
#define OP_SAVE_2 0x0  /* 00        */
#define OP_PFX    0x13 /* 10011     */
#define OP_SUBI   0x3  /* 000011    */
#define OP_STS    0x6  /* 110       */

/*
 * Instruction format parsers.
 */
static void
p_insn_RR (int i, int *op6, int *a5, int *b5)
{
    *op6 = (i >> 10) & 0x3f;
    *a5 = (i >> 5) & 0x1f;
    *b5 = i & 0x1f;
}

static void
p_insn_i8v (int i, int *op6, int *op2v, int *imm8v)
{
    *op6 = (i >> 10) & 0x3f;
    *op2v = (i >> 8) & 0x3;
    *imm8v = i & 0x3f;
}

static void
p_insn_i11 (int i, int *op5, int *i11)
{
    *op5 = (i >> 11) & 0x1f;
    *i11 = i & 0x7ff;
}

static void
p_insn_ri5 (int i, int *op6, int *i5, int *a5)
{
    p_insn_RR (i, op6, i5, a5);
}

static void
p_insn_ri8 (int i, int *op3, int *imm8, int *a5)
{
    *op3 = (i >> 13) & 0x7;
    *imm8 = (i >> 5) & 0xff;
    *a5 = i & 0x1f;
}

/*
 * Specific instruction parsers.  They return 1 if the parse was correct.
 * Using the insn_* macros to decode can result in headaches.
 */

static int
p_insn_sts (int i, int *imm8, int *a5)
{
    int op;

    p_insn_ri8 (i, &op, imm8, a5);

    return (OP_STS == op);
}

static int
p_insn_subi (int i, int *i5, int *a5)
{
    int op;

    p_insn_ri5 (i, &op, i5, a5);

    return (OP_SUBI == op);
}

static int
p_insn_mov (int i, int *a, int *b)
{
    int op;

    p_insn_RR (i, &op, a, b);

    return (OP_MOV == op);
}

static int
p_insn_save (int i, int *imm8)
{
    int op6, op2;

    p_insn_i8v (i, &op6, &op2, imm8);

    return (OP_SAVE == op6 && OP_SAVE_2 == op2);
}

static int
p_insn_pfx (int i, int *i11)
{
    int op;

    p_insn_i11 (i, &op, i11);

    return (OP_PFX == op);
}

/* Breakpoints.  */

static unsigned char *
nios_breakpoint_from_pc (CORE_ADDR *pcptr, int *lenp)
{
  /* We use `trap #e' for breakpoints.  */
#define insn_trap (insn_i5v (30, 1, 3))

  static unsigned char breakpoint[] = { low (insn_trap), high (insn_trap) };

  *lenp = sizeof (breakpoint);
  return breakpoint;
}



/* Backtraces.  */

/* A nios function's stack frame looks like this:
   (in this diagram, addresses increase as you go down; the stack grows
    downward in memory)

   before %sp      unreserved memory, might be overwritten at any time
		   if we get a signal, by the handler's frame
   %sp (%o6) ----> space to save this function's %l[0-7], %i[0-7]
                          (size: 16 * register size)
                   return struct ptr (one pointer)
                   spill slot area for outgoing arguments
                          (fixed size, determined at compile time)
                   local variables (size varies)
   %fp (%i6) ----> start of next frame: space to save caller's regs

   Each function's "save" instruction rotates the register window and
   initializes the fp from the sp, reserving the space for this
   invocation's stack frame.

   We lay out a function call as follows: we reserve one word for each
   argument in the "spill slot area".  The first six arguments are
   passed in %o0 -- %o5; any subsequent arguments are placed in their
   slots in the spill area.  The compiler sizes the spill slot area to
   accomodate the largest argument list the function will need for its
   calls.

   The spill slots for the first six arguments never get used, unless
   we're calling a varargs function.  A varargs function spills %o0 --
   %o5 into the slots reserved for them in the caller's frame, and
   then it has a nice, simple argument list, contiguous in memory and
   easy to walk with varargs macros.

   Sometimes, code compiled without optimization spills its register
   arguments back into their slots in the caller's frame, and then
   treats those memory slots as the official homes of the arguments.
   So, unoptimized code may use those slots, too.

   If the caller expects the callee to return a structure by value,
   the caller should store a pointer to a buffer for the return value
   in the "return struct ptr", just above the spill slot area.

   Don't ask me about long longs and things like that.  I don't know.
   Compile some code and figure it out.  */

/* The offset from a function's SP to where its register R is saved.
   Keep in mind that the callee's FP is the caller's SP.

   NOT FOR FLAT CODE.  */

#define SAVED_REG_OFFSET(r) \
  (((r) - first_local_regnum) * nios_register_raw_size (first_local_regnum))

/* For functions that return structures by value, this is the offset
   from the frame pointer to a word containing the address of the
   buffer for the returned structure, allocated by the caller.
   That word comes right after the register window save area.

   NOT FOR FLAT CODE?  */

#define RETURN_STRUCT_PTR_OFFSET (SAVED_REG_OFFSET (last_in_regnum + 1))
   

/* Disassemble the function prologue, and return the address of the
   first instruction after the prologue.

   At the moment, this function's specification is exactly the same as
   that of the skip_prologue function, so it looks kind of silly.
   However, in the long run, we may actually start extracting
   interesting information about the prologue, and then this function
   will be used by both the skip_prologue and init_extra_frame_info
   functions.  */
static CORE_ADDR
nios_analyze_prologue (CORE_ADDR pc)
{
  for (;;)
    {
      LONGEST op = read_memory_unsigned_integer (pc, 2);
      LONGEST next_op;

      /* save #V */
      if ((op & insn_i8v (-1, -1, 0)) == insn_i8v (30, 0, 0))
      {
          pc += 2;
      }
      else
      {
          break;
      }
    }

  verbose ("ANALYZE PROLOGUE");

  return pc;
}


static CORE_ADDR
nios_skip_prologue (CORE_ADDR pc)
{
  CORE_ADDR func_addr, func_end, new_pc;

  new_pc = pc;

  /* Find the line table entry for the function's entry point.  By
     convention, the code of that "line" is the function prologue.
     The address of the next line is where the body code starts.
     However, this information is not always available, or reliable,
     so do some sanity checks, and fall back to nios_analyze_prologue
     if necessary.  */

  /* Find the extent of the function containing PC.  */
  if (find_pc_partial_function (pc, NULL, &func_addr, &func_end))
    {
      struct symtab_and_line sal;

      /* Find the extent of the prologue "line" --- the one containing the
	 function's entry point.  */
      sal = find_pc_line (func_addr, 0);

      /* If that line ends within the function, then that's probably
	 the location of the first body instruction.  */
      if (sal.line != 0 && sal.end < func_end)
	new_pc = sal.end;
    }

  /* If all that machination didn't get us past the first instruction,
     then we probably didn't succeed in skipping the prologue.  Go and
     analyze the instruction stream, and see if we can guess what's
     going on.  */
  if (! (new_pc > func_addr))
  {
    new_pc = nios_analyze_prologue (func_addr);
  }

  return new_pc;
}

static int
read_code (CORE_ADDR pc)
{
    if (pc & 1)
    {
        internal_error ("bad pc alignment in read_code()");
    }

    return read_memory_unsigned_integer (pc, 2);
}

static int
nios_decode_flat_prologue (struct frame_info *frame, CORE_ADDR start)
/*
 * Try to decode a flat prologue of this form:
 *
 *        mov     %g1, %sp
 *        pfx     %hi(X)
 *        subi    %sp, %lo(X)
 *        sts     [%sp,Y], %fp
 *        sts     [%sp,Z], %rx
 *        ...
 *        mov     %fp, %g1
 */
{
    int insn, insn2, insn3, insn4;
    int imm8, rega, regb, i11 = 0, i8, sts_a5;
    int sub_i5, sub_a5;
    int stack_space = 0;

    /* Grab the 1st instruction */
    insn = read_code (start);

    if (p_insn_save (insn, &imm8))
    {
        frame->extra_info->save_words = imm8;
        /* Not flat */
        return 0;
    }

    start += 2;

    if (! p_insn_mov (insn, &rega, &regb))
    {
        /* Not a gcc flat compiled function? */
        return start;
    }

    if (! (rega == 14 && regb == 1))
    {
        return start;
    }

    /* Try to read a "pfx NUM" ... */
    insn2 = read_code (start);

    if (p_insn_pfx (insn2, &i11))
    {
        start += 2;
    }
    else
    {
        i11 = 0;
    }

    /* ... followed by a "subi %sp, NUM" */
    insn3 = read_code (start);
    start += 2;

    if (!p_insn_subi (insn3, &sub_i5, &sub_a5))
    {
        return start;
    }

    if (14 != sub_a5) /* %sp */
    {
        return start;
    }

    frame->extra_info->flat_stack_space = (i11 << 5) + sub_i5;

    while (1)
    {
        /* They may be a few "sts" followed by a "mov". */
        insn4 = read_code (start);
        start += 2;

        if (! p_insn_sts (insn4, &i8, &sts_a5))
        {
            break;
        }

        frame->extra_info->flat_sp_offsets[sts_a5] = i8;
    }

    return start;
}


static void nios_frame_init_saved_regs (struct frame_info *frame);
static void nios_init_extra_frame_info (int fromleaf, struct frame_info *frame);

static CORE_ADDR
nios_init_pc (struct frame_info *frame)
/*
 * Try to find a frame's own pc.
 */
{
    struct frame_info *next = frame->next;
    int reg_size = nios_register_raw_size (link_regnum);
    CORE_ADDR pc = 0, where;

    if (frame->pc)
    {
        verbose ("%s %x HAD pc %x",
                 FRAME_NAME(frame),
                 (int)FRAME_FP(frame),
                 (int)frame->pc);
       // return frame->pc;
    }

    /*
     * XXX: does it matter if 'frame' is flat?
     */

    if (next)
    {
        /* Look at the link register of the lower frame. */
        if (FRAME_FLAT(next))
        {
            where = next->saved_regs[link_regnum];
        }
        else
        {
            where = next->saved_regs[save_regnum];
        }

        if (where)
        {
            /* It's on the stack. */
            pc = read_memory_unsigned_integer(where, reg_size);
        }
        else
        {
            /* It's still in the register. The flat frame did not save it. */
            pc = read_register (link_regnum);
        }

        pc <<= 1;

        verbose ("INIT PC frame:%x pc:%x next:%x %s pc:%x where:%x",
                 (int)frame->frame,
                 (int)pc,
                 (int)next->frame,
                 FRAME_FLAT(next) ? "FLAT" : "ROUND",
                 (int)next->pc,
                 (int)where);
    }
    else
    {
        verbose ("INIT PC?");
    }

    return pc;
}

static void
nios_init_extra_frame_info (int fromleaf, struct frame_info *frame)
{
    CORE_ADDR pro_start, pro_end, flat_pro_end;
    char *name;

    frame->extra_info = frame_obstack_alloc (sizeof (struct frame_extra_info));
    bzero (frame->extra_info, sizeof (struct frame_extra_info));

    if (! find_pc_partial_function (frame->pc, &name, &pro_start, &pro_end))
    {
        frame->pc = nios_init_pc (frame);
    }

    if (find_pc_partial_function (frame->pc, &name, &pro_start, &pro_end))
    {
        flat_pro_end = nios_decode_flat_prologue (frame, pro_start);
        frame->extra_info->name = name ? name : "<unknown>";

        verbose ("EXTRA frame:%x %s startpc:%x %s %s",
                 (int)frame->frame,
                 name,
                 (int)pro_start,
                 FRAME_FLAT(frame) ? "flat" : "round",
                 FRAME_SIMPLE(frame) ? "frameless" : "");

        nios_frame_init_saved_regs (frame);
    }
    else
    {
        verbose ("Cannot find function for pc %x", (int)frame->pc);
    }
}


/* Fill in FRAME's saved_regs array.
   It's okay to call this more than once on the same frame; if
   saved_regs has already been filled in, we do nothing.  */
static void
nios_frame_init_saved_regs (struct frame_info *frame)
{
  int i, offset, reg_size = nios_register_raw_size(0);
  CORE_ADDR sp;
  struct frame_info *next = frame->next;

  /* Have we already set up this frame's saved register map?  */
  if (frame->saved_regs)
  {
      return;
  }

  frame_saved_regs_zalloc (frame);
  verbose ("INIT SAVED REGS %s %s",
           FRAME_NAME(frame),
           FRAME_FLAT(frame) ? "FLAT" : "ROUND");

  /* By convention, the stack pointer's slot in the saved_regs array
   holds the actual stack pointer, not the address at which it is
   saved.  */

  if (frame->next && ! FRAME_SIMPLE(frame->next))
  {
      /* The frame has a descendant.
         Simple frames don't count as descendants. */
      sp = FRAME_FP (frame->next);
  }
  else
  {
      /* Frame is the bottom frame. Simple frames are always bottom frames. */
      sp = read_register (sp_regnum);
  }

  if (FRAME_FLAT(frame))
  {
      for (i = 0; i < 32; i++)
      {
          offset = frame->extra_info->flat_sp_offsets[i] * reg_size;

          if (offset)
          {
              frame->saved_regs[i] = sp + offset;
              verbose ("%%%s@%x", nios_register_names[i], (int)frame->saved_regs[i]);
          }
      }
  }
  else
  {
      /* Upon entry, the caller's registers are saved at the caller's SP,
       which is our FP.  */

      for (i = first_local_regnum; i <= last_in_regnum; i++)
      {
          frame->saved_regs[i] = FRAME_FP (frame) + SAVED_REG_OFFSET (i);
      }
  }

  frame->saved_regs[sp_regnum] = sp;
  verbose ("%%sp:%x", (int)frame->saved_regs[sp_regnum]);
}

static int nios_get_register_value_before (struct frame_info *frame, int reg);

static int
nios_get_round_register (struct frame_info *frame, int reg)
/*
 * Get a register value for a round frame.  Because of the possibility
 * of a flat caller/callee, this is not simple.
 */
{
    struct frame_info *next = frame->next;
    int out = 0xdeadbeef, size;
    CORE_ADDR where;

    nios_frame_init_saved_regs (frame);
    size = nios_register_raw_size (reg);

    if (next)
    {
        nios_frame_init_saved_regs (next);
    }

    if (! next)
    {
        /* Still in the register. */
        out = read_register (reg);
    }
    else if (! FRAME_FLAT(next))
    {
        where = frame->saved_regs[reg];
        out = read_memory_unsigned_integer (where, size);
    }
    else
    {
        out = nios_get_register_value_before (next, reg);
    }

    return out;
}

static CORE_ADDR nios_frame_saved_pc (struct frame_info *frame);

static void
nios_guess_caller_frame (struct frame_info *frame, struct frame_info *caller)
{
    if (FRAME_FLAT (frame))
    {
        abort();
    }

    memset (caller, 0, sizeof (*caller));
    caller->next = frame;
    caller->pc = nios_frame_saved_pc (frame);
    verbose ("GUESS PC %x", (int)caller->pc);
    nios_init_extra_frame_info(0, caller);
    nios_frame_init_saved_regs (caller);
    verbose ("CALLER %s %s o7@%x",
             FRAME_NAME(caller),
             FRAME_FLAT(caller) ? "FLAT" : "ROUND",
             (int)caller->saved_regs[link_regnum]);
}

static int
nios_get_register_value_before (struct frame_info *frame, int reg)
/*
 * Find the value of a register before the frame's prologue is executed.
 */
{
    struct frame_info *next = frame->next, *prev = frame->prev;
    int out = 0xdeadbeef, size;
    CORE_ADDR where = 0;

    nios_frame_init_saved_regs (frame);

    if (next)
    {
        nios_frame_init_saved_regs (next);
    }

    size = nios_register_raw_size (reg);

    if (FRAME_FLAT(frame))
    {
        where = frame->saved_regs[reg];

        if (where)
        {
            /* It has explicitly saved the register. */
            out = read_memory_unsigned_integer (where, size);
        }
        else
        {
            if (next)
            {
                /* Try even lower on the stack. */
                out = nios_get_register_value_before (next, reg);
            }
            else
            {
                /* It must be in the register. */
                out = read_register (reg);
            }
        }
    }
    else if ( ! FRAME_FLAT (frame))
    {
        /*
         * It's round.  If there is a round function further down in the call
         * tree, then %i and %l regs may be in the save area above that lower
         * round function.
         *
         * A flat function down in the tree may also save regs.
         *
         * Or, by default, it may be live in the register.
         */
        if (reg <= 15 && reg >= 8)
        {
            /* An %o register, really an %i now. */
            where = frame->saved_regs[reg + 16];

        }
        else if (reg == 30) /* %fp a.k.a %i6 */
        {
            struct frame_info caller;

            nios_guess_caller_frame (frame, &caller);

#if 0 /* Will not work with alloca. */
            if (FRAME_FLAT(&caller))
            {
                out = FRAME_FP(frame) + caller.extra_info->flat_stack_space;
                verbose ("FLAT SPACE %d", caller.extra_info->flat_stack_space);
            }
            else
            {
                out = FRAME_FP(frame) + caller.extra_info->save_words * 4;
                verbose ("SAVE BYTES %d", caller.extra_info->save_words * 4);
            }
#else
            if (FRAME_FLAT(&caller))
            {
                /* Callee round, caller flat */
                where = caller.saved_regs[fp_regnum];
                out = read_memory_unsigned_integer (where, size);
                verbose ("XXX %x at %x space %d",
                         (int)out,
                         (int)where,
                         caller.extra_info->flat_stack_space);
            }
            else
            {
                /* Caller and callee are round. */
                out = read_memory_unsigned_integer
                    (
                     FRAME_FP (frame)
                     + SAVED_REG_OFFSET (fp_regnum),
                     nios_register_raw_size (fp_regnum)
                    );
            }
#endif

            verbose ("GUESS PREV reg %d %x, %s save %d",
                     reg,
                     out,
                     FRAME_NAME(&caller),
                     caller.extra_info->save_words);
        }
        else if (reg <= 31 && reg >= 24) /* %i other than %fp */
        {
            /* An %i register, the prev frame's %o */
            if (! prev)
            {
                verbose ("OOPS %s %x",
                         nios_register_names[reg],
                         (int)frame->pc);
            }
            else
            {
                verbose ("HAVE PREV %s - %s",
                         FRAME_NAME(frame),
                         FRAME_NAME(prev));
                where = prev->saved_regs[reg - 16];
            }
        }
        else
        {
            verbose ("OOPS %s reg %d", __FUNCTION__, reg);
            where = frame->saved_regs[reg];
        }

        if (where)
        {
            out = read_memory_unsigned_integer (where, size);
        }
    }

    return out;
}


static CORE_ADDR
old_nios_frame_chain (struct frame_info *frame)
{
  /* If the callee's frame is at FRAME_FP (frame), then that's
     also the address at which the caller's registers are saved.  */
  return ((CORE_ADDR)
	  read_memory_unsigned_integer (FRAME_FP (frame)
					+ SAVED_REG_OFFSET (fp_regnum),
					nios_register_raw_size (fp_regnum)));
}


static CORE_ADDR
nios_frame_chain (struct frame_info *frame)
/* Given a GDB frame, determine the address of the calling function's frame. */
{
    CORE_ADDR addr, old;

    nios_frame_init_saved_regs (frame);

    if (FRAME_SIMPLE(frame))
    {
        /* Our frame is the callers frame. */
        addr = read_register (fp_regnum);
    }
    else
    {
        addr = nios_get_register_value_before (frame, fp_regnum);
        old = old_nios_frame_chain (frame);

        verbose ("CHAIN %s (%s) %x old %x %s",
                 FRAME_NAME(frame),
                 FRAME_FLAT(frame) ? "FLAT" : "ROUND",
                 (int) addr,
                 (int) old,
                 old == addr ? "OK" : "********");
    }

    return addr;
}


static CORE_ADDR
nios_saved_pc_after_call (struct frame_info *frame)
{
  CORE_ADDR pc;
  /* Immediately after a call, the return address is in LINK.  The
     `save' instruction will move it to the SAVE register, but we
     haven't done that yet.
     And don't forget, the Nios saves the PC shifted right one bit...  */


  nios_frame_init_saved_regs (frame);
  pc = read_register (link_regnum);
  pc <<= 1;
  verbose ("AFTER CALL 0x%x", (int)pc);

  return pc;
}


static CORE_ADDR
old_nios_frame_saved_pc (struct frame_info *frame)
{
  CORE_ADDR pc, where;

  if (frame->next)
    {
      /* For all but the top frame, we extract the saved PC from the
	 frame's register window save area.  Note that *our* registers
	 are saved at the addresses given in the *next* frame's
	 `saved_regs' array.

	 We don't use the size of the PC register here; that's
	 deliberate.  According to our simulator, the PC is always 32
	 bits long, on both the Nios-16 and the Nios-32.  We're
	 actually retrieving the saved contents of the SAVE register,
	 so that's whose size we use.  */
      struct frame_info *next = frame->next;
      int size = nios_register_raw_size (save_regnum);
      nios_frame_init_saved_regs (next);
      where = next->saved_regs[save_regnum];

      if (where)
      {
          pc = read_memory_unsigned_integer (where, size);
      }
      else
      {
          pc = nios_get_register_value_before (next, link_regnum);
      }
    }
  else
  {
    /* For the youngest frame, the return address is still in $i7.  */
    pc = read_register (save_regnum);
  }

  return pc;
}



static CORE_ADDR
nios_frame_saved_pc (struct frame_info *frame)
/* Get the frame's return address. */
{
    CORE_ADDR pc;

    if (FRAME_SIMPLE(frame))
    {
        pc = read_register (link_regnum);
    }
    else if (FRAME_FLAT(frame))
    {
        pc = nios_get_register_value_before (frame, link_regnum);
    }
    else
    {
        pc = old_nios_frame_saved_pc (frame);
    }

    pc <<= 1;

    verbose ("SAVED PC of %s is 0x%x %s %s",
             FRAME_NAME (frame),
             (int) pc,
             FRAME_FLAT(frame) ? "FLAT" : "ROUND",
             FRAME_SIMPLE (frame) ? "frameless" : ""
            );

    return pc;
}


static CORE_ADDR
nios_frame_address (struct frame_info *frame)
{
    CORE_ADDR f = frame->frame;

    verbose ("FRAME ADDRESS %s %x", FRAME_NAME(frame), (int) f);
    return f;
}


/* Stack unwinding.  */

static void
nios_pop_flat_frame (struct frame_info *frame)
{
    verbose ("POP FLAT");
}

static void
nios_pop_frame_regular (struct frame_info *frame)
{
  int i;

  verbose ("POP REGULAR %s", FRAME_NAME(frame));
  nios_frame_init_saved_regs (frame);

  if (FRAME_FLAT(frame))
  {
      nios_pop_flat_frame (frame);

      return;
  }

  /* Restore the PC from the SAVE register.  */
  write_pc (read_register (save_regnum) << 1);

  /* Rotate back to the caller's window.  */
  {
    int cwp = nios_get_cwp ();
    int hi_limit = nios_get_hi_limit ();

    if (cwp < hi_limit)
      nios_set_cwp (cwp + 1);
  }

  /* Restore the caller's register values from the stack.  */
  for (i = 0; i < nios_num_regs; i++)
    if (i != sp_regnum		/* sp is special, by GDB tradition */
	&& frame->saved_regs[i])
      write_register
	(i, read_memory_unsigned_integer (frame->saved_regs[i],
					  nios_register_raw_size (i)));

  /* The caller's FP is the callee's FP.  */
  write_register (sp_regnum, FRAME_FP (frame));

  flush_cached_frames ();
}

static void
nios_pop_frame (void)
{
    verbose ("POP FRAME\b");
  /* This function checks for and handles generic dummy frames, and
     calls back to our function for ordinary frames.  */
  generic_pop_current_frame (nios_pop_frame_regular);
}

static void
nios_store_return_value (struct type *type, char *valbuf)
{
  int len = TYPE_LENGTH (type);

  /* Any value that fits in a word, we return in %o0.  */
  if (len <= nios_register_raw_size (return_regnum))
    write_register_bytes (nios_register_byte (return_regnum), valbuf, len);
  else
    {
      /* Values that don't fit in a word, we return using the struct
	 convention.  (So this function could actually handle
	 returning struct values, if the rest of GDB were prepared!)

	 By the time this function is called, we've already restored
	 the registers of the frame we're returning to.  So we can
	 find the pointer to the buffer for the return address
	 relative to the current SP.  */
      CORE_ADDR ptr_size = nios_register_raw_size (sp_regnum);
      CORE_ADDR buf_ptr = read_register (sp_regnum) + RETURN_STRUCT_PTR_OFFSET;
      CORE_ADDR buf = read_memory_unsigned_integer (buf_ptr, ptr_size);
      
      /* Now buf is the address where we should store the return value.  */
      write_memory (buf, valbuf, len);
    }
}


static int
nios_return_value_on_stack (struct type *type)
{
  /* Everything that fits in a register, we return in a register.
     Everything else, we return using the struct convention.  */
  return TYPE_LENGTH (type) > nios_register_raw_size (return_regnum);
}


static int
nios_use_struct_convention (int gcc_p, struct type *type)
{
  /* Same rules as above.  */
  return nios_return_value_on_stack (type);
}

static void
nios_extract_return_value (struct type *type, char *regbuf, char *valbuf)
{
  memcpy (valbuf, regbuf + nios_register_byte (first_out_regnum),
	  TYPE_LENGTH (type));
}


static CORE_ADDR
nios_extract_struct_value_address (char *regbuf)
{
  /* On the Nios, %o0 points to the buffer.  Remember that we're
     getting %o0's value from regbuf, not the live registers.  */
  char *reg_bits = regbuf + nios_register_byte (return_regnum);
  int reg_size = nios_register_raw_size (return_regnum);
  CORE_ADDR buf = extract_unsigned_integer (reg_bits, reg_size);
      
  return buf;
}



/* Inferior function calls.  */

static LONGEST nios_call_dummy_words[] =
{0};


/* Round down ADDR to the nearest multiple of UNIT.  */
static int
round_down (CORE_ADDR addr, int unit)
{
  return addr - (addr % unit);
}

/* Round up VALUE to the nearest multiple of UNIT.  */
static int
round_up (int value, int unit)
{
  value = value + unit - 1;
  return value - (value % unit);
}

static CORE_ADDR
nios_push_arguments (int nargs, value_ptr *args, CORE_ADDR sp,
		      int struct_return, CORE_ADDR struct_addr)
{
  /* This function is responsible for setting up the stack frame and
     call context.  We are aided by:

     - generic_push_dummy_frame, which saves the enclosing register state
		in a generic dummy frame
     - nios_return_value_on_stack and generic_use_struct_convention_always,
		which tell GDB to allocate a buffer for the return value
     - nios_push_return_address, which stores the right thing in the
		LINK register
     - hand_function_call itself, which will write the new sp returned by
		this function and nios_push_return_address,
     - run_stack_dummy, which sets up the pc.

     So this function is responsible for laying out the stack in
     memory, and putting the proper values in %o0 -- %o5.

     Our new frame is going to look like this:
     (in this diagram, addresses increase as you go down; the stack grows
     downward in memory)

     dummy frame %sp ->	save area for dummy frame's `local' and `in'
				registers.  These are really saved in
				the generic dummy frame within GDB,
				but the window spill handler doesn't
				know that. (size: 16 * register size)
			return struct ptr (size: one pointer)
			outgoing argument area
				size: at least six registers long, possibly
				longer if the argument list requires it
			buffers for aggregates passed by value
			buffer for return value (created by hand_function_call)
			nios_call_dummy_words - this only exists because
				we're afraid of the effect of a zero-length
				call dummy.  (size: one instruction)
     dummy frame %fp -> start of last real inferior frame

     The nios passes anything four words or less in length directly
     in the argument area / registers.  Anything larger than four
     words, it passes by reference.  */

  int i;
  int register_size = REGISTER_SIZE;
  int pointer_size = (TARGET_PTR_BIT / TARGET_CHAR_BIT);

  /* Sizes of various components of the frame.  */

  /* The total size of all buffers allocated to hold aggregates being
     passed and returned by value.  */
  int aggregate_buf_size = 0;

  /* The total size of the argument area area.  We actually compute this
     by pretending every argument is being passed in memory, and then
     bumping it up to the minimum six-register size if necessary.  */
  int argument_area_size = 0;

  /* The size of the register save area, plus the return struct pointer.  */
  int register_save_size = 16 * register_size + pointer_size;

  /* Addresses of those components.  */
  CORE_ADDR aggregate_buf_start;
  CORE_ADDR argument_area_start;
  CORE_ADDR register_save_start;

    verbose ("PUSH ARGS");

  for (i = 0; i < nargs; i++)
    {
      struct type *arg_type = VALUE_TYPE (args[i]);
      int arg_size = TYPE_LENGTH (arg_type);

      /* If it fits in four words, it's passed by directly in the
         argument area.  Otherwise, it's passed by reference.  */
      if (arg_size <= 4 * register_size)
	argument_area_size += round_up (arg_size, register_size);
      else
	{
	  /* The strictest alignment required is to a register size
             boundary.  */
	  aggregate_buf_size += round_up (arg_size, register_size);
	  argument_area_size += round_up (pointer_size, register_size);
	}

      if (argument_area_size % register_size)
	internal_error ("nios-tdep.c (nios_push_arguments): "
			"spill slots have become unaligned");
      
      if (aggregate_buf_size % register_size)
	internal_error ("nios-tdep.c (nios_push_arguments): "
			"aggregate buffers have become unaligned");
    }

  /* The spill slot area must always be at least large enough to hold
     %o0 -- %o5.  */
  if (argument_area_size < 6 * register_size)
    argument_area_size = 6 * register_size;

  /* Now that we know everything's size, compute their addresses.
     Make sure we start with an aligned SP.  */
  sp = round_down (sp, register_size);
  aggregate_buf_start = sp - aggregate_buf_size;
  argument_area_start = aggregate_buf_start - argument_area_size;
  register_save_start = argument_area_start - register_save_size;

  /* For simplicity, we write everything to memory, and then load the
     `out' registers from that memory.  */

  /* Place the arguments in their proper places on the stack.  */
  {
    /* The end of the buffer for the next aggregate passed by value.  */
    CORE_ADDR aggregate_ptr = aggregate_buf_start;

    /* The address of the next slot in the argument area.  */
    CORE_ADDR arg_ptr = argument_area_start;

    for (i = 0; i < nargs; i++)
      {
	struct type *arg_type = VALUE_TYPE (args[i]);
	int arg_size = TYPE_LENGTH (arg_type);
	char *contents = VALUE_CONTENTS (args[i]);

	/* If it fits in four words, it's passed by directly in the
	   argument area.  Otherwise, it's passed by reference.  */
	if (arg_size <= 4 * register_size)
	  {
	    write_memory (arg_ptr, contents, arg_size);
	    arg_ptr += round_up (arg_size, register_size);
	  }
	else
	  {
	    /* First, store the pointer to the copy in the argumement
               area.  */
	    char *buf = alloca (pointer_size);
	    store_address (buf, pointer_size, aggregate_ptr);
	    write_memory (arg_ptr, buf, pointer_size);
	    arg_ptr += round_up (pointer_size, register_size);

	    /* Then, fill in the copy.  */
	    write_memory (aggregate_ptr, contents, arg_size);
	    aggregate_ptr += round_up (arg_size, register_size);
	  }
      }

    /* Make sure everything filled the space allocated.  */
    if (arg_ptr > argument_area_start + argument_area_size)
      internal_error ("nios-tdep.c (nios_push_arguments): "
		      "argument area size is incorrect");
    if (aggregate_ptr != aggregate_buf_start + aggregate_buf_size)
      internal_error ("nios-tdep.c (nios_push_arguments): "
		      "aggregate buffer size is incorrect");
  }

  /* Load the `out' registers from the stack.  */
  {
    int shadow_size = 6 * register_size;
    char *buf = alloca (shadow_size);
    read_memory (argument_area_start, buf, shadow_size);
    write_register_bytes (nios_register_byte (first_out_regnum),
			  buf, shadow_size);
  }

  /* Store the pointer to the struct return buffer in the right place.  */
  if (struct_return)
    {
      char *buf = alloca (pointer_size);
      store_address (buf, pointer_size, struct_addr);
      write_memory (register_save_start + RETURN_STRUCT_PTR_OFFSET,
		    buf, pointer_size);
    }

  /* Like any other frame, the dummy frame's SP is the address of its
     register save area.  */
  return register_save_start;
}


static CORE_ADDR
nios_push_return_address (CORE_ADDR pc, CORE_ADDR sp)
{
    CORE_ADDR tmp = CALL_DUMMY_ADDRESS ();

    verbose ("PUSH RETURN pc:0x%x sp:0x%x", (int) pc, (int) sp);

    write_register (link_regnum, tmp >> 1);

    return sp;
}

static void
nios_store_struct_return (CORE_ADDR addr, CORE_ADDR sp)
{
  /* We don't need to do anything; nios_push_arguments takes care of
     this for us.  */
}


/* Initialization: this module, and the nios arch object.  */

/* When we see an executable whose BFD architecture is bfd_arch_nios,
   the gdbarch machinery will call this function to get an appropriate
   `struct gdbarch' for the executable.

   The nios comes in two variants --- 16-bit registers and 32-bit
   registers --- and we have a separate `struct gdbarch' for each of
   them.  

   INFO is all the info GDB could gather from the executable file.
   ARCHES is a cache of the `struct gdbarch' objects we've built
   for bfd_arch_nios so far; we check it first, to see if we've
   already built an appropriate architecture object.  */

static struct gdbarch *
nios_gdbarch_init (struct gdbarch_info info, struct gdbarch_list *arches)
{
  struct gdbarch *gdbarch;
  struct gdbarch_tdep *tdep;

  /* Check to see if we've already built an appropriate architecture
     object for this executable.  */
  arches = gdbarch_list_lookup_by_info (arches, &info);
  if (arches)
    return arches->gdbarch;

  /* Make sure it's a machine we support.  */
  if (info.bfd_arch_info->mach != bfd_mach_nios16
      && info.bfd_arch_info->mach != bfd_mach_nios32)
    return 0;

  /* Nope --- we need to build a new one.  */
  tdep = xmalloc (sizeof (*tdep));
  memset (tdep, 0, sizeof (*tdep));
  gdbarch = gdbarch_alloc (&info, tdep);

  /* With rational default values for these things, a lot of this
     would be unnecessary.  */

  if (info.bfd_arch_info->mach == bfd_mach_nios16)
    {
      set_gdbarch_register_size (gdbarch, 2);
      set_gdbarch_ptr_bit       (gdbarch, 16);
      set_gdbarch_int_bit       (gdbarch, 16);
    }
  else if (info.bfd_arch_info->mach == bfd_mach_nios32)
    {
      set_gdbarch_register_size (gdbarch, 4);
      set_gdbarch_ptr_bit       (gdbarch, 32);
      set_gdbarch_int_bit       (gdbarch, 32);
    }
  else 
    {
      gdbarch_free (gdbarch);
      return 0;
    }

  /* Set up values common to both architectures.  */

  /* Data type sizes.  */
  set_gdbarch_short_bit       (gdbarch, 16);
  set_gdbarch_long_bit        (gdbarch, 32);
  set_gdbarch_long_long_bit   (gdbarch, 64);
  set_gdbarch_float_bit       (gdbarch, 32);
  set_gdbarch_double_bit      (gdbarch, 64);
  set_gdbarch_long_double_bit (gdbarch, 64);
  set_gdbarch_ieee_float      (gdbarch, 1);

  /* The nios uses odd pointer formats.  */
  set_gdbarch_address_to_pointer (gdbarch, nios_address_to_pointer);
  set_gdbarch_pointer_to_address (gdbarch, nios_pointer_to_address);

  /* The register set.  */
  tdep->registers = make_register_map (info.bfd_arch_info->mach);
  set_gdbarch_num_regs  (gdbarch, nios_num_regs);
  set_gdbarch_sp_regnum (gdbarch, sp_regnum);
  set_gdbarch_fp_regnum (gdbarch, fp_regnum);
  set_gdbarch_pc_regnum (gdbarch, pc_regnum);
  set_gdbarch_register_name (gdbarch, nios_register_name);
  set_gdbarch_max_register_raw_size (gdbarch, 4);
  set_gdbarch_max_register_virtual_size (gdbarch, 4);
  set_gdbarch_register_bytes (gdbarch, tdep->registers->total_size);
  set_gdbarch_register_raw_size (gdbarch, nios_register_raw_size);
  set_gdbarch_register_byte (gdbarch, nios_register_byte);
  set_gdbarch_register_virtual_type (gdbarch, nios_register_virtual_type);
  set_gdbarch_register_virtual_size (gdbarch, nios_register_virtual_size);
  set_gdbarch_register_convertible (gdbarch, nios_register_convertible);
  set_gdbarch_register_convert_to_virtual (gdbarch, nios_convert_to_virtual);
  set_gdbarch_register_convert_to_raw (gdbarch, nios_convert_to_raw);

  set_gdbarch_read_pc (gdbarch, generic_target_read_pc);
  set_gdbarch_write_pc (gdbarch, generic_target_write_pc);
  set_gdbarch_read_fp (gdbarch, generic_target_read_fp);
  set_gdbarch_write_fp (gdbarch, generic_target_write_fp);
  set_gdbarch_read_sp (gdbarch, generic_target_read_sp);
  set_gdbarch_write_sp (gdbarch, generic_target_write_sp);

  /* Breakpoints.  */
  set_gdbarch_decr_pc_after_break (gdbarch, 0);
  set_gdbarch_function_start_offset (gdbarch, 0);
  set_gdbarch_breakpoint_from_pc (gdbarch, nios_breakpoint_from_pc);

  /* Backtraces.  */
//  set_gdbarch_get_saved_register (gdbarch, default_get_saved_register);
  set_gdbarch_get_saved_register (gdbarch, generic_get_saved_register);
  set_gdbarch_init_extra_frame_info (gdbarch, nios_init_extra_frame_info);
  set_gdbarch_inner_than (gdbarch, core_addr_lessthan);
  set_gdbarch_frame_chain (gdbarch, nios_frame_chain);
  set_gdbarch_frame_chain_valid (gdbarch, generic_func_frame_chain_valid);
  set_gdbarch_frame_init_saved_regs (gdbarch, nios_frame_init_saved_regs);
  set_gdbarch_frame_saved_pc (gdbarch, nios_frame_saved_pc);
  set_gdbarch_saved_pc_after_call (gdbarch, nios_saved_pc_after_call);
  set_gdbarch_frameless_function_invocation (gdbarch, 
					     frameless_look_for_prologue);
  set_gdbarch_frame_num_args (gdbarch, frame_num_args_unknown);
  set_gdbarch_frame_args_address (gdbarch, nios_frame_address);
  set_gdbarch_frame_locals_address (gdbarch, nios_frame_address);
  set_gdbarch_skip_prologue (gdbarch, nios_skip_prologue);

  /* Stack unwinding.  */
  set_gdbarch_pop_frame (gdbarch, nios_pop_frame);
  set_gdbarch_store_return_value (gdbarch, nios_store_return_value);
  set_gdbarch_return_value_on_stack (gdbarch, nios_return_value_on_stack);
  set_gdbarch_use_struct_convention (gdbarch, nios_use_struct_convention);
  set_gdbarch_extract_return_value (gdbarch, nios_extract_return_value);
  set_gdbarch_extract_struct_value_address
    (gdbarch, nios_extract_struct_value_address);

  /* Inferior calls.  */
  set_gdbarch_call_dummy_p (gdbarch, 1);
  set_gdbarch_call_dummy_location (gdbarch, AT_ENTRY_POINT);
  set_gdbarch_call_dummy_address (gdbarch, entry_point_address);
  set_gdbarch_call_dummy_words (gdbarch, nios_call_dummy_words);
  set_gdbarch_sizeof_call_dummy_words (gdbarch,
				       sizeof (nios_call_dummy_words));
  set_gdbarch_use_generic_dummy_frames (gdbarch, 1);
  set_gdbarch_call_dummy_length (gdbarch, 0);
  set_gdbarch_coerce_float_to_double (gdbarch, 
				      standard_coerce_float_to_double);
  set_gdbarch_push_arguments (gdbarch, nios_push_arguments);
  set_gdbarch_push_return_address (gdbarch, nios_push_return_address);
  set_gdbarch_store_struct_return (gdbarch, nios_store_struct_return);
  set_gdbarch_call_dummy_breakpoint_offset_p (gdbarch, 1);
  set_gdbarch_call_dummy_breakpoint_offset (gdbarch, 0);
  set_gdbarch_call_dummy_start_offset (gdbarch, 0);
  set_gdbarch_pc_in_call_dummy (gdbarch, generic_pc_in_call_dummy);
  set_gdbarch_call_dummy_stack_adjust_p (gdbarch, 0);
  set_gdbarch_push_dummy_frame (gdbarch, generic_push_dummy_frame);
  set_gdbarch_fix_call_dummy (gdbarch, generic_fix_call_dummy);

  set_gdbarch_remote_translate_xfer_address
    (gdbarch, generic_remote_translate_xfer_address);

  return gdbarch;
}


void
_initialize_nios_tdep (void)
{
  register_gdbarch_init (bfd_arch_nios, nios_gdbarch_init);

  tm_print_insn = print_insn_nios;

  /* Some sanity checks.  */
  if (sizeof (nios_register_names) / sizeof (nios_register_names[0])
      != nios_num_regs)
    error ("_initialize_nios_tdep: bad size for register name array");
}


/* Remove all this when we make a public release of the nios port: */
/* Local variables: */
/* change-log-default-name: "ChangeLog.Cygnus" */
/* End: */
